/************************************************************************
 *																		*
 *			file booter.c												*
 *			author: Marco Kesseler										*
 *																		*
 ************************************************************************/

#include "booter.h"

#ifndef _TYPES_
#include "types.h"
#endif

#ifndef _SYSTEM_
#include "system.h"
#endif

#ifndef _HARDWARE_
#include "hardware.h"
#endif

#include "macros.h"
#include "netboot.h"
#include "connect.h"
#include "errorno.h"
#include "messageno.h"
#include "scanner.h"
#include "memory.h"
#include "graphs.h"
#include "hostserver.h"
#include "instructions.h"
#include "debugger.h"
#include "io.h"
#include "loader.h"
#include "kernel.h"
#include "heapalloc.h"
#include "packets.h"

#ifdef __INCLUDE_STANDARD__

#include <string.h>
#include <time.h>
#include <node.h>

#endif

extern OPTIONS			options;
extern NODE_INFO		node_info;
extern NET_INFO			net_info;
extern PROCESSOR_INFO	processor_info;

extern SYS				sys;

/************************************************************************************
 *																					*
 *		routines to intialise memory												*
 *																					*
 ************************************************************************************/

static BOOL initialise_memory(VOID)
{
	VRED_ON_CHIP_DATA *on_chip;
	
	/* This routine is only executed during startup; it initialises permament
	   data structures of the runtime system. Additionally it allocates the
	   heap at the end of the available memory. In this way it is possible to
	   change the size of the heap and everyting in it during runtime. To avoid
	   having to implement all kinds of exceptions the heap is kept in a
	   consistent state as much as possible
	*/

#if (VariableArraySize != 0)
	on_chip=push_int_mem(sizeof(VRED_ON_CHIP_DATA)+(node_info.max_reducers-1)*sizeof(REDUCER_LOCALS));
#else
	on_chip=push_int_mem(sizeof(VRED_ON_CHIP_DATA)+node_info.max_reducers*sizeof(REDUCER_LOCALS));
#endif

	if (on_chip == NULL)
		return False;

	/* intialise some pseudo-hard addresses */
	
	chreset(hard_signal_address());
	on_chip->root						=&sys.root;
	on_chip->descriptors				=NULL;			/* initialised by prelude of application */

	/* use macros sys_root and sys_descriptors instead of sys.root and sys.descriptors */
		
	sys.debug.processor_state			=&sys.processor_state;
	sys_root->comm_info					=&sys.comm_info;
	sys_root->reducer_info				=&sys.reducer_info;
	sys_root->general_info				=&sys.general_info;
	sys_root->copy_info					=&sys.copy_info;
	sys_root->descriptor_info			=&sys.descriptor_info;
	sys_root->code_info					=&sys.code_info;

	/* stacks grow from high to low */
	
	sys_root->system_stack				=((ADDRESS)&on_chip->system_stack)+sizeof(SYSTEM_STACK);	
	sys_root->process_stack				=((ADDRESS)&on_chip->process_stack)+sizeof(PROCESS_STACK);	
		
	chreset(&sys.debug.signal_channel);
	
	/* set flags for reducers */
	
	sys.general_info.runtime_flags		= 0 | options.basic_value_only;
	sys.reducer_info.init_stack_size	= options.init_stack_size;		/* in words */
	sys.reducer_info.init_heap_size		= options.init_heap_size;		/* in words */
	sys.reducer_info.ext_heap_size		= options.increase_heap_size;	/* in words */
		
	/****************************************************************
	 *																*
	 *	initialisation of the pseudo-random number generator		*
	 *	normally some unpredictable seed is used, but when a fixed	*
	 *	one is needed the option -f can be used						*
	 *																*
	 *	See: ACM Transactions on Mathematical Software				*
	 *	december 1987												*
	 *	pp 368-381. Author: A. Haas									*
	 *																*
	 ****************************************************************/
	
	if (!options.fixed_random)
	{
		/* initialised only once */
		
		sys.general_info.random_rand		= (4181+clock()+node_info.id)%10000;
		sys.general_info.random_m			= 971;
		sys.general_info.random_a			= 11113;
		sys.general_info.random_b			= 104322;
	}

	processor_info.id=node_info.id;

	/* initialise static global network info */
	
	sys.comm_info.transputer_id	= node_info.id;
	sys.comm_info.n_transputers	= net_info.size;
	sys.comm_info.dest_info		= node_info.rout_data;

	/* indicate there is no heap yet */
	
	invalid_heap();

	return True;
}

/************************************************************************************
 *																					*
 *		main booter routines														*
 *																					*
 ************************************************************************************/

static VOID set_scheduling_bias(VOID)
{
	WORD i;

	/*	determine speed of scheduling,	
		consider 'high' macro: it uses a word in internal memory,
		stack should be off-chip to be most reliable
	*/
		
	node_info.scheduling_bias=0;
	for (i=0; i<10; ++i)
	{
		UWORD sched_time;
		
		sched_time=ldtimer();
		low();
		
		/*	...this is the time we are measuring: the time needed to schedule
			the additional scheduler, so that this time is ignored in computing
			the idle time... */
		
		high();
		node_info.scheduling_bias+=(ldtimer()-sched_time);
	}
	node_info.scheduling_bias=node_info.scheduling_bias/10;
}

static VOID booter (VOID)
{
	/************************************************************************
	 *																		*
	 *		booter started at high priority									*
	 *																		*
	 ************************************************************************/
		
	set_scheduling_bias();
		
	/************************************************************************
	 *																		*
	 *		allocate and initialise memory structures						*
	 *																		*
	 ************************************************************************/

	if (!initialise_memory())
		error_number(EInitSystem);

	if (!initialise_packets(100))
		error_number(EInitRouter);
	
	if (!initialise_scanner())
		error_number(EInitSystem);
	
	if (!initialise_debugger())
		error_number(EInitDebug);
	
	if (!initialise_io())
		error_number(EInitIO);

	if (!initialise_loader())
		error_number(EInitIO);
		
	if (!initialise_graph_manager())
		error_number(EInitHeap);
		
	if (!initialise_instructions())
		error_number(EInitDebug);
	
	if (node_info.id == RootId && !init_ioserver(100))
		error_number(EInitIO);
				
	/****************************************************************************
	 *																			*
	 *		basic initialisation succeeded, tell host about it					*
	 *																			*
	 ****************************************************************************/
	
	error_number(ENoErr);
	
	/****************************************************************************
	 *																			*
	 *		start up router														*
	 *																			*
	 ****************************************************************************/
		
	if(!start_packets())
	{
		/* if this does not work one has to route information to the host
		   explicitly and make it appear on the screen. This has not been
		   implemented.
		*/
		
		if (node_info.id == RootId)
			error_number (EInitRouter);
		else
			procstop();
	}
	
	/****************************************************************************
	 *																			*
	 *		router started, all transputers have access to any other processor	*
	 *		in the network. In particular this means they can print on the		*
	 *		screen.																*
	 *																			*
	 ****************************************************************************/
	
	/****************************************************************************
	 *																			*
	 *		booted, start kernel												*
	 *																			*
	 ****************************************************************************/
		
	kernel();
}

/********************************************************************************
 *																				*
 *		C entry point															*
 *																				*
 ********************************************************************************/

int main (WORD argc, CHAR *argv[])
{
	/*	this routine has its stack allocated in internal memory (use link
		flag too ensure it takes up the entire internal memory). It
		performs basic intialisation (for which its stack should be
		sufficient) and then starts up a new booter process and finally
		dies without releasing the internal memory, so that it is free
		to be used by anyone. */

	/************************************************************************
	 *																		*
	 *		start with basic tests (for processor type)						*
	 *																		*
	 *		at this moment only T800 is supported							*
	 *																		*
	 ************************************************************************/
	
	if (_Node.memstart != UsrMemStart)
		return;

	/************************************************************************
	 *																		*
	 *		go to high prio (note this will write a value in low			*
	 *		internal memory													*
	 *																		*
	 ************************************************************************/
	
	high();

	/************************************************************************
	 *																		*
	 *		boot rest of network											*
	 *																		*
	 ************************************************************************/
	
	netboot(argc,argv);
	
	/************************************************************************
	 *																		*
	 *		memory has been 'reserved' by netboot routine already, 			*
	 *		be careful not to use (initialise) internal memory in this		*
	 *		routine as the stack is in it.									*
	 *																		*
	 ************************************************************************/

	/************************************************************************
	 *																		*
	 *		connect to neighbours											*
	 *																		*
	 ************************************************************************/
	
	connect_neighbours();
	
	/************************************************************************
	 *																		*
	 *		compute routeing tables											*
	 *																		*
	 ************************************************************************/
	
	if (!compute_route_tables())
		return;

	/****************************************************************************
	 *																			*
	 *		print basic system info												*
	 *																			*
	 ****************************************************************************/

	show_system_parameters();
	
	push_process(booter,4*KiloByte,0);
	procstop();	
}
