#include "commands.h"

#include "messageno.h"
#include "macros.h"
#include "scanner.h"
#include "memory.h"
#include "io.h"
#include "packets.h"

#ifdef __INCLUDE_STANDARD__

#include <ctype.h>
#include <stddef.h>
#include <string.h>

#endif

extern NODE_INFO node_info;

/********************************************************************************
 *																				*
 *		global variables														*
 *																				*
 ********************************************************************************/

/*	the command line is a global variable, as normally only one process will be
	reading commands. This avoids the need to allocate buffers on the stack or
	in the heap (with all problems of stack-overflow or failing mallocs)
	
	Separate words of this line are 'scanned' into the global string
	'command_word' (which is globally known for the same reason as the
	global command line)
	
	some predefined command routines here work on these global data structure
*/

static COMMAND_LINE	command_line;
static CHAR			command_word[StringLength];

/********************************************************************************
 *																				*
 *		command table construction												*
 *																				*
 ********************************************************************************/
	
static WORD command_hash (CHAR *s)
{
	CHAR *cp;
	WORD h;
	
	/* add the character values of the string */
	
	for (cp = s, h = 0; *cp; cp++)
		h += (WORD) *cp;
	
	return h;
}

VCOMMAND_TABLE *get_command_table(WORD size)
{
	VCOMMAND_TABLE *command_table;
	WORD i;
	
	if (!(command_table=(VCOMMAND_TABLE *)push_mem(sizeof(VCOMMAND_TABLE)+size*sizeof(COMMAND_DEFINITION))))
		return NULL;
	
	command_table->size=size;
	command_table->running=False;
	
	for (i=0; i<size; ++i)
	{
		command_table->commands[i].definition		=NULL;
		command_table->commands[i].next				=NULL;
	}
	return command_table;
}

static BOOL define_command (VCOMMAND_TABLE *command_table, COMMAND *command)
{
	/* the command should be allocated by the caller... */
	
	WORD n, h;
	
	h = command_hash (command->name);
	n = h = h % command_table->size;
	
	if (command_table->commands[h].definition != NULL)
	{
		/* entry already defined, go to end of list */
		
		while (command_table->commands[h].next)
			h = command_table->commands[h].next;
		
		/* find a free entry */
	
		for (n=1; n<command_table->size && command_table->commands[n].definition != NULL; ++n)
			;
		
		/* check for table overflow */
		
		if (n==command_table->size)
			return False;
		
		/* link in list */
		
		command_table->commands[h].next=n;
	}
	
	command_table->commands[n].definition=command;
	
	return True;
}

BOOL define_commands (VCOMMAND_TABLE *command_table, COMMAND *command, WORD amount)
{
	WORD i;
	
	for (i=0; i<amount; ++i,++command)
		if (!define_command(command_table,command))
			return False;
	
	return True;
}

BOOL exec_command(VCOMMAND_TABLE *command_table,const CHAR *command_line)
{
	CHAR	*line;
	COMMAND	*command;
	WORD	h;
	
	line=(CHAR *)command_line;
	
	if(!scan_word(&line,command_word,StringLength))
		return False;									/* empty string */

	/* there's a non-empty command name */
	
	h = command_hash (command_word) % command_table->size;
	
	while ((command_table->commands[h].definition != NULL) &&
			strcmp (command_word, command_table->commands[h].definition->name))
	{
		if (command_table->commands[h].next)
			h = command_table->commands[h].next;
		else
			return False;								/* command not found */
	}
	
	command=command_table->commands[h].definition;
	
	if (command == NULL)
		return False;									/* command not found */
	
	/* command has been found: parse args (and exec) */
	
	(*(command->parse_function))(&line,command);
	
	return True;
}

/********************************************************************************
 *																				*
 *		standard distributed command execution									*
 *																				*
 ********************************************************************************/

static VOID handle_command(WORD kind,WORD source,WORD size,VCOMMAND_PACKET *packet)
{
	/* default packet handling routine when MOutCommand packet arrives:
	   copy received command into the appropriate command line
	*/
		
	move(packet->command_dest->line,packet->command.characters,packet->command.length);
	packet->command_dest->line[packet->command.length]='\0';
	
	|packet->command_dest->channel=packet->command.length;
}

VOID put_command(WORD dest, WORD command, CHAR *message)
{
	VCOMMAND_PACKET *new_packet;
	WORD length;
	
	/* determine length of message */
	
	length=strlen(message);
	
	/* send command packet to dest */
	
	new_packet=alloc_packet(True);
		
	new_packet->command.length=length;
	move(new_packet->command.characters,message,length);
	
	send_packet(command,unique_rout(dest),
				vsizeof(VCOMMAND_PACKET)+length,new_packet,True);
}

static VOID get_command(CHAR *prompt)
{
	VCOMMAND_PACKET *new_packet;
	WORD length;
	
	/* reset command line */
	
	chreset(&command_line.channel);

	/* determine length of prompt */
	
	length=strlen(prompt);
	
	/* send command packet to host */
	
	new_packet=alloc_packet(True);
	
	new_packet->command_source=&command_line;
	new_packet->command.length=length;	
	move(new_packet->command.characters,prompt,length);
	
	send_packet(MInCommand,unique_rout(ToHost),
				vsizeof(VCOMMAND_PACKET)+length,new_packet,True);
	
	/* wait until answer has been received */
	
	command_line.size=|command_line.channel;
}

static VOID exec_host_command(VOID)
{
	/*	command could not be executed locally, see if host knowns about
		this command
	*/
	
	VCOMMAND_PACKET *new_packet;
	
	/* reset command line */
	
	chreset(&command_line.channel);

	/* send command to host */
	
	new_packet=alloc_packet(True);
	
	new_packet->command_source=&command_line;
	new_packet->command.length=command_line.size;
	move(new_packet->command.characters,command_line.line,command_line.size+1);
	
	send_packet(MHostCommand,unique_rout(ToHost),
				vsizeof(VCOMMAND_PACKET)+command_line.size+1,new_packet,True);

	/* wait until command has been executed */
	
	command_line.size=|command_line.channel;
}

static RECEIVE_ACTION_TABLE command_actions[]=
{
	{MOutCommand,		{handle_command,	handle_command}}
};

VOID execute_commands (VCOMMAND_TABLE *command_table, CHAR *prompt)
{
	/* use default command handling routine (and global command line) */
			
	set_packet_actions(command_actions,sizeof(command_actions)/sizeof(RECEIVE_ACTION_TABLE));

	command_table->running=True;
	
	while(command_table->running)
	{		
		get_command(prompt);
		
		if (!exec_command(command_table,command_line.line))
		{
			/*	this command is locally unknown, so forward it to host, to
				see if the host is able to execute it */
			
			if (command_line.line[0]!='\n')
				file_print(StdErr,"unknown command\n");
			
				/* exec_host_command(); */
		}
	}
}

/********************************************************************************
 *																				*
 *		some predefined parser routines											*
 *																				*
 ********************************************************************************/

VOID parse_e(CHAR **string_handle, COMMAND *command)
{
	(*command->exec_function)();
}

VOID parse_n(CHAR **string_handle, COMMAND *command)
{
	WORD number;
	BOOL avail;
	
	avail=scan_value(string_handle,&number);
	(*command->exec_function)(avail,number);
}

VOID parse_s(CHAR **string_handle, COMMAND *command)
{
	BOOL avail;
	
	avail=scan_string(string_handle,command_word,StringLength);
	(*command->exec_function)(avail,command_word);
}

VOID parse_ident(CHAR **string_handle, COMMAND *command)
{
	BOOL avail;
	
	avail=scan_ident(string_handle,command_word,StringLength);
	(*command->exec_function)(avail,command_word);
}

VOID parse_word(CHAR **string_handle, COMMAND *command)
{
	BOOL avail;
	
	avail=scan_word(string_handle,command_word,StringLength);
	(*command->exec_function)(avail,command_word);
}

VOID parse_nn(CHAR **string_handle, COMMAND *command)
{
	WORD number1;
	WORD number2;
	BOOL avail1;
	BOOL avail2;
	
	avail1=scan_value(string_handle,&number1);
	avail2=scan_value(string_handle,&number2);
	
	(*command->exec_function)(avail1,number1,avail2,number2);
}

/********************************************************************************
 *																				*
 *		some predefined error messages											*
 *																				*
 ********************************************************************************/

VOID missing_string(VOID)
{
	file_print(StdErr,"a string was expected\n");
}

VOID missing_ident(VOID)
{
	file_print(StdErr,"an identifier was expected\n");
}

VOID missing_number(VOID)
{
	file_print(StdErr,"a number was expected\n");
}

VOID missing_address(VOID)
{
	file_print(StdErr,"an address was expected\n");
}
