/************************************************************************
 *																		*
 *			file connect.c												*
 *			author: Marco Kesseler										*
 *																		*
 ************************************************************************/

#include "connect.h"

#ifndef _SYSTEM_
#include "system.h"
#endif

#ifndef _HARDWARE_
#include "hardware.h"
#endif

#include "memory.h"
#include "messageno.h"
#include "errorno.h"
#include "scanner.h"

#ifdef __INCLUDE_STANDARD__

#include <stdlib.h>
#include <stdio.h>

#endif

extern OPTIONS			options;
extern NODE_INFO		node_info;
extern NET_INFO			net_info;
extern PROCESSOR_INFO	processor_info;

#define SYNC_REPEAT		5

/********************************************************************************
 *																				*
 *		auxiliary routines														*
 *																				*
 ********************************************************************************/

static VOID fprintsize(FILE *fp,WORD size)
{
	if (size < KiloByte)
	{
		fprintf(fp,"%d Bytes", size);
		return;
	}
	if (size < MegaByte)
	{
		fprintf(fp,"%g KByte", (FLOAT)size/KiloByte);		
		return;
	}
	fprintf(fp,"%g MByte", (FLOAT)size/MegaByte);
}

VOID show_system_parameters(VOID)
{
	if (node_info.id == RootId)
	{
		if (options.extra_verbose)
		{
			fprintf(stderr,"diam %d, %d(+%d) buffs of ",net_info.diameter,options.n_buffers,net_info.Nclass);
			fprintsize(stderr,options.packet_size);
			fprintf(stderr,"\nmain heap ");
			fprintsize(stderr,options.heap_size);
			fprintf(stderr,"\nlocal heap %d(+%d) (words), stack %d (words)\n",options.init_heap_size,options.increase_heap_size,options.init_stack_size);
			fprintf(stderr,"%d on chip reducers\n",node_info.max_reducers);
			
		} else
		{
			/* echo amount of processors */
			
			if (options.verbose)
			{
				if (net_info.size == 1)
					fprintf(stderr,"1 processor\n");
				else
					fprintf(stderr,"%d processors\n",net_info.size);
			}
		}
	}
}

/********************************************************************************
 *																				*
 *		Part 1																	*
 *		basic connection routines (to determine neighbours, etc.)				*
 *																				*
 ********************************************************************************/

static VOID write_id (WORD link)
{
	BYTE sync;
	
	/* first reset the channel to get it in orderly state */
	
	chreset(link_to_chanout(link));
	
	/* send a number of sync bytes */
	
	for (sync=1; sync<=SYNC_REPEAT; sync++)
		sendlink(link,&sync,sizeof(BYTE));
	
	/* now asume the receiver expects the id */
	
	sendlink(link,&node_info.id,sizeof(WORD));
}

static VOID read_id (WORD link)
{
	BYTE sync;
	BYTE i;
	
	/*	first reset channel to throw away unwanted bytes, which possibly are
		sync bytes, channel is now in orderly state. Normally mostly one sync
		byte is thrown away (as reception of some next byte will only proceed
		after acknowledgement when receiver is ready)
	*/
	
	chreset(link_to_chanin(link));

	for (;;)
	{
		/* try grap sync byte: throw away everything that is strange */
		
		recvlink(link,&sync,sizeof(BYTE));
		while (sync <1 || sync >SYNC_REPEAT)
			recvlink(link,&sync,sizeof(BYTE));
	
		/* some sync byte received, try read rest of sync sequence */
		
		for (i=sync; i<SYNC_REPEAT; ++i)
		{
			recvlink(link,&sync,sizeof(BYTE));
			if (sync != i+1)
			{	sync = 0;
				break;								/* not a byte from the sync seq, try again */
			}
		}
		if (sync == SYNC_REPEAT)
			break;									/* last byte from sync seq read */
	}
	
	/* sync == repeat, so now id should follow */
	
	recvlink(link,&node_info.neighbour[link],sizeof(WORD));
}

static VOID determine_neighbours(VOID)
{
	WORD i;
		
	for (i=0; i<NLinks; ++i)
	{
		if (node_info.neighbour[i] == LUnknown && node_info.nbooted[i] == 0)
		{
			/* it is unknown which transputer is connected to this link and
			   no transputer has been booted over it, so start processes to
			   determine id's of neighbours
			*/
			
			push_process(write_id,	64,1,i);
			push_process(read_id,	64,1,i);
		}
	}
	
	/* make sure every high prio process has started before proceeding */
	
	procresch();
	
	for (i=0; i<NLinks; ++i)
	{
		if (node_info.nbooted[i] > 0)
		{
			/* get id from each booted processor */
			
			recvlink(i,&node_info.neighbour[i],sizeof(WORD));
		}
	}
}

static VOID kill_probes(VOID)
{
	WORD i;
	
	/*	tell others to 'kill' probes by sending a word. This routine is
		used as a synchronisation point: it tells all transputers that all
		have started probes and id's of neighbours should be available by
		now
	*/

	for (i=0; i<NLinks; ++i)
	{
		if (node_info.nbooted[i]>0)
			sendlink(i,&i,sizeof(WORD));
	}
	
	/* high prio avoids problems with shared data here */
	
	for (i=0; i<NLinks; ++i)
	{
		if (node_info.neighbour[i] == LUnknown)
		{
			/*	apparantly this link is not connected and the probes will be
				waiting on it. Kill them by resetting the corresponding link
				hardware... */
			
			chreset(link_to_chanout(i));
			chreset(link_to_chanin(i));
								
			node_info.neighbour[i]=LDumb;
		}
	}
	/*	reset external memory admin, the probe memory is not needed anymore */
	
	reset_external_memory();
}

/************************************************************************************
 *																					*
 *		main connection routine														*
 *																					*
 ************************************************************************************/

VOID connect_neighbours (VOID)
{
	/* start up processes to determine neighbours */
	
	determine_neighbours();

	if (node_info.id == RootId)
	{
		/* all procs have started up probes, tell them to kill after waiting
		   a little to allow them to communicate (waiting is not very important,
		   the communication procs run at high prio and it is unlikely they
		   have not finished already, or by the time the id's of neighbours
		   are needed
		*/
			
		wait(10000); /* 10000 high prio ticks = 0.01 sec */
		
	} else
	{
		WORD dummy;
				
		/* pass own id upwards, and then read sync word and kill probes */
		
		chwrite(node_info.to_host,&node_info.id,sizeof(WORD));
		chread(node_info.from_host,&dummy,sizeof(WORD));
	}
	
	kill_probes();
}

/************************************************************************************
 *																					*
 *		Part 2:																		*
 *		routines to compute routing tables											*
 *																					*
 ************************************************************************************/

static VOID send_rout_table_and_diameter(DESTINATION_INFO **rout_table)
{
	WORD i;

	for (i=0;i<NLinks;i++)
	{
		WORD j;	/* separate j for each process */
		
		if (node_info.nbooted[i] > 0)
		{
			for (j=0 ; j<net_info.size+2 ; j++)
				sendlink(i,rout_table[j],(net_info.size+2)*sizeof(DESTINATION_INFO));
			
			sendlink(i,&net_info.diameter,sizeof(WORD));
		}
	}
}

static VOID pass_rout_table_and_diameter(DESTINATION_INFO **rout_table)
{
	WORD i;
	
	for (i=0;i<net_info.size+2;i++)
	{
		WORD l;
		
		chread(node_info.from_host,rout_table[i],(net_info.size+2)*sizeof(DESTINATION_INFO));
		
		/*	pass information on directly, to reduce delay! For large networks
			this is quite important... */

		for (l=0; l<NLinks; ++l)
			if (node_info.nbooted[l] > 0)
					sendlink(l,rout_table[i],(net_info.size+2)*sizeof(DESTINATION_INFO));
	}
	
	chread(node_info.from_host,&net_info.diameter,sizeof(WORD));
	
	for (i=0; i<NLinks; ++i)
		if (node_info.nbooted[i] > 0)
			sendlink(i,&net_info.diameter,sizeof(WORD));
}

static BOOL check_neighbours(WORD **neighbour)
{
	WORD transputer;
	WORD l;
	WORD error;
	
	if (!options.check)
		return True;
	
	error=False;
	
	for (transputer=0; transputer<=net_info.size; ++transputer)
	{
		for (l=0; l<NLinks; ++l)
		{
			WORD nbour;
			WORD nbourcount1;
			WORD nbourcount2;
			WORD link2;
			
			nbour=neighbour[transputer][l];
			if (nbour<0)
				continue;							/* link not connected */

			if (nbour>net_info.size)
			{
				fprintf(stderr,"transputer %d has non-existing neighbour %d\n",transputer,nbour);
				error=True;
				continue;
			}
			
			nbourcount1=0;
			nbourcount2=0;

			for (link2=0; link2<NLinks; ++link2)
			{	if (neighbour[transputer][link2]==nbour)
					++nbourcount1;
			}
			
			for (link2=0; link2<NLinks; ++link2)
			{	if (neighbour[nbour][link2]==transputer)
					++nbourcount2;
			}
			
			if (nbourcount1!=nbourcount2)
			{
				fprintf(stderr,"proc %d found %d links to proc %d\n",transputer,nbourcount1,nbour);
				fprintf(stderr,"proc %d found %d links to proc %d\n",nbour,nbourcount2,transputer);
				error=True;
			}
		}
	}
	return !error;
}

static VOID init_rout_table(DESTINATION_INFO **rout_table)
{
	WORD s, d;
		
	for (s=0; s<net_info.size+2; s++)
	{	for (d=0; d<net_info.size+2; d++)
		{	rout_table[s][d].proc_info				=net_info.size+1;		/* shortest path from s to d not yet found	*/
			rout_table[s][d].distance_info			=net_info.size+1;		/* shortest path of length d not yet found	*/
			rout_table[s][d].outlinks				=0;						/* no link to dest yet */
			rout_table[s][d].outblinks				=0;						/* no links to all yet */
			rout_table[s][d].distance				=INT_MAX;				/* infinity */
			rout_table[s][d].proc					=net_info.size+1;
		}
	}
}

static WORD compute_rout_table(WORD **neighbour, DESTINATION_INFO **rout_table)
{	
	/* This routine computes all shortests paths from every source to every
	   destination. This algorithm is O(n*n), because the maximum number of links
	   from a certain transputer is fixed. It returns the longest path that
	   was found (diameter of the network). For each transputer an ordered
	   array is returned that contains routing information. This information
	   is ordered on pathlength. It holds for all transputers including the
	   host.
	*/
	
	WORD s;						/* source */
	WORD length;				/* pathlength			*/
	WORD l;						/* to neighbour			*/
	WORD i;
	WORD new_path_found;
	
	/* compute paths of length 0 from s */
	
	for (s=0; s<=net_info.size; s++)
	{
		rout_table[s][s].proc_info				=0;		/* information about dest s at 0		*/
		rout_table[s][0].distance_info			=0;		/* information about distance 0 at 0	*/
		rout_table[s][1].distance_info			=1;		/* information about distance 1 at 1	*/
		
		/* fill in record 0 */
		
		rout_table[s][0].proc			=s;				/* destination is proc itself	*/
		rout_table[s][0].distance		=0;				/* distance is 0				*/
		rout_table[s][0].outlinks		=(1<<NLinks);	/* outlink is 4 (internal)		*/
	}
	
	/* paths of length 0 have been computed, now compute shortest paths of length
	   1..Ntransputers. This is done by looking at the neighbours of each source.
	   A path of length L is built by taking a path of length L-1 from a neighbour.
	   The maximum length of a path is net_info.size (host + net_info.size transputers).
	*/
	
	for (length=1; length<=net_info.size; length++)
	{
		new_path_found=False;
		
		for(s=0; s<=net_info.size; s++)
		{
			/* try find new path of length 'length' from source 's' */
			
			WORD new_path;
			
			/* determine index where info of paths of length 'length' has to
			   be stored. */
			
			new_path=rout_table[s][length].distance_info;
			
			for(l=0; l<NLinks; l++)
			{
				/* try find new path of length 'length' from source 's' via neighbour 'l' */
				
				WORD nbour;
				
				/* find all paths of length-1 from neighbour */

				nbour=neighbour[s][l];
								
				if (nbour<0)
					continue;	/* no neighbour 'l' exists */
								
				for(i=rout_table[nbour][length-1].distance_info; i<rout_table[nbour][length].distance_info; i++)
				{
					WORD new_dest, old_dest_info;
					
					new_dest=rout_table[nbour][i].proc;
					old_dest_info=rout_table[s][new_dest].proc_info;
					
					/* found a path of length 'length-1' from 'l' to 'new_dest',
					   see if already a path to 'new_dest' had been found earlier */
					
					if(old_dest_info>net_info.size)
					{
						/* no path was found from source to destination earlier:
						   path has 'infinite' length. So new path found from s
						   to 'new_dest' via 'l' */
						
						rout_table[s][new_dest].proc_info=new_path;
						rout_table[s][new_path].proc=new_dest;
						rout_table[s][new_path].distance=length;
						rout_table[s][new_path].outlinks=(1<<l);
						new_path++;
						
						new_path_found=True;
					} else
					{
						/* a path from source to dest already has been found earlier */
						
						if(rout_table[s][old_dest_info].distance==length)
						{
							/* another shortest path found from 's', so add link */
							
							rout_table[s][old_dest_info].outlinks|=(1<<l);
						}
						/* else forget this path: a shorter one already has been found */
					}
				}							
			}
			/* now new paths of length 'length' have been found via the neighbours
			   of source. 'new_path' contains the first index where information
			   about length+1 paths should be stored later. */
			
			rout_table[s][length+1].distance_info=new_path;
			
			/* proceed with next s */
		}
		
		/* proceed with 'length+1', unless no new path of length 'length' has been found */
		
		if (!new_path_found)
			return length-1;
	}
	
	/* maximum length reached ('length' is max length+1) */
	
	return length-1;
}

static VOID compute_broad_table(WORD **neighbour, DESTINATION_INFO **rout_table, WORD *visited)
{
	/* after computing point-to-point routing tables with the routine above,
	   this routine computes for each source the links to use for broadcasting.
	   This is done in a straightforward way for each source separately: the
	   source should use all its links and mark its neighbours. Next the neighbours
	   must use all links to unmarked neighbours.
	*/
	
	WORD source;
	
	for (source=0; source<=net_info.size; source++)
	{
		WORD dest;
		DESTINATION_INFO *dest_info;

		/* no nodes visited from source yet */
		
		for (dest=0; dest<=net_info.size; dest++)
			visited[dest]=False;
		
		/* visit nodes from source in increasing distance from source */

		for(dest_info=rout_table[source]; dest_info<=&rout_table[source][net_info.size]; dest_info++)
		{
			WORD link;
			DESTINATION_INFO *source_info;
		
			dest=dest_info->proc;
			source_info=&rout_table[dest][rout_table[dest][source].proc_info];
			
			/* broadcast over internal link... */

			source_info->outblinks|=(1<<NLinks);
			visited[dest]=True;

			/* find neighbours of dest */
			
			for (link=0; link<NLinks; link++)
			{
				WORD neighb;
				
				neighb=neighbour[dest][link];
				
				if (neighb >= 0)
				{
					/* a neighbour exists, see if it has been reached earlier */
					
					if (visited[neighb])
						continue;
				
					/* the neighbour has no yet been reached, so dest may be used
					   to broadcast to it
					*/
				
					source_info->outblinks|=(1<<link);
					visited[neighb]=True;											
				}
			}
		}
	}	
}

static VOID relocate_host_info(DESTINATION_INFO *dest_info)
{
	DESTINATION_INFO host_info;
	WORD host_index,i,distance;
	
	/*	relocate host info to the end of dest_info array. First save host
		info. Then shift all info down that is above it, i.e. of processors
		at greater or equal distance as the host. Finally, set the host info
		at the end.
	*/
	
	host_index=dest_info[ToHost].proc_info;
	host_info=dest_info[host_index];
	
	/* shift all info blocks above host down */
	
	for (i=host_index, distance=host_info.distance; i<net_info.size; ++i)
	{
		WORD proc;
		
		dest_info[i]=dest_info[i+1];
		
		/*	proc, distance, and outlinks remain the same of course, but links
			to this record (proc_info and distance_info) must be adjusted. The
			proc info must be set to exactly this record, the distance info must
			be set to this record or before it */
		
		proc		=dest_info[i].proc;
		distance	=dest_info[i].distance;
		
		dest_info[proc].proc_info=i;
		
		if (dest_info[distance].distance_info > i)
			dest_info[distance].distance_info = i;
	}
	
	/* adjust distance info for values greater than distances already adjusted */
	
	for (++distance; distance<=net_info.size; ++distance)
		dest_info[distance].distance_info=net_info.size;
	
	/* and put host info at the end */
	
	dest_info[net_info.size]=host_info;
	dest_info[ToHost].proc_info=net_info.size;
}

typedef struct NEIGHBOUR_INFO
{
	WORD	id;
	WORD	memsize;
	WORD	neighbours[NLinks];
	
} NEIGHBOUR_INFO;

BOOL compute_route_tables(VOID)
{
	WORD i;
	DESTINATION_INFO **rout_table;

	/* Destination info will be computed for n transputers + root; 1 extra field
	   is used to indicate an infinitly long path. First allocate resident memory,
	   note that the total routeing information is distributed over all processors
	   and each picks what it needs.
	*/
	
	if(!(node_info.rout_data=push_mem((net_info.size+2)*sizeof(DESTINATION_INFO))))
		return False;

	if(!(rout_table=push_mem((net_info.size+2)*sizeof(DESTINATION_INFO *))))
		return False;
		
	for (i=0;i<net_info.size+2;i++)
	{	if (i != node_info.id)
		{	if(!(rout_table[i]=push_mem((net_info.size+2)*sizeof(DESTINATION_INFO))))
				return False;
		}
		else
			rout_table[i]=node_info.rout_data;
	}
	
	if (node_info.id == RootId)
	{
		BOOL *visited;
		WORD *memsizes;
		WORD **neighbour;
		
		/* neighbours of n transputers + host will be determined, reserve memory
		   to store neighbour info */
	
		if(!(neighbour=push_mem((net_info.size+1)*sizeof(WORD *))))
			return False;
	
		for (i=0;i<net_info.size+1;i++)
		{	if(!(neighbour[i]=push_mem(NLinks*sizeof(WORD))))
				return False;
		}
		
		if(!(visited=push_mem((net_info.size+1)*sizeof(BOOL))))
			return False;
		if(!(memsizes=push_mem((net_info.size+1)*sizeof(WORD))))
			return False;

		/* receive info about neighbours...
		   the root is the neighbour of the host (link unimportant) */
	
		memsizes[ToHost]=0;
		neighbour[ToHost][0]=RootId;
		for (i=1; i<NLinks; ++i)
			neighbour[ToHost][i]=LDumb;
	
		/* gather global information about links */

		for (i=0; i<NLinks ;i++)
		{
			/* note each process has a copy of the vars defined below */
			
			NEIGHBOUR_INFO remote;
			WORD j,k,Tn;

			/* set neighbours of root */
			
			neighbour[RootId][i]	=node_info.neighbour[i];
			memsizes[RootId]		=node_info.ext_mem_size;
	
			/* get info of procs booted over this link */
			
			for (j=0;j<node_info.nbooted[i];j++)
			{
				recvlink(i,&remote,sizeof(NEIGHBOUR_INFO));
				
				Tn				=remote.id;
				memsizes[Tn]	=remote.memsize;
				
				for (k=0;k<NLinks;k++)
					neighbour[Tn][k]=remote.neighbours[k];
			}
		}
		
		if (options.extra_verbose)
			fprintf(stderr,"computing routing tables\n");
		
		init_rout_table(rout_table);
		
		if (!check_neighbours(neighbour))
			net_info.diameter=-1;
		else
		{
			net_info.diameter=compute_rout_table(neighbour,rout_table);
			compute_broad_table(neighbour,rout_table,visited);
		}

		if (options.extra_verbose)
		{
			/* print connection information */
			
			for (i=0; i<=net_info.size; i++)
			{
				WORD j;	

				fprintf(stderr,"proc %6d [",i);
				
				for (j=0; j<NLinks; j++)
				{
					WORD n;
					
					n=neighbour[i][j];
					
					if (n >= 0)
						fprintf(stderr,"%6d",n);
					else
						fprintf(stderr,"      ");
					
					if (j<NLinks-1)
						putc(',',stderr);					
				}
				fprintf(stderr,"], ");
				fprintsize(stderr,memsizes[i]);
				putc('\n',stderr);
			}
		}

		/* visited array and memsizes not needed anymore */
		
		pop_mem((net_info.size+1)*sizeof(BOOL));
		pop_mem((net_info.size+1)*sizeof(WORD));

		/* neighbour info is not needed anymore, so discard it */
		
		for (i=0;i<net_info.size+1;i++)
			pop_mem(NLinks*sizeof(WORD));
		pop_mem((net_info.size+1)*sizeof(WORD *));
		
		/* send info down */
		
		send_rout_table_and_diameter(rout_table);
	
	} else
	{
		/* this is not the root, send info about neighbours to root processors,
		   note that the parallel booter must use a semaphore to assure safe
		   access to the (shared) host link */

		NEIGHBOUR_INFO local;

		for (i=0;i<NLinks;i++)
		{
			NEIGHBOUR_INFO remote;
			WORD j;

			for (j=0;j<node_info.nbooted[i];j++)
			{
				recvlink(i,&remote,sizeof(NEIGHBOUR_INFO));
				chwrite(node_info.to_host,&remote,sizeof(NEIGHBOUR_INFO));
			}
		}

		local.id		=node_info.id;
		local.memsize	=node_info.ext_mem_size;
		
		for (i=0 ; i<NLinks ; i++)
			local.neighbours[i]=node_info.neighbour[i];
		
		chwrite(node_info.to_host,&local,sizeof(NEIGHBOUR_INFO));

		/* now wait for routing info and diameter to arrive, and pass these */
		
		pass_rout_table_and_diameter(rout_table);
	}
		
	if (net_info.diameter < 0)
		return False;
	
	for (i=0;i<net_info.size+2;i++)
	{	if (i != node_info.id)
			pop_mem((net_info.size+2)*sizeof(DESTINATION_INFO));
	}
	pop_mem((net_info.size+2)*sizeof(DESTINATION_INFO *));
	
	/*	the host processor plays an important role in the system: it should
		not be electable to start up processes, but it might come in handy
		to have some information about it at hand. For this reason the destination
		info of the host will be moved to the end of the row.
	*/
	
	relocate_host_info(node_info.rout_data);
	
	return True;
}

/************************************************************************************
 *																					*
 *		routine that passes error numbers upward after first initialisation			*
 *																					*
 ************************************************************************************/

VOID error_number (WORD errno)
{
	WORD i;
	WORD par_errno[NLinks];
		
	/*
		if no local error, gather and pass errno of booted proces, else
		pass local error.
	*/

	if (errno == ENoErr)
	{
		for (i=0; i<NLinks; ++i)
		{	if (node_info.nbooted[i] > 0)
				recvlink(i,&par_errno[i],sizeof(WORD));
			else
				par_errno[i]=ENoErr;
		}
		
		/* set errno to first error encountered of booted procs */
		
		for (i=0; i<NLinks; ++i)
		{
			if (par_errno[i]!=ENoErr)
			{
				errno=par_errno[i];
				break;
			}
		}
	}

	processor_info.error_number=errno;

	/* pass error number upward if not root proc, else tell about error */
	
	if (node_info.id != RootId)
	{
		chwrite(node_info.to_host,&errno,sizeof(WORD));
		
		if (errno != ENoErr)
			procstop();						/* an error occured, so stop */
	} else
	{
		if (errno != ENoErr)
		{	fprintf(stderr,"\n\ninitialisation error, id = %d\n",errno);
			exit(errno);
		}
	}
}

