#include "debugger.h"

#include "macros.h"
#include "commands.h"
#include "io.h"
#include "system.h"
#include "packets.h"
#include "instructions.h"
#include "nodes.h"
#include "hardware.h"
#include "processes.h"
#include "kernel.h"

#ifdef __INCLUDE_STANDARD__

#include <stdlib.h>
#include <stdio.h>

#endif

/********************************************************************************
 *																				*
 *		global variables														*
 *																				*
 ********************************************************************************/

static VCOMMAND_TABLE *debug_command_table;

static BOOL		debug_verbose;

/********************************************************************************
 *																				*
 *		import system vars														*
 *																				*
 ********************************************************************************/

extern NODE_INFO	node_info;

extern SYS 			sys;

/********************************************************************************
 *																				*
 *		debug commands															*
 *																				*
 ********************************************************************************/

static VOID quit(VOID)
{
	debug_command_table->running=False;
	hexit(0);
}

static VOID run(VOID)
{
	debug_command_table->running=False;
}

static VOID jump_absolute_code(BOOL avail, ADDRESS address)
{
	if (!avail)
	{
		missing_address();
		return;
	}
	
	sys.processor_state.Iptr=address;
}	

static VOID jump_code(BOOL avail, WORD offset)
{
	if (!avail)
	{
		missing_number();
		return;
	}

	sys.processor_state.Iptr+=offset;
}

static VOID step_instruction(BOOL avail, WORD n)
{	
	INSTRUCTION instruction;
	
	if(!avail)
		n=1;

	if (debug_verbose)
	{
		if (n > 50)
			file_print(StdErr,"step verbose %d (ignored)\n",n);
		else
		{
			for(; n>0; --n)
			{
				if(!step(&instruction))
				{	file_print(StdErr,"can't step\n");
					return;
				}
				show_basic_instruction(StdErr,&instruction);
			}
		}
	} else
	{
		WORD i=0;
		WORD amount;
				
		for(amount=0;amount<n;++amount)
		{
			if (++i > 10000)
			{
				i=1;
				
				/* procresch();		so that other high prio procs can run... */
				
				file_print(StdErr,"\r%d",amount);
			}
			
			if(!step(&instruction))
			{	file_print(StdErr,"\r%d: can't step\n",amount);
				return;
			}
		}
	}
}

static VOID simulate_instructions(VOID)
{	
	INSTRUCTION instruction;
			
	while(step(&instruction))
	{	if(debug_verbose)
			show_basic_instruction(StdErr,&instruction);
	}
	file_print(StdErr,"can't step\n");
}

static VOID unlock_heap(VOID)
{
	unblock_heap_messages();
}

static VOID prio(VOID)
{
	if(sys.processor_state.prio)
		file_print(StdErr,"low priority\n");
	else
		file_print(StdErr,"high priority\n");
}

static VOID heap(VOID)
{

	ADDRESS curr, garb, end, free;
	WORD size, free_nr;
		
	curr = sys_root->current_heap;
	garb = sys_root->garbage_heap;
	end  = sys_root->heap_end_ptr;
	free = sys_root->heap_free_ptr;
	size = sys_root->half_heap_size;
	free_nr=(WORD)end - (WORD)free;
	
	file_print(StdErr,"start of current heap is at  : %X\n", curr);
	file_print(StdErr,"start of garbage heap is at  : %X\n", garb);
	file_print(StdErr,"pointer to first free byte   : %X\n", free);
	file_print(StdErr,"pointer to end of heap       : %X\n", end);
	
	file_print(StdErr,"\ncurrent heap half:\n");	
	file_print(StdErr,"size: %d used: %d free: %d (bytes)\n", size, size-free_nr, free_nr);
}

static WSP *get_param_wsp(BOOL available, WORD reducer)
{
	if(!available)
		return (WSP *)sys.processor_state.Wptr;
	
	/*	if positive, the argument is supposed to be a pid, else it
		is a workspace pointer. If a pid is used, check if the
		table contains a forwarding address.
	*/
	
	if (reducer >= 0)
	{
		REDUCER_ENVIRONMENT *env;

		if (sys.comm_info.id_table[reducer].process.refnum != RefProcess)
		{
			file_print(StdErr,"%d is not a pid\n",reducer);
			return NULL;
		}
		
		env=pid_to_env(reducer);

		if (!test_word_address(env))
			return NULL;

		while (is_forward(((NODE *)env)))
		{
			env=(REDUCER_ENVIRONMENT *)get_forward(((NODE *)env));
		
			if (!test_word_address(env))
				return NULL;
		}
		
		/*	obtained a valid environment, get workspace */
		
		return env_to_regs(env);
	}
	
	if (test_word_address((ADDRESS)reducer))
		return (WSP *)reducer;
		
	return NULL;
}

static VOID locals(BOOL available, WORD reducer)
{
	WSP *wsp;
	
	wsp=get_param_wsp(available, reducer);
	
	if (wsp == NULL)
		return;
		
	if (wsp != (WSP *)sys.processor_state.Wptr)
	{
		/* this is not the current process, print the stored iptr */
		
		file_print(StdErr,"iptr: %08X\n\n",get_iptr(wsp));
	}
	
	file_print(StdErr,"return address: %08X    ",	wsp->return_address);
	file_print(StdErr,"a2            : %08X\n",		wsp->a2);
	
	file_print(StdErr,"b1            : %08X    ",	wsp->b1);
	file_print(StdErr,"a3            : %08X\n",		wsp->a3);

	file_print(StdErr,"b2            : %08X    ",	wsp->b2);
	file_print(StdErr,"a4            : %08X\n",		wsp->a4);

	file_print(StdErr,"b3            : %08X    ",	wsp->b3);	
	file_print(StdErr,"extra local   : %08X\n",		wsp->extra_local);

	file_print(StdErr,"b4            : %08X    ",	wsp->b4);
	file_print(StdErr,"b-stack ptr   : %08X\n", 	wsp->b_stack_ptr);

	file_print(StdErr,"b5            : %08X    ",	wsp->b5);
	file_print(StdErr,"a-stack ptr   : %08X\n",		wsp->a_stack_ptr);

	file_print(StdErr,"b6            : %08X    ",	wsp->b6);
	file_print(StdErr,"heap start    : %08X\n",		wsp->heap_start);
	
	file_print(StdErr,"a1            : %08X    ",	wsp->a1);
	file_print(StdErr,"heap end      : %08X\n",		wsp->heap_end);
}

static VOID stacks(BOOL available, WORD reducer)
{
	WORD *a_top, *a_start;
	WORD *b_top, *b_start;
	WSP * wsp;
	REDUCER_ENVIRONMENT *env;

	wsp=get_param_wsp(available,reducer);
	if (wsp == NULL)
		return;
	
	env=get_process_env(wsp);
	
	a_top = (WORD *)wsp->a_stack_ptr;
	b_top = (WORD *)wsp->b_stack_ptr;	
	
	a_start = (WORD *)env->a_stack_start;
	b_start = (WORD *)env->b_stack_start;

	file_print(StdErr,"stack:     size:      free:      top:       start:     end\n");
	file_print(StdErr,"AS         %-10d %-10d %08X  %08X  %08X\n", a_top-a_start, b_top-a_top, a_top, a_start, b_top);
	file_print(StdErr,"BS         %-10d %-10d %08X  %08X  %08X\n", b_start-b_top, b_top-a_top, b_top, b_start, a_top);
}

static VOID anodes(BOOL avail, WORD reducer)
{
	NODE **a_top,**a_start;
	WSP *wsp;
	REDUCER_ENVIRONMENT *env;
	
	wsp=get_param_wsp(avail,reducer);
	if (wsp == NULL)
		return;
	
	env=get_process_env(wsp);

	a_top	= wsp->a_stack_ptr;
	a_start	= env->a_stack_start;

	if (test_word_address((ADDRESS)a_top) && test_word_address((ADDRESS)a_start))
	{	
		while (a_top > a_start)
		{
			--a_top;
			
			if (test_word_address((ADDRESS)a_top))
			{
				file_print(StdErr,"%08X: ",*a_top);

				if (test_word_address((ADDRESS)*a_top))
					show_node(StdErr,*a_top);
			}
		}
	}
}

static VOID regs(VOID)
{
	file_print(StdErr,"Areg         : %08X    ", sys.processor_state.Areg);
	file_print(StdErr,"fAreg        : %0.16e\n", sys.processor_state.fAreg);
	file_print(StdErr,"Breg         : %08X    ", sys.processor_state.Breg);
	file_print(StdErr,"fBreg        : %0.16e\n", sys.processor_state.fBreg);
	file_print(StdErr,"Creg         : %08X    ", sys.processor_state.Creg);
	file_print(StdErr,"fCreg        : %0.16e\n", sys.processor_state.fCreg);
	
	if(sys.processor_state.error_flag)
		file_print(StdErr,"error flag   : clear       ");
	else
		file_print(StdErr,"error flag   : set         ");
	
	if(sys.processor_state.float_error_flag)
		file_print(StdErr,"f error flag : clear\n\n");
	else
		file_print(StdErr,"f error flag : set\n\n");
	
	file_print(StdErr,"Wdesc        : %08X    ", (WORD)sys.processor_state.Wptr | sys.processor_state.prio);
	file_print(StdErr,"low_front    : %08X\n", sys.processor_state.low_front);
	
	file_print(StdErr,"Iptr         : %08X    ", sys.processor_state.Iptr);
	file_print(StdErr,"low_back     : %08X\n", sys.processor_state.low_back);
	
	file_print(StdErr,"Time         : %08X\n",sys.processor_state.time);
}

static VOID instructions(BOOL a1, ADDRESS from, BOOL a2, WORD amount)
{
	if (a1)
	{
		if (!a2)
			amount=1;
		
		show_code(StdErr,0,from,amount);
		
	} else
		missing_address();
	
}

static VOID relative_instructions(BOOL avail1, WORD from, BOOL avail2, WORD to)
{
	ADDRESS start_address;
	WORD amount;
	
	start_address=get_instr_address(sys.processor_state.Iptr);
	
	if(avail1)
	{	if(avail2)
		{
			amount=to-from+1;
			
			if(amount<=0)
			{	file_print(StdErr,"invalid interval\n");
				return;
			}
			
			start_address=get_offset_instr_address(start_address,from);
		
		} else
		{
			/* only 1 param available, if negative it is 'from' else 'to */
			
			if(from<0)
			{	start_address=get_offset_instr_address(start_address,from);
				amount=-from+1;
			} else
				amount=from;
		}
	} else
	{
		/* no params available */
		
		amount=1;
	}
	show_code(StdErr,sys.processor_state.Iptr,start_address,amount);
}

static VOID instruction_page(VOID)
{
	ADDRESS start_address;

	start_address=get_offset_instr_address(sys.processor_state.Iptr,-5);
	show_code(StdErr,sys.processor_state.Iptr,start_address,20);
}
	
static VOID node(BOOL avail, NODE *node)
{
	if (!avail)
	{
		missing_address();
		return;
	}
	show_node(StdErr,node);
}

static VOID anode(BOOL avail, WORD offset)
{
	NODE **a_top;
	WSP *wsp;
	
	if (!avail)
	{
		missing_number();
		return;
	}
	
	wsp=(WSP *)sys.processor_state.Wptr;

	a_top = wsp->a_stack_ptr;
	
	if (test_word_address((ADDRESS)a_top))
	{
		file_print(StdErr,"AS[%d]: ",offset);
		show_node(StdErr,a_top[-offset-1]);
	}
}

static VOID lnode(BOOL avail, WORD offset)
{
	NODE **local;
	
	local = (NODE **)sys.processor_state.Wptr;
	file_print(StdErr,"AL[%d]: ",offset);
	show_node(StdErr,local[offset]);
}
	
static VOID verbose(VOID)
{
	debug_verbose=True;
}

static VOID brief(VOID)
{
	debug_verbose=False;
}

static VOID low_mem(BOOL available, WORD transputer)
{
	if(!available)
		transputer=node_info.id;
	
	file_print(StdErr,"link 0 output         : %08X    ",sys.processor_state.link0_out);
	file_print(StdErr,"high prio timer queue : %08X\n",sys.processor_state.high_timer_list);

	file_print(StdErr,"link 1 output         : %08X    ",sys.processor_state.link1_out);
	file_print(StdErr,"low prio timer queue  : %08X\n",sys.processor_state.low_timer_list);

	file_print(StdErr,"link 2 output         : %08X    ",sys.processor_state.link2_out);
	file_print(StdErr,"Wdesc save            : %08X\n",sys.processor_state.WdescIntSaveLoc);
	
	file_print(StdErr,"link 3 output         : %08X    ",sys.processor_state.link3_out);
	file_print(StdErr,"Iptr save             : %08X\n",sys.processor_state.IptrIntSaveLoc);
	
	file_print(StdErr,"link 0 input          : %08X    ",sys.processor_state.link0_in);
	file_print(StdErr,"Areg save             : %08X\n",sys.processor_state.AregIntSaveLoc);
	
	file_print(StdErr,"link 1 input          : %08X    ",sys.processor_state.link1_in);
	file_print(StdErr,"Breg save             : %08X\n",sys.processor_state.BregIntSaveLoc);
	
	file_print(StdErr,"link 2 input          : %08X    ",sys.processor_state.link2_in);
	file_print(StdErr,"Creg save             : %08X\n",sys.processor_state.CregIntSaveLoc);
	
	file_print(StdErr,"link 3 input          : %08X    ",sys.processor_state.link3_in);
	file_print(StdErr,"STATUS save           : %08X\n",sys.processor_state.STATUSIntSaveLoc);
	
	file_print(StdErr,"Event                 : %08X    ",sys.processor_state.event_in);
	file_print(StdErr,"Ereg save             : %08X\n",sys.processor_state.EregIntSaveLoc);
}

static VOID lowps (VOID)
{
	WORD *wsp;
	
	file_print(StdErr,"interrupted proc: %08X\n",(WSP *)sys.processor_state.Wptr);
	
	wsp=(WORD *)sys.processor_state.low_front;
	
	if (wsp==NotProcessP)
		file_print(StdErr,"empty low prio scheduling list\n");
	else
	{
		file_print(StdErr,"scheduling list:\n");
		while (wsp != (WORD *)sys.processor_state.low_back)
		{
			file_print (StdErr,"%08X\n",(int)wsp);
			wsp=get_next_process(wsp);
		}
		file_print (StdErr,"%08X\n",wsp);
	}
}

static VOID show_process(WSP *wsp,ADDRESS iptr)
{
	REDUCER_ENVIRONMENT *env;
	BOOL reducer;
	
	wsp=(WSP *)word_truncate((WORD)wsp);
	
	if (wsp==NULL || wsp==(WSP *)NotProcessP)
		return;

	if (!test_word_address((ADDRESS)wsp))
		return;
	
	if (wsp >= sys.reducer_info.on_chip_first_regs &&
		wsp <= sys.reducer_info.on_chip_last_regs)
	{
		/* on-chip reducer */
		
		file_print(StdErr,"on-chip ");
		reducer=True;
	
	} else if ((ADDRESS)wsp >= sys_root->current_heap &&
		(ADDRESS)wsp < sys_root->current_heap + sys_root->half_heap_size)
	{
		/* off-chip reducer */
		
		file_print(StdErr,"off-chip");
		reducer=True;
		
	} else
		reducer=False;

	if (reducer)
	{
		env=get_process_env(wsp);
		file_print(StdErr," wsp: %08X, iptr: %08X, pid: %d, status: ",wsp,iptr,get_pid(env));
		if (is_active(env))
			file_print(StdErr,"active");
		else
			file_print(StdErr,"inactive");
	} else
	{	
		/*	wsp not inside current heap, nor on-chip, this is a system
			process
		*/
		
		file_print(StdErr,"system   wsp: %08X, iptr: %08X, system process",wsp,iptr);
	}
	
	if (wsp == (WSP *)word_truncate((WORD)sys.processor_state.WdescIntSaveLoc))
		file_print(StdErr," interrupted\n");
	else
		file_print(StdErr,"\n");
}

static VOID show_processes (VOID)
{
	WSP *wsp;
	
	show_process((WSP *)sys.processor_state.Wptr,sys.processor_state.Iptr);
	show_process((WSP *)sys.processor_state.WdescIntSaveLoc,sys.processor_state.IptrIntSaveLoc);
	
	wsp=(WSP *)sys.processor_state.low_front;
	if (wsp != (WSP *)NotProcessP)
	{
		while (wsp != (WSP *)sys.processor_state.low_back)
		{
			show_process(wsp,get_iptr(wsp));
			wsp=get_next_process(wsp);	
		}
		show_process(wsp,get_iptr(wsp));
	}
		
	for (wsp=sys.reducer_info.on_chip_first_regs;wsp<=sys.reducer_info.on_chip_last_regs;wsp=(WSP *)((REDUCER_LOCALS *)wsp+1))
	{
		/*	if the reducer has an environment and it is inactive, it is a
			suspended reducer (stopped ones have the 'active' bit set!) */

		if (wsp->heap_start != NULL && is_inactive(get_process_env(wsp)))
			show_process(wsp,get_iptr(wsp));
	}
		
	wsp=sys.reducer_info.off_chip_first_regs;
	if (wsp != NULL)
	{
		while (wsp != sys.reducer_info.off_chip_last_regs)
		{
			show_process(wsp,get_iptr(wsp));
			wsp=get_next_process(wsp);	
		}
		show_process(wsp,get_iptr(wsp));
	}
	
	wsp=sys.reducer_info.off_chip_suspended_regs;	
	while (wsp != NULL)
	{
		show_process(wsp,get_iptr(wsp));
		wsp=get_next_process(wsp);
	}
}

static VOID tablegraphs(VOID)
{
	WORD i;

	for (i=0; i<sys.comm_info.id_table_size; ++i)
	{
		ENTRY *entry;

		entry=&sys.comm_info.id_table[i];

		if (entry->graph.refnum > RefFree &&
			entry->graph.refnum <= RefMax &&
			entry->graph.graph != NULL)
		{
			file_print(StdErr,"table[%d]:\n",i);
			show_node(StdErr,entry->graph.graph);
		}
	}
}

static VOID table(BOOL avail, WORD index)
{
	
	if (avail && index >= 0 && index < sys.comm_info.id_table_size)
	{
		ENTRY *entry;

		entry=&sys.comm_info.id_table[index];
		
		file_print(StdErr,"Table[%d]: ",index);
		
		switch (entry->graph.refnum)
		{
			case RefFree:
				file_print(StdErr,"free entry, link to %d",entry->free.next);
				break;
				
			case RefNull:
				file_print(StdErr,"null entry");
				break;

			case RefProcess:
				file_print(StdErr,"process entry, environment %08X",entry->process.env);
				break;

			case RefAllocFail:
				file_print(StdErr,"fail entry, discarded %d bytes",entry->alloc_fail.discarded);
				break;

			case RefIdBuff:
				file_print(StdErr,"id buffer entry, buff1 %08X, buff2 %08X, wait %08X",entry->idbuff.buff1,entry->idbuff.buff2,entry->idbuff.wait);
				break;

			default:
			{
				file_print(StdErr,"refnum %d, ",entry->graph.refnum);
				
				if (entry->graph.refnum > RefFree && entry->graph.refnum <= RefMax)
					file_print(StdErr,"(graph %08X requests %08X pid %08X)",entry->graph.graph,entry->graph.requests,entry->graph.pid);
				else
					file_print(StdErr,"invalid refnum");
			}
		}
		file_print(StdErr,"\n");
	
	} else
	{
		WORD i, nfree, nbuff, nalloc_fail, nproc, nnull, ngraph;

		file_print(StdErr,"table size: %d\n",sys.comm_info.id_table_size);
		
		nfree		=0;
		nbuff		=0;
		nalloc_fail	=0;
		nproc		=0;
		nnull		=0;
		ngraph		=0;
		
		for (i=0; i<sys.comm_info.id_table_size; ++i)
		{
			switch (sys.comm_info.id_table[i].graph.refnum)
			{
				case RefFree:		++nfree;		break;
				case RefNull:		++nnull;		break;
				case RefProcess:	++nproc;		break;
				case RefAllocFail:	++nalloc_fail;	break;
				case RefIdBuff:		++nbuff;		break;
				default:			++ngraph;
			}
		}
		file_print(StdErr,"%d free, %d graphs, %d id buffers, %d processes\n",nfree,ngraph,nbuff,nproc);
	}
}

static VOID cstring (BOOL avail, ADDRESS address)
{
	file_print(StdErr,"string: %s\n",address);
}

/************************************************************************************
 *																					*
 *		auxiliary routines															*
 *																					*
 ************************************************************************************/

VOID remove_process(ADDRESS wsp)
{
	/* removes process from simulated scheduling list */
	
	if (sys.processor_state.low_front==NotProcessP)
		return;
	
	if (sys.processor_state.low_front==wsp)
	{
		if (sys.processor_state.low_back==wsp)
			sys.processor_state.low_front=NotProcessP;
		else
			sys.processor_state.low_front=get_next_process(wsp);
	} else
	{
		ADDRESS current_process;
		ADDRESS next_process;
	
		for(current_process=sys.processor_state.low_front;current_process!=sys.processor_state.low_back;current_process=next_process)
		{
			next_process=get_next_process(current_process);
			
			if (next_process==wsp)
			{
				if (next_process==sys.processor_state.low_back)
					sys.processor_state.low_back=current_process;
				else
					set_next_process(current_process,get_next_process(wsp));
				return;
			}
		}
	}
}

/************************************************************************************
 *																					*
 *		the debugger process														*
 *																					*
 ************************************************************************************/

VOID handle_debug_commands(VOID)
{
	execute_commands(debug_command_table,"debugger> ");
}

VOID debug_process(VOID)
{		
	for(;;)
	{
		SIGNUM					signal;
		WSP						*caller;
		ADDRESS					resume_code;
		REDUCER_START_ROUTINE	saved_restart_routine;
		
		signal=|sys.debug.signal_channel;		/* wait for work */
				
		/*	calls to debugger are always made by high prio processes. These
			should have saved vital information (regs etc.) already. When the
			debugger is ready it should restart this high prio process at the
			place it stopped. debugging could overwrite the continuation address
			however, so this address is saved first
		*/
		
		caller=(WSP *)((WORD)signal.wsp&~1);
		resume_code=get_iptr(caller);
		
		set_round_mode(ToNearest);
		
		/* save high prio time of interrupt */
		
		sys.processor_state.time=ldtimer();
		
		/* save low prio scheduling list and empty it */
		
		savel(&sys.processor_state.low_front);
		stlf(NotProcessP);
		saved_restart_routine=get_restart_routine();
		set_restart_routine(simulate_run_reducer);

		/* save current contents of low memory and kill interrupted low prio proc */
		
		move(&sys.processor_state.link0_out,mint(),18*sizeof(WORD));
		*(WORD *)hard_WdescIntSaveLoc=(WORD)NotProcessP;

		/*	the process that is being debugged is still in the low prio
			scheduling list (at the end normally). It should be removed
			(which keeps the corespondence with the real transputer scheduling
			list where running processes don't reside either)
		*/
		
		remove_process(sys.processor_state.Wptr);
		
		block_heap_messages();
		execute_commands(debug_command_table,"debugger> ");
		unblock_heap_messages();

		/*	restore contents of saved low memory (except for link words,
			event word, and timer lists... */
		
		move(hard_WdescIntSaveLoc,&sys.processor_state.WdescIntSaveLoc,7*sizeof(WORD));

		/* restore low prio scheduling list */
		
		stlf(sys.processor_state.low_front);
		stlb(sys.processor_state.low_back);
		set_restart_routine(saved_restart_routine);

		/* and restart caller at high prio */

		set_iptr(caller,resume_code);
		runp(caller,HighPrio);		
	}
}

/********************************************************************************
 *																				*
 *		initialisation															*
 *																				*
 ********************************************************************************/

static COMMAND debug_commands[]=
{
	{"quit",		parse_e,		quit},
	
	/* flow of control */
	
	{"run",			parse_e,		run},
	{"sim",			parse_e,		simulate_instructions},
	{"s",			parse_n,		step_instruction},
	{"ja",			parse_n,		jump_absolute_code},
	{"j",			parse_e,		jump_code},
	
	/*	reducer data structures */
	
	{"locals",		parse_n,		locals},
	{"stacks",		parse_n,		stacks},
	{"regs",		parse_e,		regs},
	
	/* heap data structures */
	
	{"heap",		parse_e,		heap},
	{"node",		parse_n,		node},
	{"anode",		parse_n,		anode},
	{"anodes",		parse_n,		anodes},
	{"lnode",		parse_n,		lnode},
	{"table",		parse_n,		table},
	{"tablegraphs",	parse_e,		tablegraphs},
	
	/* instructions */
	
	{"i",			parse_nn,		instructions},
	{"ri",			parse_nn,		relative_instructions},
	{"ip",			parse_e,		instruction_page},
	
	/* settings */
	
	{"verbose",		parse_e,		verbose},
	{"brief",		parse_e,		brief},
	
	/* general data */
	
	{"lowmem",		parse_e,		low_mem},
	{"lowps",		parse_e,		lowps},
	{"prio",		parse_e,		prio},
	{"ps",			parse_e,		show_processes},

	{"hunlock",		parse_e,		unlock_heap},
	
	{"cstring",		parse_n,		cstring}
	
};

BOOL initialise_debugger(VOID)
{	
	if (!(debug_command_table=get_command_table(50)))
		return False;
	
	if (!define_commands(debug_command_table,debug_commands,sizeof(debug_commands)/sizeof(COMMAND)))
		return False;
		
	debug_verbose=False;
	
	return True;
}