#include "graphs.h"

#include "messageno.h"
#include "system.h"
#include "nodes.h"
#include "io.h"
#include "packets.h"
#include "processes.h"
#include "hardware.h"
#include "depend.h"
#include "heapalloc.h"
#include "heapcopy.h"
#include "heapcollect.h"

#ifdef __INCLUDE_STANDARD__
#include <stdlib.h>
#include <string.h>
#endif

/********************************************************************************
 *																				*
 *		import system routines													*
 *																				*
 ********************************************************************************/

extern NODE_INFO	node_info;
extern NET_INFO		net_info;
extern OPTIONS		options;

extern BOOL			collector_running;

extern SYS			sys;

/********************************************************************************
 *																				*
 *		typedefs																*
 *																				*
 ********************************************************************************/

typedef struct VHEAP_GRAPH
{
	VHEAP_BLOCK_ADMIN	admin;
	NODE				*remote_address;
	BYTE				graph[VariableArraySize];
	
} VHEAP_GRAPH;

#define first_node(graph)	((NODE *)&graph->graph)
#define last_node(graph)	((NODE *)((ADDRESS)&graph->remote_address+graph->admin.initialised))

/********************************************************************************
 *																				*
 *		decrement reference count												*
 *																				*
 ********************************************************************************/

static VOID handle_decref(WORD kind,WORD source,WORD size,DECREF_PACKET *packet)
{
	/* this should go well during garbage collections */
	
	decrement_reference_count(packet->id.dest,1<<packet->weight);
}

/********************************************************************************
 *																				*
 *		remote creation of channels												*
 *																				*
 ********************************************************************************/

static VOID handle_in_ids(WORD kind,WORD source,WORD size,GETID_PACKET *packet)
{
	WORD n, allocated;
	
	/* handle channel creation */
		
	n=packet->number;
	
	if (n > 0 && n <= (2*options.idbuffsize))
	{
		WORD			*p;
		PUTID_PACKET	*ids;
	
		/* allocate return message (must be changed) */
		
		if(!(ids=(PUTID_PACKET *)alloc_packet(False)))
			return;
		
		/*	perform global entry allocation, at this moment the requesting side
			expects this to succeed */
		
		p=ids->i;
		for (allocated=0; allocated<n;)
		{
			WORD id;
			
			if ((id=halloc_entry())==0)
				continue;					/* retry */
			
			*p++=id;
			++allocated;
		}
				
		ids->number=allocated;
		
		send_packet(MOutIds,source,(allocated+1)*sizeof(WORD),ids,False);
		return;
	}
	fatal_error(False,"invalid no channels requested");
}

extern BOOL channel_from_buffer(WORD pid,WORD proc);

static VOID out_ids(WORD proc)
{
	RELEASE_NODE *wait;

	if (table_idwait(proc)==NULL)
		return;							/* no buffers allocated */
	
	/*	ensure safe status of heap: this is required by the channel_from_buffer
		routine. Note that it may change heap addresses.
	*/
	
	low1();
	high();
	
	/*	restart suspended reducers */
	
	while ((wait=table_idwait(proc))!=NULL)
	{
		WORD pid;
		
		pid=wait->pid;
		
		if (!channel_from_buffer(pid,proc))
			break;

		/*	remove reducer from waiting list and run it */
		
		table_idwait(proc)=table_idwait(proc)->wait;
		run_reducer(pid);
	}
}

static VOID handle_out_ids(WORD kind,WORD source,WORD size,PUTID_PACKET *packet)
{
	WORD	n;
	WORD	*buffer;
	
	n=packet->number;
	
	if (n <= options.idbuffsize)
	{
		/*	first try to fill first buffer */
		
		buffer=(WORD *)(&table_idbuff1(source)->string_node.string->characters);

		if (*buffer != 0)
		{
			/* the first buffer is not empty, fill the other one */
			
			buffer=(WORD *)(&table_idbuff2(source)->string_node.string->characters);
		}
		move(buffer,&packet->number,(n+1)*sizeof(WORD));
		out_ids(source);
		return;
	}
		
	if (n == 2*options.idbuffsize)
	{
		/* fill both buffers */

		buffer=(WORD *)(&table_idbuff1(source)->string_node.string->characters);
		*buffer++=options.idbuffsize;
		move(buffer,&packet->i,options.idbuffsize*sizeof(WORD));
		
		buffer=(WORD *)(&table_idbuff2(source)->string_node.string->characters);
		*buffer++=options.idbuffsize;
		move(buffer,(((WORD *)(&packet->i))+options.idbuffsize),options.idbuffsize*sizeof(WORD));
		
		out_ids(source);
		return;
	}
	fatal_error(False,"put ids");
}

/********************************************************************************
 *																				*
 *		reservation of graphs/reducers											*
 *																				*
 ********************************************************************************/

static VOID allocate_lazy_reducer (WORD id)
{
	/* allocate a reducer (if none exists) */

	if(table_graph_reducer(id) == 0)
	{
		RESERVED_NODE		*reserved_node;
		WSP					*regs;
		WORD				pid;
	
		/* allocate reducer */
		
		pid=alloc_reducer(sys.code_info.ea_process_code);
		table_graph_reducer(id)=pid;
		
		/*	allocate reserved node and put it on top of the stack of
			the reducer */
	
		reserved_node			=alloc_words(3);
		reserved_node->code		=sys.code_info.reserve_code;
		reserved_node->wait		=NULL;

		regs				=pid_to_regs(pid);
		regs->a1			=(NODE *)reserved_node;
		regs->extra_local	=1;
	}
}

static VOID allocate_reducer(WORD id)
{
	RESERVED_NODE	*reserved_node;
	WSP				*regs;
	WORD			pid;

	if(table_graph_reducer(id) != 0)
		return;
	
	/*	no reducer available yet, try allocate one. */

	if ((reserved_node=halloc_words(3))==NULL)
		return;

	if((pid=halloc_reducer(sys.code_info.ea_process_code))==0)
		return;
	
	/*	if allocation of reducer fails, the reserved node will
		be removed by the garbage collector
	
		the allocated reducer is inactive. It will get
		a reserved node on top of its stack; the graph
		entry will get the pid of the reducer later
	*/

	table_graph_reducer(id)=pid;
	
	reserved_node->code		=sys.code_info.reserve_code;
	reserved_node->wait		=NULL;
		
	regs				=pid_to_regs(pid);	
	regs->a1			=(NODE *)reserved_node;
	regs->extra_local	=1;
}

static VOID get_data(WORD sourceproc,WORD sourceid,WORD destid,BOOL is_graph,BOOL reserved,BOOL reduce)
{
	IN_PACKET	*in;
	WORD		kind;

	/* allocate return message (must be changed) */
	
	if(!(in=alloc_packet(False)))
		return;
	
	/*	return message to source of message, to force a (re)transmission,
		reserved should be False if another reserve message is required */
	
	in->id.dest			=sourceid;
	in->id.source		=destid;	
	in->reserved		=reserved;
	in->reduce			=reduce;

	if (is_graph)
		kind=MInGraph;
	else
		kind=MInBlock;
	
	send_packet(kind,sourceproc,sizeof(IN_PACKET),in,False);
}

static VOID remove_data(WORD sourceproc,WORD sourceid)
{
	DECREF_PACKET	*dref;

	/* allocate return message (must be changed) */
	
	if(!(dref=alloc_packet(False)))
		return;
	
	/*	return message to source of message, to force a (re)transmission,
		reserved should be False if another reserve message is required */
	
	dref->id.dest		=sourceid;
	dref->weight		=MaxWeight;
	dref->processor		=sourceproc;
	
	send_packet(MDecRef,sourceproc,sizeof(DECREF_PACKET),dref,False);
}

static VOID handle_reserve (WORD kind,WORD source,WORD size,ALLOCATION_PACKET *packet)
{
	/*	handle reservation (reserve messages are only generated by
		the runtime system) */
	
	WORD	id;
	BOOL	reserved;
	
	/* perform allocation */
	
	id=packet->id.dest;
	
	/*	first try to allocate a reducer if necessary. If this fails, the
		graph can be sent anyway. It will then be allocated later */
	
	if (packet->reduce)
		allocate_reducer(id);

	reserved=True;
	
	if (packet->size > 0 && !halloc_gblock(id,packet->size))
		reserved=False;

	get_data(source,packet->id.source,packet->id.dest,kind==MReserveGraph,reserved,packet->reduce);	
}

/********************************************************************************
 *																				*
 *		reception of graphs														*
 *																				*
 ********************************************************************************/

static VOID release_reducers(RELEASE_NODE *release_node)
{
	/*	only called when heap is in consistent state. This has to do with
		request list administration that is performed at low priority: if
		a reducer is adding itself to the waiting list, one should wait
		for this to complete.
		
		The waiting list is not NULL and should be saved in a table
		entry temporarily so that it is preserved over garbage collections
		when returning graphs to remote processors		
	*/
	
	WORD list;
	
	list=alloc_entry();
	
	while(release_node != NULL)
	{
		WORD release_id;

		table_graph(list)	=(NODE *)release_node->wait;
		release_id			=release_node->pid;

		if (release_node->code == sys.code_info.release_proc1 ||
			release_node->code == sys.code_info.release_proc)
			run_reducer(release_id);
		else if (release_node->code == sys.code_info.release_request1 ||
				release_node->code == sys.code_info.release_request)
		{
			/*
				release remote reducers that are waiting for the graph with
				the given id. This situation occurs when 
			*/
			
			REQUEST_NODE *request_node;

			request_node=table_graph_requests(release_id);
	
			while (request_node != NULL)
			{
				REQUEST	request;
		
				request			=request_node->request;
				request_node	=request_node->wait;
				
				table_graph_requests(release_id)=request_node;
				
				return_hnf(release_id,&request);
		
				if (request_node != NULL)
					request_node=table_graph_requests(release_id);
			}
			
		} else
			fatal_error(False,"release reducer");

		release_node=(RELEASE_NODE *)table_graph(list);
	}
	
	/* free the previously allocated entry */
	
	decrement_reference_count(list,RefMax);
}

static VOID block_received(WORD id, BOOL is_graph)
{
	if (table_graph(id)==NULL)
	{
		/*	the table does not contain a graph to be overwritten, this can only
			happen for graphs (closures) */
		
		if (is_graph)
		{
			GRAPH_ENTRY		*entry;
			VHEAP_GRAPH		*graph;
			UWORD			start_time;
	
			start_time=ldtimer();

			entry=&sys.comm_info.id_table[id].graph;
			
			graph=(VHEAP_GRAPH *)entry->reserved->string_node.string;
			AdjustPointers(first_node(graph),last_node(graph),graph->remote_address);
			
			entry->reserved		=NULL;
			entry->graph		=first_node(graph);
			
			/*	if a reducer has been allocated, it will already known the
				id of the graph it should evaluate. */
			
			add_copy_time(start_time);
			
		} else
			fatal_error(False,"block_received: graph expected");

	}else
	{
		/*	a graph (channel node) must be overwritten. This node should
			be a reserved one */

		GRAPH_ENTRY		*entry;
		RELEASE_NODE	*wait_list;

		/* needed at this moment as adding an entry in the waitlist
		   of a channel node happens at low priority and adjusting status etc.
		   as well */
	 	
	 	low1();
	 	high();
	 	
		/*	only now the graph may be restored,
			otherwise pointers may have become incorrect
			blocks do not need to be restored */

		entry=&sys.comm_info.id_table[id].graph;
		wait_list=entry->graph->reserved_node.wait;
		
		if (is_graph)
		{
			VHEAP_GRAPH *graph;
			UWORD start_time;

			start_time=ldtimer();
			
			graph=(VHEAP_GRAPH *)entry->reserved->string_node.string;
			AdjustPointers(first_node(graph),last_node(graph),graph->remote_address);
			
			add_copy_time(start_time);

			/* overwrite channel with root of graph */
			
			move(entry->graph,first_node(graph),BasicNodeSize);
			
		} else
		{
			STRING_NODE *block;

			block=(STRING_NODE *)entry->reserved;

			/*
				overwrite channel with block, note that the 'intialised' field of the
				block should not be included. It indicates the size field of the
				block though and it will become the size field
			*/
		
			entry->graph->string_node.descriptor	=block->descriptor;
			entry->graph->string_node.string		=(VCLEAN_STRING *)&((VHEAP_BLOCK *)block->string)->admin.initialised;
		}

		/* release waiting reducers */

		entry->reserved=NULL;	

		if (wait_list != NULL)
			release_reducers(wait_list);
		
		/* decrement ref count of channel node at correct index */
			
		decrement_reference_count(id,RefMax);
	}
}

/*	a reducer may only be started if the whole graph
	is available (true if the graph is not NIL) and the reducer
	has been allocated.
	
	A reducer will either be allocated explicitly by a
	reserve message, or implicitly when a request comes in.

	At this moment the startup code of a reducer will be
	ignored. It has been set already during allocation.
	
	NOTE: at this moment the instruction 'new_ext_reducer' has
	become obsolete. The reduction code is sent together with
	the graph that must be reduced. This has the advantage that
	the code cannot arrive before (or after) the graph does:
	-	no check necessary at end of graph reception if code already available
		(nil stored in regs if not, it may arrive later)
	-	allocation of a reducer is done together with reservation
		of the graph. One cannot start up reducers on some graph
		more than once and allocation means: a reducer WILL be started
		on this graph.
*/

static VOID start_reducer (WORD id)
{
	/*	intialise reducer registers: a1 should already contain a
		reserved node (possibly containing requests already). The
		graph entry must be made to refer this reserved node. This
		will avoid starting up multiple reducers on the same graph.
	*/

	GRAPH_ENTRY	*entry;
	WORD		pid;
	
	entry=&sys.comm_info.id_table[id].graph;	
	pid=entry->pid;
			
	if (pid != 0)
	{
		WSP *regs;

		regs=pid_to_regs(pid);

		regs->a2			=entry->graph;
		regs->extra_local	=2;

		entry->graph		=regs->a1;						
		
		run_reducer(pid);
	}
}

static VOID handle_out(WORD kind,WORD source,WORD size,VPACKET *packet)
{	
	WORD	id,sourceid,position,total;
	ADDRESS	data, reduce;
	BOOL	is_graph;
	
	VOUT_GRAPH_PACKET	*graph;
	VOUT_BLOCK_PACKET	*block;
	
	is_graph=(kind==MOutGraph);
	
	if (is_graph)
	{
		graph		=(VOUT_GRAPH_PACKET *)packet;
		id			=graph->id.dest;
		sourceid	=graph->id.source;
		total		=graph->size;
		reduce		=graph->reduce;
		position	=graph->position;
		data		=(ADDRESS)graph->graph;
		size		=size-vsizeof(VOUT_GRAPH_PACKET);
	} else
	{
		block		=(VOUT_BLOCK_PACKET *)packet;
		id			=block->id.dest;
		sourceid	=block->id.source;
		total		=block->size;
		reduce		=NULL;
		position	=block->position;
		data		=(ADDRESS)block->data;
		size		=size-vsizeof(VOUT_BLOCK_PACKET);
	}

	while(collector_running)
		procresch();

	/* first check if reducer should be allocated */
	
	if (reduce != NULL)
		allocate_reducer(id);

	/* try store packet */
	
	if (size > 0)
	{
		VHEAP_BLOCK	*reserved_area;

		/* total should be non-zero as well */
		
		if (table_graph_reserved(id)==NULL)
		{
			/*	no memory has been reserved yet, this means that no reserve
				message has been sent for this graph. Its reference count
				must be decreased explicitly after allocation has succeeded.
				if allocation fails, all packets should be discarded. The
				graph should then be retransmitted */
			
			if (table_refnum(id)!=RefAllocFail)
			{
				/*	first packet that arrives, try allocate memory. If this fails,
					go to failed mode and discard all subsequent packets
				*/
	
				if (!halloc_gblock(id,total))
				{
					/* if a reducer was allocated, it should be released */
					
					if (table_discarded(id) != 0)
						runtime_warning(False,"discarding reducer");
						
					table_refnum(id)	=RefAllocFail;
					table_discarded(id)	=0;
				}
			}
			
			if (table_refnum(id)==RefAllocFail)
			{
				/*	allocation has failed earlier (or above),
					this packet should be discarded
				*/
				
				WORD *discarded;
				
				discarded=&table_discarded(id);
				
				*discarded+=size;
				
				if (*discarded==total)
				{
					/* restore table entry */
					
					*discarded=0;
					table_refnum(id)=RefMax;
					
					/*	get graph again, after trying to reserve memory */
					
					get_data(source,sourceid,id,is_graph,halloc_gblock(id,total),reduce!=NULL);
				}
				return;
			}
		
			/*	allocation succeeded here, decrement the reference count
				of the original graph: it can (and will) be stored here
			*/
			
			remove_data(source,sourceid);
		}
	
		reserved_area=reserved_graph(id);
		move(((ADDRESS)&reserved_area->data)+position,data,size);
		reserved_area->admin.initialised+=size;
		
		if (reserved_area->admin.initialised==total)
			block_received(id,is_graph);
	}
	
	if (reduce != NULL && table_graph(id)!=NULL)
	{
		/*	ensure a reducer is available, the allocation routine may
			cause this routine to wait a little */
		
		while (table_graph_reducer(id)==0)
			allocate_reducer(id);
		
		start_reducer(id);
	}
}

static VOID handle_out_basic(WORD kind,WORD source,WORD size,OUT_BASIC_PACKET *packet)
{
	/*	overwrite channel node with small HNF node */
	
	WORD				id;
	GRAPH_ENTRY			*entry;
	RELEASE_NODE		*wait_list;
	
	id=packet->id.dest;
	
	while(collector_running)
		procresch();

	if (table_graph(id)==NULL)
	{
		fatal_error(False,"OutBasic: no channel node to overwrite");
		return;
	}

	/* needed at this moment as adding an entry in the waitlist
	   of a channel node happens at low priority and adjusting status etc.
	   as well */
 	
 	low1();
 	high();
 	
	entry=&sys.comm_info.id_table[id].graph;
	
	wait_list=entry->graph->reserved_node.wait;
	
	/* overwrite channel node */
	
	move(entry->graph,packet->basic,BasicNodeSize);
	
	entry->reserved=NULL;
	
	if (wait_list != NULL)
		release_reducers(wait_list);
		
	/* decrement ref count of channel node at correct index */
	
	decrement_reference_count(id,RefMax);
}

/********************************************************************************
 *																				*
 *		sending of graphs														*
 *																				*
 ********************************************************************************/

static VOID handle_in(WORD kind,WORD source,WORD size,IN_PACKET *packet)
{
	WORD	table_id;
	WORD	rsize;
	
	/* size in cons1->tuple2->string */

	table_id=packet->id.dest;	
	rsize=table_graph(table_id)->hnf2_node.arg1->hnf2_node.arg2->string_node.string->length;
	
	if(!packet->reserved)
	{
		/* reservation failed, try again */

		if (rsize > ReserveBoundary)
		{
			ALLOCATION_PACKET *reserve_packet;
				
			/*	no memory could be reserved at remote processor for result.
				Just try again
			*/
			
			if(table_graph(table_id)==NULL)
			{	fatal_error(0,"result graph is missing");
				return;
			}
			
			if(!(reserve_packet=(ALLOCATION_PACKET *)alloc_packet(False)))
				return;
			
			reserve_packet->id.dest				=packet->id.source;
			reserve_packet->id.source			=table_id;
			reserve_packet->size				=rsize;
			reserve_packet->reduce				=packet->reduce;
			
			if (kind == MInGraph)
				kind=MReserveGraph;
			else
				kind=MReserveBlock;
			
			send_packet(kind,source,sizeof(ALLOCATION_PACKET),reserve_packet,False);
			return;
		}
		
		/* send graph again, do not decrement reference count yet */
				
		send_heap_message(table_graph(table_id));
		return;
	}

	/* allocation succeeded: graph can be removed afterwards */
	
	send_heap_message(table_graph(table_id));
	decrement_reference_count(table_id,RefMax);
}

/************************************************************************************
 *																					*
 *		remote graph copying														*
 *																					*
 ************************************************************************************/

VOID return_hnf(WORD id,REQUEST *request)
{
	NODE *graph;
	WORD result_size,copy,reserve_kind;

	/*	build linear copy of HNF graph in separate table entry. Note that for
		basic nodes no copy has to be made: flat data can be copied from the
		heap directly (e.g. strings) and 'small' nodes do not even require
		reservation as they fit in a channel node */

	graph=table_graph(id);
	
	if (request->address.processor == node_info.id)
		runtime_warning(False,"return_hnf: sending to this processor");
	
	if (hnf_arity(graph) == 0)
	{
		if (get_code(graph) != (ADDRESS)sys.descriptor_info.string2_offset)
		{
			/* small basic node */
			
			OUT_BASIC_PACKET *out;
			NODE *message;
			
			message=alloc_hpacket(MOutBasic,request->address.processor,sizeof(OUT_BASIC_PACKET),&out);
			
			out->id.dest			=request->address.id;
			out->id.source			=0;			
			move(out->basic,graph,BasicNodeSize);
			
			send_heap_message(message);
			
			decrement_reference_count(id,1<<request->address.weight);
			return;
			
		}

		/* string node */
		
		copy=alloc_entry();
		table_graph(copy)=table_graph(id);
		block_message(copy,request->address.processor,request->address.id);		
		reserve_kind=MReserveBlock;
		
	} else
	{
		/* general graph */

		copy_graph(table_graph(id),copy);
		graph_message(copy,request->address.processor,request->address.id,NULL);
		reserve_kind=MReserveGraph;
	}
	
	/*	a linear copy is available, decrement ref count of original graph. */

	decrement_reference_count(id,1<<request->address.weight);
	
	result_size=table_graph(copy)->hnf2_node.arg1->hnf2_node.arg2->string_node.string->length;

	/* if the request indicates some space has been allocated already, no
		reserve message is needed */
	
	if (request->reserved > 0)
	{
		/*	check for matching sizes, and send copy if allright. Decrement ref
			count of copy as soon as it has been saved in a message
		*/
		
		if (request->reserved < result_size)
		{
			fatal_error(False,"reserved size %d < result size %d",request->reserved,result_size);
			return;
		}
		
		send_heap_message(table_graph(copy));
		decrement_reference_count(copy,RefMax);
	
	} else
	{
		transmit(copy,reserve_kind,result_size,False,
				request->address.processor,request->address.id);
	}
}

/********************************************************************************
 *																				*
 *		request handling routines												*
 *																				*
 ********************************************************************************/

static WORD request_status;
static WSP *request_wsp;

static WORD		current_request_id;
static REQUEST	current_request;

static VOID handle_request_graph(WORD kind,WORD source,WORD size,REQUEST_PACKET *packet)
{
	/* temporary solution to avoid problems when some low prio process isn't ready */
	
	int wait_repeat;
		
	wait_repeat=0;
	
	while(request_status==PActive)
	{
		/* give request handler opportunity to run, do not use low1 */
		
		low();
		high();
		
		if(++wait_repeat >= 10)
		{
			runtime_warning(False,"failed %d times to administer request. (ignored!)",wait_repeat);
			return;
		}
	}
	
	current_request_id	=packet->id;
	current_request		=packet->request;
		
	request_status=PActive;	
	runp1(request_wsp);	
}

/************************************************************************************
 *																					*
 *		request handling process													*
 *																					*
 ************************************************************************************/

static VOID enter_request(VOID)
{
	REQUEST_NODE	*request_node;
	GRAPH_ENTRY		*entry;

	/*	if this is the first request over this entry, the id
		of this entry must be put in the waiting list of the reserved
		graph. Furthermore the request must be put in the request
		list of this entry.
		
		If there is no graph yet, a reducer will have been allocated
		for it and a reserved node can be found on the stack of this reducer.
		If a graph exists, it will have been reserved already.
	*/

	if (table_graph_requests(current_request_id)==NULL)
	{
		RELEASE_NODE	*release_node;
		RESERVED_NODE	*reserved_node;
		
		/* this is the first request */
		
		request_node=alloc_bytes(sizeof(RELEASE_NODE)+sizeof(REQUEST_NODE));
		request_node->code=sys.code_info.request_code;
		
		release_node=(RELEASE_NODE *)(request_node+1);
		reserved_node=&table_graph(current_request_id)->reserved_node;
		
		if (reserved_node == NULL)
		{
			WSP *regs;
						
			/*	extract reserved node from reducer stack */
			
			regs=pid_to_regs(table_graph_reducer(current_request_id));
			reserved_node=(RESERVED_NODE *)regs->a1;
		}
		
		/*	create release node */
		
		if (reserved_node->wait==NULL)
		{
			release_node->code	=sys.code_info.release_request;
			release_node->wait	=NULL;
		} else
		{
			release_node->code	=sys.code_info.release_request1;
			release_node->wait	=reserved_node->wait;
		}
		release_node->pid=current_request_id;
		
		/* put release node in reserved node */
		
		reserved_node->code	=sys.code_info.reserve1_code;
		reserved_node->wait	=release_node;

	} else
	{
		/* requests already exist */
		
		request_node=alloc_bytes(sizeof(REQUEST_NODE));
		request_node->code=sys.code_info.request1_code;
	}

	entry=&sys.comm_info.id_table[current_request_id].graph;
	
	/* enter request node */
	
	request_node->wait		=entry->requests;
	request_node->request	=current_request;

	entry->requests=request_node;
}

static VOID safe_handle_request(VOID)
{
	NODE *graph;
	
	graph=table_graph(current_request_id);
	
	if (graph==NULL)
	{
		/*	If the graph is not yet available the request should be
			administered and a reducer should be allocated. As soon as the
			graph arrives, the reducer will be started on it and this reducer
			will answer the request later.
		*/

		allocate_lazy_reducer(current_request_id);
		enter_request();
		return;
	}

	if (is_closure(graph))
	{		
		/*	If the graph is not yet in HNF, the request should be administered
			as well. If the root node is reserved, the request should merely be
			stored. Otherwise a reducer should be started as well...
		   
			requests for channel nodes form no problem: if they are reserved, they will
			become updated and releasing them will forward the result. If they are
			not yet reserved, a reducer will be started on them and this will forward
			the request.
		*/
		
		if (!is_closure_type(graph,ReservedNode))
		{
			allocate_lazy_reducer(current_request_id);
			start_reducer(current_request_id);
		}
		
		enter_request();
		return;
	}
	
	/*	HNF graph, return it */
	
	return_hnf(current_request_id,&current_request);
}

/* these are process that perform tasks that must be started at low priority to
   ensure the heap is in a consistent state. They are started explicitly by some
   other process at low priority. If necessary they increase priority themselves */

VOID request_process(VOID)
{	
	request_status=PWaiting;
	
	stopp(&request_wsp);
	
	for(;;)
	{	
		/* you've been started at low prio, so the heap is allright */
		
		high();

		safe_handle_request();
		
		request_status=PWaiting;
		stopp(&request_wsp);
	}
}

/********************************************************************************
 *																				*
 *		initialisation															*
 *																				*
 ********************************************************************************/

static RECEIVE_ACTION_TABLE graph_actions[]=
{
	{MReserveGraph,		{handle_reserve,			ignore_reception}},
	{MReserveBlock,		{handle_reserve,			ignore_reception}},
	{MInGraph,			{handle_in,					ignore_reception}},
	{MInBlock,			{handle_in,					ignore_reception}},
	{MOutGraph,			{handle_out,				ignore_reception}},
	{MOutBlock,			{handle_out,				ignore_reception}},
	{MOutBasic,			{handle_out_basic,			ignore_reception}},
	
	{MRequestGraph,		{handle_request_graph,		ignore_reception}},
	{MInIds,			{handle_in_ids,				ignore_reception}},
	{MOutIds,			{handle_out_ids,			ignore_reception}},
	
	{MDecRef,			{handle_decref,				ignore_reception}}
};

BOOL initialise_graph_manager(VOID)
{
	request_status=PActive;
	
	set_packet_actions(graph_actions,sizeof(graph_actions)/sizeof(RECEIVE_ACTION_TABLE));
	
	return True;
}
