/********************************************************************************
 *																				*
 *			routines to allocate/initialise memory in the heap					*
 *																				*
 ********************************************************************************/

#include "heapalloc.h"

#include "errorno.h"
#include "heapcollect.h"
#include "depend.h"
#include "processes.h"
#include "io.h"
#include "memory.h"
#include "packets.h"
#include "macros.h"

#ifdef __INCLUDE_STANDARD__

#include <string.h>

#endif

extern OPTIONS			options;
extern NODE_INFO		node_info;
extern NET_INFO			net_info;
extern PROCESSOR_INFO	processor_info;

extern SYS				sys;

extern BOOL				collector_running;

/********************************************************************************
 *																				*
 *		initialisation of the heap												*
 *																				*
 ********************************************************************************/

VOID invalid_heap(VOID)
{
	/*	used to indicate that the heap is not valid anymore, executed at
		startup and after running an application */
	
	REDUCER_LOCALS	*reducer_locals;
	WORD			i,n_reducers;

	n_reducers=node_info.max_reducers-1;

	/*	set some pointers to NULL, messages in the heap are blocked */
	
	block_heap_messages();

	sys.comm_info.first_message	=NULL;
	sys.comm_info.last_message	=NULL;

	/* initialise global table, assuming heap is big enough to hold it ********/

	sys.comm_info.id_table			= NULL;
	sys.comm_info.id_table_size		= 0;
	sys.comm_info.id_table_free		= 0;
	
	/* reset channel admininstration */
	
	sys.comm_info.active_channels			= NULL;
	sys.comm_info.first_inactive_channel	= NULL;
	sys.comm_info.last_inactive_channel		= NULL;
	
	/* reset register administration */

	reducer_locals=&(node_info.int_mem_start->reducer_locals[0]);
	
	sys.reducer_info.on_chip_free_list		= &(reducer_locals[0].wsp);
	sys.reducer_info.on_chip_first_regs		= &(reducer_locals[0].wsp);
	sys.reducer_info.on_chip_last_regs		= &(reducer_locals[n_reducers].wsp);
	
	sys.reducer_info.off_chip_free_list			= NULL;
	sys.reducer_info.off_chip_suspended_regs	= NULL;
	sys.reducer_info.off_chip_first_regs		= NULL;

	for (i=0; i<n_reducers; i++)
	{
		reducer_locals[i].special.next_proc=&(reducer_locals[i+1].wsp);
		reducer_locals[i].wsp.heap_start=NULL;
	}
	reducer_locals[n_reducers].special.next_proc	=NULL;
	reducer_locals[n_reducers].wsp.heap_start		=NULL;	
}

VOID valid_heap (VOID)
{	
	/*	used to (re)intialise heap (before running an application), the
		heap should have been allocated earlier (pointer in 'heap_free_ptr')
	 */

	WORD				i,half_heap_size;
	ADDRESS				heap;
	NULL_ENTRY			*null_entry;
	
	half_heap_size = (options.heap_size)>>1;			/* half of heapsize */
	half_heap_size = (half_heap_size>>2)<<2;			/* multiple of wordlength */
	
	/* initialise heap administration */
	
	heap=sys_root->heap_free_ptr;
	
	sys_root->heap_end_ptr		= heap+half_heap_size;
	sys_root->current_heap		= heap;
	sys_root->garbage_heap		= heap+half_heap_size;
	sys_root->half_heap_size	= half_heap_size;

	/* initialise message list in heap, let ptrs point to nil node ************/
	
	sys.comm_info.first_message							=(NODE *)sys_root->heap_free_ptr;
	sys.comm_info.last_message							=(NODE *)sys_root->heap_free_ptr;
	sys.comm_info.first_message->hnf0_node.descriptor	=0;
	sys_root->heap_free_ptr+=(2*sizeof(WORD));

	/*	initialise global table, assuming heap is big enough to hold it,
		the first entry will be the null entry, that terminates lists,
		the following entries are used to hold buffers for obtaining
		ids. Initially these are empty. */

	sys.comm_info.id_table			= (ENTRY *)sys_root->heap_free_ptr;
	sys.comm_info.id_table_size		= options.init_id_table_size+net_info.size;
	sys.comm_info.id_table_free		= 1+net_info.size;
	
	/* entry 0 is used to end internal waiting lists */
	
	null_entry=&sys.comm_info.id_table[0].null;
	
	null_entry->refnum	=RefNull;
	null_entry->p1		=NULL;
	null_entry->p2		=NULL;
	null_entry->p3		=NULL;
	null_entry->i		=0;
	
	/* the following netsize entries are use to hold id buffers */

	for(i=1; i<=net_info.size; ++i)
	{
		IDBUFF_ENTRY *entry;
		
		entry=&sys.comm_info.id_table[i].idbuff;

		entry->refnum	=RefIdBuff;
		entry->buff1	=NULL;
		entry->buff2	=NULL;
		entry->wait		=NULL;
		entry->i		=0;
	}
	
	for(; i<sys.comm_info.id_table_size; ++i)
	{
		FREE_ENTRY *entry;
		
		entry=&sys.comm_info.id_table[i].free;
		
		entry->refnum	=RefFree;
		entry->next		=i+1;
	}
	
	sys_root->heap_free_ptr+=(sys.comm_info.id_table_size*sizeof(ENTRY));
	
	/* initialise swap space */
	
	sys.reducer_info.swap=(WSP *)(sys_root->heap_free_ptr+sizeof(HARD_WSP));	
	sys_root->heap_free_ptr+=(sizeof(REDUCER_LOCALS));

	if (options.fixed_random)
	{
		/* reset pseudo random generator to fixed value */
		
		sys.general_info.random_rand=4181;
		sys.general_info.random_m	=971;
		sys.general_info.random_a	=11113;
		sys.general_info.random_b	=104322;
	}

	unblock_heap_messages();
}

/********************************************************************************
 *																				*
 *		allocation routine for high prio processes								*
 *		these routines assume the heap is in consistent state and				*
 *		the system stack is empty. The garbage collector cannot be running		*
 *		when such a routine is called.											*
 *																				*
 ********************************************************************************/

VOID *alloc_words(WORD n)
{
	WORD *result;
		
	while ((WORD *)sys_root->heap_free_ptr+n > (WORD *)sys_root->heap_end_ptr)
		collect(n);
	
	result=(WORD *)sys_root->heap_free_ptr;
	sys_root->heap_free_ptr=(ADDRESS)(result+n);
	
	return result;
}

VOID *alloc_bytes(WORD n)
{
	BYTE *result;
	
	n=word_up(n);
	
	while ((BYTE *)sys_root->heap_free_ptr+n > (BYTE *)sys_root->heap_end_ptr)
		collect(n/sizeof(WORD));
	
	result=(BYTE *)sys_root->heap_free_ptr;
	sys_root->heap_free_ptr=(ADDRESS)(result+n);
	
	return result;
}

WORD alloc_entry (VOID)
{
	WORD id;
	
	for (;;)
	{
		id=NewGlobalAddress(NULL,RefMax);

		if (id > 0)
			return id;

		collect(-id);
	}
}

WORD alloc_reducer(VOID *code)
{
	WSP					*new_regs;
	REDUCER_ENVIRONMENT	*new_env;
		
	/* get reducer from freelist */
	
	new_regs=sys.reducer_info.on_chip_free_list;
	
	if (new_regs == NULL)
	{
		/* try allocate an off-chip reducer */
	
		new_regs=sys.reducer_info.off_chip_free_list;

		if (new_regs == NULL)
		{
			new_regs=init_reducer(NULL);

			while ((WORD)new_regs >= 0)
			{
				collect((WORD)new_regs);
				new_regs=init_reducer(NULL);
			}
		} else
			sys.reducer_info.off_chip_free_list=get_next_process(new_regs);

		/*	regs allocated, put in suspended list to keep garbage collector
			from throwing away the reducer */
		
		off_chip_suspend(new_regs);
					
	} else
	{
		/* on-chip reducer available */
		
		if (new_regs->heap_start == NULL)
		{	
			new_regs=init_reducer(new_regs);
			
			while ((WORD)new_regs >= 0)
			{
				collect((WORD)new_regs);
				new_regs=init_reducer(new_regs);
			}
		}
		sys.reducer_info.on_chip_free_list=get_next_process(new_regs);
	}

	new_env=get_process_env(new_regs);
	
	/*	the pid of the reducer should be positive (active), make it inactive
		for the moment, as it is not yet in the scheduling list and it
		should remain possible to swap the reducer out */
	
	set_inactive(new_env);
	new_regs->extra_local=0;
	
	set_iptr(new_regs,code);
	
	return get_pid(new_env);
}

VOID alloc_gblock(WORD id,WORD size)
{
	STRING_NODE			*block_ptr;
	VHEAP_BLOCK_ADMIN	*block;
	WORD				allocated_bytes;

	/*	This routine allocates a global string of the given size, it contains
		an extra word that indicates how much of it has been initialised already
		this is used for remote fills.
	
		the size should be bigger than zero. for table entries that have a
		reserved area already this routine has no effect: it may be possible that
		an earlier attempt to allocate a block has succeeded, while allocation
		of a reducer (or another component) failed.
	*/

	if(sys.comm_info.id_table[id].graph.reserved!=NULL)
		return;
	
	size+=sizeof(STRING_NODE)+sizeof(VHEAP_BLOCK_ADMIN);
	
	/* allocate required amount of bytes in the heap */
	
	allocated_bytes=word_up(size);
	
	while (sys_root->heap_free_ptr+allocated_bytes>sys_root->heap_end_ptr)
		collect(allocated_bytes>>2);
	
	block_ptr	=(STRING_NODE *)sys_root->heap_free_ptr;
	block		=(VHEAP_BLOCK_ADMIN	*)(block_ptr+1);
	
	sys_root->heap_free_ptr+=allocated_bytes;
	
	block_ptr->descriptor=sys.descriptor_info.string2_offset;
	block_ptr->string=(VCLEAN_STRING *)block;
	
	/*	store the amount of reserved data in the the appropriate field
		and indicate nothing has been initialised */
	
	block->reserved		=size-sizeof(STRING_NODE)-sizeof(WORD);
	block->initialised	=0;
	
	sys.comm_info.id_table[id].graph.reserved=(NODE *)block_ptr;
}

/********************************************************************************
 *																				*
 *		allocation routines. The heap may not be in a consistent state,			*
 *		in particular the garbage colletor may be running.						*
 *																				*
 ********************************************************************************/

VOID *halloc_words(WORD n)
{
	WORD *result;
	
	if (collector_running)
		return NULL;
	
	if((WORD *)sys_root->heap_free_ptr+n > (WORD *)sys_root->heap_end_ptr)
	{
		invoke_garbage_collector(n);
		return NULL;
	}
	
	result=(WORD *)sys_root->heap_free_ptr;
	sys_root->heap_free_ptr=(ADDRESS)(result+n);

	return result;
}

WORD halloc_reducer(VOID *code)
{
	/* return pid of reducer or zero */
	
	WSP					*new_regs;
	REDUCER_ENVIRONMENT	*new_env;
	
	if (collector_running)
		return 0;
			
	/* get reducer from freelist */
	
	new_regs=sys.reducer_info.on_chip_free_list;
	
	if (new_regs == NULL)
	{
		/* try allocate an off-chip reducer */
	
		new_regs=sys.reducer_info.off_chip_free_list;

		if (new_regs == NULL)
		{
			new_regs=init_reducer(NULL);

			if ((WORD)new_regs >= 0)
			{
				invoke_garbage_collector((WORD)new_regs);
				return 0;
			}
			
		} else
			sys.reducer_info.off_chip_free_list=get_next_process(new_regs);

		/* regs allocated, put in suspended list */
		
		off_chip_suspend(new_regs);
					
	} else
	{
		/* on-chip reducer available */
		
		if (new_regs->heap_start == NULL)
		{	
			new_regs=init_reducer(new_regs);
			
			if ((WORD)new_regs >= 0)
			{
				invoke_garbage_collector((WORD)new_regs);
				return 0;
			}
		}
		sys.reducer_info.on_chip_free_list=get_next_process(new_regs);
	}

	new_env=get_process_env(new_regs);
	
	/*	the pid of the reducer should be positive (active), make it inactive
		for the moment, as it is not yet in the scheduling list */
	
	set_inactive(new_env);
	new_regs->extra_local=0;
	
	set_iptr(new_regs,code);
	
	return get_pid(new_env);
}

WORD halloc_entry (VOID)
{
	WORD id;
	
	if (collector_running)
		return 0;
	
	id=NewGlobalAddress(NULL,RefMax);

	if (id > 0)
		return id;
	
	invoke_garbage_collector(-id);
	return 0;
}

BOOL halloc_gblock(WORD id,WORD size)
{
	STRING_NODE			*block_ptr;
	VHEAP_BLOCK_ADMIN	*block;
	WORD				allocated_bytes;

	/*	This routine allocates a global string of the given size, it contains
		an extra word that indicates how much of it has been initialised already
		this is useful for remote fills.
	
		the size should be bigger than zero. for table entries that have a
		reserved area already this routine has no effect: it may be possible that
		an earlier attempt to allocate a block has succeeded, while allocation
		of a reducer (or another component) failed.
	*/

	if(sys.comm_info.id_table[id].graph.reserved!=NULL)
	{
		/* block already allocated, indicate success */
		
		return True;
	}

	if (collector_running)
		return False;
	
	size+=sizeof(STRING_NODE)+sizeof(VHEAP_BLOCK_ADMIN);
	
	/* allocate required amount of bytes in the heap */
	
	allocated_bytes=word_up(size);
	
	if(sys_root->heap_free_ptr+allocated_bytes>sys_root->heap_end_ptr)
	{
		invoke_garbage_collector(allocated_bytes>>2);
		return False;
	}
	
	block_ptr	=(STRING_NODE *)sys_root->heap_free_ptr;
	block		=(VHEAP_BLOCK_ADMIN	*)(block_ptr+1);
	
	sys_root->heap_free_ptr+=allocated_bytes;
	
	block_ptr->descriptor=sys.descriptor_info.string2_offset;
	block_ptr->string=(VCLEAN_STRING *)block;
	
	/*	store the amount of reserved data in the the appropriate field
		and indicate nothing has been initialised */
	
	block->reserved		=size-sizeof(STRING_NODE)-sizeof(WORD);
	block->initialised	=0;
	
	sys.comm_info.id_table[id].graph.reserved=(NODE *)block_ptr;
	return True;
}