#include "heapcollect.h"

#include "system.h"
#include "depend.h"
#include "io.h"
#include "hardware.h"
#include "router.h"
#include "processes.h"
#include "packets.h"

#ifdef __INCLUDE_STANDARD__
#include <string.h>
#endif

extern OPTIONS			options;
extern NODE_INFO		node_info;
extern PROCESSOR_INFO	processor_info;

extern SYS				sys;

/********************************************************************************
 *																				*
 *		constants																*
 *																				*
 ********************************************************************************/

#define GarbWaitTime 100

/********************************************************************************
 *																				*
 *		garbage collector (must be started at low prio...)						*
 *																				*
 ********************************************************************************/

static WORD		collect_status;
static WSP		*collect_wsp;
static WORD		collect_heap_need;

BOOL			collector_running;

/* after a garbage collection has taken place, the channel node lists are
   updated. This is done by inspecting the old channel lists. All inactive channels
   remain inactive, furthermore the old active channels that haven't been copied
   (they have not been marked) are deactivated as well. Old active channels that
   contain a mark are put in the active list again; they are still reachable.
   For all inactive channels a decrement reference count message is sent. Note that
   channel nodes over which a request has been sent are globally known, so these cannot
   be deactivated by this routine alone. These will be deactivated explicitly as soon
   as the request is sent: the channel node will be removed automatically when
   overwritten by an HNF node.
*/

static VOID activate_channel_node(CHANNEL_NODE *channel_node)
{
	/* the channel node has already been copied */
	
	channel_node->prev = NULL;
	channel_node->next = sys.comm_info.active_channels;
	
	if (sys.comm_info.active_channels != NULL)	
		sys.comm_info.active_channels->prev = channel_node;
	
	sys.comm_info.active_channels = channel_node;
}

static VOID deactivate_channel_node(CHANNEL_NODE *channel_node)
{
	/* the channel node doesn't contain a forwarding address, copy the node to
	   the new heap space and put it in the list of inactive channels
	*/
	
	move(sys_root->heap_free_ptr,channel_node,sizeof(CHANNEL_NODE));
	channel_node=(CHANNEL_NODE *)sys_root->heap_free_ptr;
	sys_root->heap_free_ptr+=sizeof(CHANNEL_NODE);
	
	channel_node->prev = sys.comm_info.last_inactive_channel;
	channel_node->next = NULL;
	
	if (sys.comm_info.first_inactive_channel == NULL)
		sys.comm_info.first_inactive_channel = channel_node;
	else
		sys.comm_info.last_inactive_channel->next = channel_node;
	
	sys.comm_info.last_inactive_channel=channel_node;
}

static VOID adjust_channel_lists(VOID)
{
	CHANNEL_NODE *old_inactive, *old_active;
	
	old_inactive	=sys.comm_info.first_inactive_channel;
	old_active		=sys.comm_info.active_channels;
	
	/* empty channel lists */
	
	sys.comm_info.first_inactive_channel	=NULL;
	sys.comm_info.last_inactive_channel		=NULL;
	sys.comm_info.active_channels			=NULL;
	
	/* put all old inactive channels in inactive list again */

	for (;old_inactive != NULL; old_inactive=old_inactive->next)
		deactivate_channel_node(old_inactive);

	/* put old active channels in active list if reachable, in inactive list otherwise */

	for (;old_active != NULL; old_active=old_active->next)
	{
		if (((WORD)old_active->code) & 1)
			activate_channel_node((CHANNEL_NODE *)((WORD)old_active->code-1));
		else
			deactivate_channel_node(old_active);
	}
	
	if (sys.comm_info.first_inactive_channel != NULL)
	{
		/* wake up router to send decrement ref count messages */
		
		wake_virtual_link(NLinks);	
	}
}

static VOID remove_special_entry(WORD index, WORD refnum)
{
	/* the reference count should hold the argument refnum */
	
	FREE_ENTRY *entry;
	
	entry=&sys.comm_info.id_table[index].free;
	
	if (entry->refnum != refnum)
		fatal_error(False,"remove_special_entry");
	
	entry->refnum	=RefFree;
	entry->next		=sys.comm_info.id_table_free;
	
	sys.comm_info.id_table_free	=index;
}

/********************************************************************************
 *																				*
 *		decrement reference count...											*
 *																				*
 ********************************************************************************/

VOID decrement_reference_count(WORD index, UWORD amount)
{
	FREE_ENTRY	*entry;
	
	if (index <= 0)
	{
		fatal_error(False,"decref: invalid index %d",index);
		return;
	}
	
	entry=&sys.comm_info.id_table[index].free;
	
	if (entry->refnum >= amount)
		entry->refnum-=amount;
	else
		fatal_error(False,"ref count: %d (-%d)",entry->refnum,amount);

	if (entry->refnum == RefFree)
	{	
		entry->next=sys.comm_info.id_table_free;
		sys.comm_info.id_table_free=index;
	}
}

VOID call_collect(VOID)
{
	UWORD				start_time, stop_time;
	ADDRESS				dest_ptr, end_ptr;
	WSP					*regs;
	REDUCER_ENVIRONMENT *env;	
	
	collector_running=True;
	block_heap_messages();
		
	start_time=ldtimer();
	
	dest_ptr	=sys_root->garbage_heap;
	end_ptr		=dest_ptr+sys_root->half_heap_size;
		
	/************************************************************************
	 *																		*
	 *		allocate new swap space											*
	 *																		*
	 ************************************************************************/
	
	sys.reducer_info.swap=(WSP *)(dest_ptr+sizeof(HARD_WSP));
	dest_ptr+=sizeof(REDUCER_LOCALS);

	/************************************************************************
	 *																		*
	 *		remove stopped off-chip regs									*
	 *																		*
	 ************************************************************************/	
	
	for (regs=sys.reducer_info.off_chip_free_list; regs != NULL; regs=get_next_process(regs))
	{
		/* stopped reducers have positive pids */
		
		env=get_process_env(regs);
		remove_special_entry(env->pid,RefProcess);
	}
	sys.reducer_info.off_chip_free_list=NULL;
	
	/************************************************************************
	 *																		*
	 *		copy suspended off-chip regs if any (doubly linked)				*
	 *																		*
	 ************************************************************************/

	regs=sys.reducer_info.off_chip_suspended_regs;

	if (regs != NULL)
	{
		WSP *new_regs;
		
		new_regs=(WSP *)(dest_ptr+sizeof(HARD_WSP));
		sys.reducer_info.off_chip_suspended_regs=new_regs;
		
		for (;;)
		{
			/* copy regs, ptrs from and to left have been adjusted */
			
			move(dest_ptr,(ADDRESS)regs-sizeof(HARD_WSP),sizeof(REDUCER_LOCALS));
			dest_ptr+=sizeof(REDUCER_LOCALS);

			regs=get_next_process(regs);
			
			if (regs == NULL)
				break;
			
			/* adjust ptrs to and from the right */
			
			set_prev_process(regs,new_regs);
			new_regs=set_next_process(new_regs,dest_ptr+sizeof(HARD_WSP));
		}
	}

	/************************************************************************
	 *																		*
	 *		copy active off-chip regs										*
	 *																		*
	 ************************************************************************/
	
	regs=sys.reducer_info.off_chip_first_regs;
	
	if (regs != NULL)
	{
		WSP *new_regs;
			
		set_next_process(sys.reducer_info.off_chip_last_regs,NULL);
		
		new_regs=(WSP *)(dest_ptr+sizeof(HARD_WSP));
		sys.reducer_info.off_chip_first_regs=new_regs;
		
		for (;;)
		{
			/* copy regs, ptrs from and to left have been adjusted */
			
			move(dest_ptr,(ADDRESS)regs-sizeof(HARD_WSP),sizeof(REDUCER_LOCALS));
			dest_ptr+=sizeof(REDUCER_LOCALS);

			regs=get_next_process(regs);
			
			if (regs == NULL)
				break;
			
			/* adjust ptr to neighbour on the right */
			
			new_regs=set_next_process(new_regs,dest_ptr+sizeof(HARD_WSP));
		}
		sys.reducer_info.off_chip_last_regs=new_regs;
	}

	/********************************************************************************
	 *																				*
	 *		all regs copied, now copy environments of reducers						*
	 *																				*
	 ********************************************************************************/

	sys.copy_info.copy_first_env=(WORD)dest_ptr;
	
	/*	copy environments via on-chip regs, first search for on-chip stopped
		reducers that contain a valid structure */

	for (regs=sys.reducer_info.on_chip_free_list; regs != NULL; regs=get_next_process(regs))
	{
		if (regs->heap_start != NULL)
		{
			/* stopped reducers have positive pids */
			
			env=get_process_env(regs);
			remove_special_entry(env->pid,RefProcess);
			regs->heap_start=NULL;
		}
	}
	
	for (regs=sys.reducer_info.on_chip_first_regs; regs <= sys.reducer_info.on_chip_last_regs; regs=(WSP *)((REDUCER_LOCALS *)regs+1))
	{
		if (regs->heap_start == NULL)
			continue;
		
		env=get_process_env(regs);
		
		/* environment must be copied, leave forwarding pointer in old one, regs
		   should stay the same */
		
		move(dest_ptr,env,sizeof(REDUCER_ENVIRONMENT));
		*(WORD *)env=(WORD)dest_ptr+1;
		
		dest_ptr+=sizeof(REDUCER_ENVIRONMENT);
		
		regs->heap_start	=(WORD *)dest_ptr;
		regs->heap_end		=(WORD *)dest_ptr;
	}
	
	/* copy environments via off-chip regs */

	for (regs=sys.reducer_info.off_chip_suspended_regs; regs != NULL; regs=get_next_process(regs))
	{
		env=get_process_env(regs);
		
		/* environment must be copied, leave forwarding pointer in old one, regs
		   should be adjusted */
		
		move(dest_ptr,env,sizeof(REDUCER_ENVIRONMENT));
		*(WORD *)env=(WORD)dest_ptr+1;
		((REDUCER_ENVIRONMENT *)dest_ptr)->registers=regs;
		
		dest_ptr+=sizeof(REDUCER_ENVIRONMENT);
		
		regs->heap_start	=(WORD *)dest_ptr;
		regs->heap_end		=(WORD *)dest_ptr;
	}
	
	for (regs=sys.reducer_info.off_chip_first_regs; regs != NULL; regs=get_next_process(regs))
	{
		env=get_process_env(regs);
		
		/* environment must be copied, leave forwarding pointer in old one, regs
		   should be adjusted */
		
		move(dest_ptr,env,sizeof(REDUCER_ENVIRONMENT));
		*(WORD *)env=(WORD)dest_ptr+1;
		((REDUCER_ENVIRONMENT *)dest_ptr)->registers=regs;
		
		dest_ptr+=sizeof(REDUCER_ENVIRONMENT);
		
		regs->heap_start	=(WORD *)dest_ptr;
		regs->heap_end		=(WORD *)dest_ptr;
	}

	sys.copy_info.copy_last_env=(WORD)dest_ptr;

	/********************************************************************************
	 *																				*
	 *		call low level garbage collector										*
	 *																				*
	 ********************************************************************************/
	
	sys_root->heap_free_ptr	=dest_ptr;
	sys_root->heap_end_ptr	=end_ptr;
		
	CopyCollect();
	
	/********************************************************************************
	 *																				*
	 *		switch heaps, use end_ptr temporarily									*
	 *																				*
	 ********************************************************************************/
	
	end_ptr=sys_root->garbage_heap;
	sys_root->garbage_heap=sys_root->current_heap;
	sys_root->current_heap=end_ptr;
	
	adjust_channel_lists();
	unblock_heap_messages();
	
	stop_time=(ldtimer()-start_time);

	/************************************************************************
	 *																		*
	 *		statistics														*
	 *																		*
	 ************************************************************************/
	
	i64incu32(processor_info.ticks_collecting,stop_time);
	
	/************************************************************************
	 *																		*
	 *		if only little memory freed, keep reducers pending until		*
	 *		outstanding heap messages have been sent						*
	 *																		*
	 ************************************************************************/

	dest_ptr=sys_root->heap_free_ptr+(sys_root->half_heap_size/10)+sys.copy_info.words_needed;

	if (dest_ptr > sys_root->heap_end_ptr)
	{
		while (sys.comm_info.first_inactive_channel != NULL || !(is_nil(sys.comm_info.first_message)))
			procresch();
		
		if (options.verbose)
			runtime_warning(False,"less than 10%% freed");
	}

	collector_running=False;
	
	/************************************************************************
	 *																		*
	 *		messages														*
	 *																		*
	 ************************************************************************/
		
	if (options.show_garbage_collection)
	{
		WORD used;
		
		used=(WORD)sys_root->heap_free_ptr-(WORD)sys_root->current_heap;
	
		runtime_warning(False,"%d bytes in use after collection",used);
		return;
	}
}

VOID collector(VOID)
{
	collect_heap_need	=0;
	collect_status		=PWaiting;
	collector_running	=False;
	
	stopp(&collect_wsp);
	
	for(;;)
	{
		high();

		/*	it could be possible that some previous low priority process
			has invoked the garbage collector already. This could mean
			enough space has become available meanwhile.
		*/
		
		if ((WORD *)sys_root->heap_free_ptr+collect_heap_need > (WORD *)sys_root->heap_end_ptr)
		{
			sys_root->copy_info->words_needed=collect_heap_need;
			call_collect();
		}
		
		collect_status=PWaiting;
		stopp(&collect_wsp);
	}
}

VOID collect(WORD need)
{
	/*	This routine may only be used by high priority processes. It calls the
		garbage collector directly. The caller must ensure that:
		-	the system stack is empty (the garbage collector uses it)
		-	the heap is in a consistent state
	*/
	
	sys_root->copy_info->words_needed=need;
	call_collect();
}

VOID invoke_garbage_collector(WORD need)
{
	/*	temporary solution to avoid problems when low prio process isn't ready.
		This routine should not be called when the collector is already running */
	
	if(collect_status==PActive)
	{
		/*	The collector process has not yet started up, wait a little to
			give it time to start up */
		
		wait(GarbWaitTime);
		return;
	}
	
	collect_status=PActive;	
	collect_heap_need=need;
	
	/* put process at start of low prio scheduling list */
	
	runp1(collect_wsp);
}

