/********************************************************************
 *																	*
 *		file server													*
 *																	*
 ********************************************************************/

#include "hostio.h"

#ifndef _TYPES_
#include "types.h"
#endif

#include "macros.h"
#include "messageno.h"
#include "packets.h"
#include "hostserver.h"
#include "memory.h"
#include "graphs.h"

#ifdef __INCLUDE_STANDARD__

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <errno.h>
#include <sys/types.h>
#include <unistd.h>

#endif

extern NODE_INFO		node_info;
extern NET_INFO			net_info;
extern PROCESSOR_INFO	processor_info;
extern OPTIONS			options;

/********************************************************************
 *																	*
 *		types and constants											*
 *																	*
 ********************************************************************/
 
#define FILE_NAME_LENGTH 20
#define MODE_NAME_LENGTH 5

#define MAX_N_FILES 50

#define READ_FILE	1
#define WRITE_FILE	2

#define MAX_FILE_NAME_LENGTH 128
#define FILE_IO_BUFFER_SIZE (4*1024)
#define FIRST_REAL_FILE 4

typedef struct CFILE_NAME
{
	/* tree allows fast checking of names */
	
	CHAR				*file_name;
	struct CFILE_NAME	*left;
	struct CFILE_NAME	*right;
	WORD				table_offset;
	
} CFILE_NAME;

typedef struct CFILE
{
	/* array allows fast access via file number */
	
	FILE				*file;
	WORD				position;
	WORD				file_length;
	WORD				mode;
	struct CFILE_NAME	*name;
	
} CFILE;

/********************************************************************************
 *																				*
 *		global variables														*
 *																				*
 ********************************************************************************/

static CFILE		file_table [MAX_N_FILES];
static CFILE_NAME	*file_names;
static WORD			number_of_files;

/********************************************************************************
 *																				*
 *		low level file routines													*
 *																				*
 ********************************************************************************/
 
static WORD get_first_file_number (CHAR *name)
{
	CFILE_NAME **name_pp;
	WORD result;
	CFILE_NAME *new_file_name;

	if (name[0]=='/')
	{	fprintf(stderr,"enter_file_name: absolute filenames not allowed\n");
		exit_ioserver();
		return -1;
	}
	
	name_pp = &file_names;
	while (*name_pp!=NULL)
	{	result = strcmp (name, (*name_pp)->file_name);
		if (result==0)
		{	file_table[number_of_files].name=(*name_pp);
			return (*name_pp)->table_offset;
		}
		if (result<0)
			name_pp=&(*name_pp)->left;
		else
			name_pp=&(*name_pp)->right;
	}
	
	new_file_name = (CFILE_NAME *) user_allocate (sizeof (CFILE_NAME));
	
	if (new_file_name == NULL)
	{	fprintf(stderr,"enter_file_name: out of memory\n");
		exit_ioserver();
		return -1;
	}
	
	file_table[number_of_files].name=new_file_name;
	
	new_file_name->left=NULL;
	new_file_name->right=NULL;
	new_file_name->table_offset=number_of_files;
	new_file_name->file_name = (char *) user_allocate (strlen(name) + 1);
	
	if (new_file_name->file_name == NULL)
	{	fprintf(stderr,"enter_file_name: out of memory\n");
		exit_ioserver();
		return -1;
	}
	
	strcpy(new_file_name->file_name, name);
		
	*name_pp=new_file_name;
	return number_of_files;
}

static WORD get_file_length (FILE *file_descriptor)
{
	WORD file_length;
	
	if (fseek (file_descriptor,0l,SEEK_END)!=0)
	{	fclose (file_descriptor);
		fprintf(stderr,"get_file_length: cannot determine end of file\n");
		exit_ioserver();
		return -1;
	}
	file_length=ftell (file_descriptor);
	if (fseek (file_descriptor,0l,SEEK_SET)!=0)
	{	fclose (file_descriptor);
		fprintf(stderr,"get_file_length: cannot determine begin of file\n");
		exit_ioserver();
		return -1;
	}
	return file_length;
}

/********************************************************************************
 *																				*
 *		basic IO operations														*
 *																				*
 ********************************************************************************/

VOID host_write_string (FILE *file,CHAR *string,WORD length)
{
	WORD i;
	for(i=0;i<length;i++)		/* replace by pointers */
		putc(string[i],file);
}

VOID host_write_character(FILE *file,CHAR *character)
{
	putc(*character,file);
}

WORD host_read_string (FILE *file,CHAR *s,WORD length)
{
	register WORD i;
	register WORD c;
		
	if (file == stdin)
	{	fflush(stdout);
		fflush(stderr);
	}
	
	i=0;
	c=getc(file);	
	while (c!='\n' && c!=EOF && i<length)
	{	s[i++]=c;
		c=getc(file);
	}
		
	if (i<length && c=='\n')
		s[i++]=c;
		
	if (i==0)
	{	fprintf(stderr,"host_read_string: read past end of file\n");
		exit_ioserver();
		return -1;
	}
	
	return i;
}

VOID host_write_integer(FILE *file,WORD *integer)
{
	fprintf(file,"%d",*integer);
}

#define INFINITY_L		0
#define INFINITY_H		2146435072

#define UNSTABLE_L		0
#define UNSTABLE_H		2146435073
#define	UNDEFINED_L		0
#define UNDEFINED_H		2146435074

#define	INACCURATE_L	(~0x7fffffff) /* -2147483648 */
#define INACCURATE_H	2146435072

VOID host_write_real (FILE *file,FLOAT *real)
{
	WORD	high;
		
	high = *((int *)real +1);
	
	if (high >= INFINITY_H)
	{	
		if (high == INFINITY_H)
		{
			if (*(int *)real==INACCURATE_L)
				fprintf(file,"NaN: INACCURATE");
			else
				fprintf(file,"NaN: INFINITY");
		}
	
		else if (high == UNDEFINED_H)
			fprintf(file,"NaN: UNDEFINED");
			
		else if (high == UNSTABLE_H)
			fprintf(file,"NaN: UNSTABLE");
		
		else
			fprintf(file,"NaN: UNKNOWN");
			
	} else
		fprintf (file,"%lg", *real);
}

VOID host_write_boolean(FILE *file,BOOL *b)
{
	if (*b)
		host_write_string (file, "TRUE", 4);
	else
		host_write_string (file, "FALSE", 5);
}

static WORD open_file (VCLEAN_STRING *file_name,VCLEAN_STRING *file_mode)
{
	FILE 			*file_descriptor;
	WORD			file_number;
	WORD			first_file_number;
	CFILE			*file_entry;
	WORD			c;
	WORD			mode;
	CHAR			c_file_name[FILE_NAME_LENGTH+1];
	CHAR			c_file_mode[MODE_NAME_LENGTH+1];
	
	if (file_name->length > FILE_NAME_LENGTH)
	{	fprintf(stderr,"open_file: file name too long\n");
		exit_ioserver();
		return -1;
	}
	
	if (file_mode->length > MODE_NAME_LENGTH)
	{	fprintf(stderr,"open_file: mode name too long\n");
		exit_ioserver();
		return -1;
	}
	
	strncpy(c_file_name,file_name->characters,file_name->length);
	strncpy(c_file_mode,file_mode->characters,file_mode->length);
	
	c_file_name[file_name->length]='\0';
	c_file_mode[file_mode->length]='\0';
	
	if (number_of_files >= MAX_N_FILES)
	{	fprintf(stderr,"open_file: too many files\n");
		exit_ioserver();
		return -1;
	}
			
	c = c_file_mode[0] & 0xdf;	/* lowercase to uppercase */
	if (file_mode->length==1 && (c=='R' || c=='W'))
		mode = (c=='W' ? WRITE_FILE : READ_FILE);
	else
	{	fprintf(stderr,"open_file: invalid file mode\n");
		exit_ioserver();
		return -1;
	}
	
	first_file_number = get_first_file_number (c_file_name);
	
	file_number = number_of_files;
	file_entry = &file_table[file_number];
	
	if (file_number != first_file_number)
	{	if (mode==WRITE_FILE)
		{	fprintf(stderr,"open_file: cannot write to existing file\n");
			exit_ioserver();
			return -1;
		}
		else
		{	if (file_table[first_file_number].mode==WRITE_FILE)
			{	fprintf(stderr,"open_file: cannot read from open output file\n");
				exit_ioserver();
				return -1;
			}
		}
	}
	
	file_descriptor = fopen (c_file_name, c_file_mode);
	
	if (file_descriptor==NULL)
	{	if (mode==WRITE_FILE)
		{	fprintf(stderr,"open_file: failed to create file\n");
			exit_ioserver();
			return -1;
		}
		else
		{	fprintf(stderr,"open_file: failed to open input file %s\n",c_file_name);
			exit_ioserver();
			return -1;
		}
	}
	
	setvbuf (file_descriptor,NULL,_IOFBF,FILE_IO_BUFFER_SIZE);
			
	file_entry->file=file_descriptor;
	file_entry->mode=mode;
	file_entry->position=0;
	if (mode==WRITE_FILE)
		file_entry->file_length=0;
	else
		file_entry->file_length=get_file_length(file_descriptor);

	++number_of_files;
	
	return file_number;
}

static VOID reopen_file (WORD file_number)
{
	FILE *file_descriptor;
	CFILE *file_entry;
	
	if (file_number<FIRST_REAL_FILE)
	{	fprintf(stderr,"reopen_file: StdIn, StdOut, StdIO or StdErr can not be reopened\n");
		exit_ioserver();
		return;
	}
	
	if (file_number>=number_of_files)
	{	fprintf(stderr,"reopen_file: invalid file number\n");
		exit_ioserver();
		return;
	}
	
	file_entry=&file_table[file_number];
	if (file_entry->mode==READ_FILE)
	{	fprintf(stderr,"reopen_file: cannot reopen an input file\n");
		exit_ioserver();
		return;
	}
		
	if (fclose (file_entry->file)!=0)
	{	file_entry->file=NULL;
		fprintf(stderr,"reopen_file: failed to close output file\n");
		exit_ioserver();
		return;
	}
	
	file_descriptor=fopen(file_entry->name->file_name,"r");
		
	if (file_descriptor==NULL)
	{	fprintf(stderr,"reopen_file: failed to reopen file for reading\n");
		exit_ioserver();
		return;
	}
		
	setvbuf (file_descriptor,NULL,_IOFBF,FILE_IO_BUFFER_SIZE);
	
	file_entry->file=file_descriptor;
	file_entry->mode=READ_FILE;
	file_entry->position=0;
	file_entry->file_length=get_file_length(file_descriptor);
}

static WORD end_file (WORD file_number, WORD position)
{
	if (file_number<FIRST_REAL_FILE)
	{	switch (file_number)
		{	case StdIn:
			case StdInOut:
				return 0;
			default:
				fprintf(stderr,"end_file: not allowed for StdOut or StdErr\n");
				exit_ioserver();
				return 0;
		}
	} else
	{	CFILE *f;
		
		f=&file_table[file_number];
		if (f->mode==WRITE_FILE)
		{	fprintf (stderr,"end_file: not allowed for output files\n");
			exit_ioserver();
			return 0;
		}
		return f->file_length==position ? 1 : 0;
	}
}

static WORD file_getC (WORD file_number,WORD position)
{
	CFILE *f;
	WORD c;
	
	f=&file_table[file_number];
	
	if (file_number<FIRST_REAL_FILE)
	{	switch (file_number)
		{	case StdOut:
				fprintf(stderr,"file_getC: can not read from StdOut\n");
				exit_ioserver();
				return 0;
			case StdErr:
				fprintf(stderr,"file_getC: can not read from StdErr\n");
				exit_ioserver();
				return 0;
			default:
				if (f->position!=position)
				{	fprintf(stderr,"file_getC: invalid position\n");
					exit_ioserver();
					return 0;
				}
				fflush(stdout);
				c=getc(stdin);
				++f->position;
		}
	} else
	{	if (f->mode==WRITE_FILE)
		{	fprintf(stderr,"file_getC: read from an output file\n");
			exit_ioserver();
			return 0;
		}
		if (f->position!=position)
		{	if (fseek (f->file,(long)position,SEEK_SET)!=0)
			{	fprintf(stderr,"file_getC: seek failed\n");
				exit_ioserver();
				return 0;
			}
			f->position=position;
		}
		c=getc(f->file);
		if (c==EOF)
		{	fprintf(stderr,"file_getC: read past end of file\n");
			exit_ioserver();
			return 0;
		}
		++f->position;
	}
	return c;
}

static WORD file_getS (WORD file_number,WORD position,CHAR *string,WORD length)
{
	CFILE *f;
	
	f=&file_table[file_number];
	
	if (file_number<FIRST_REAL_FILE)
	{	switch (file_number)
		{	case StdOut:
				fprintf(stderr,"file_getS: can not read from StdOut\n");
				exit_ioserver();
				return -1;
			case StdErr:
				fprintf(stderr,"file_getS: can not read from StdErr\n");
				exit_ioserver();
				return -1;
			default:
				if (f->position!=position)
				{	fprintf(stderr,"file_getS: invalid position %d in file: %d\n",position,file_number);
					exit_ioserver();
					return -1;
				}
				length=host_read_string(stdin,string,length);
				f->position+=length;
				return length;
		}
	} else
	{	if (f->mode==WRITE_FILE)
		{	fprintf(stderr,"file_getS: read from an output file\n");
			exit_ioserver();
			return -1;
		}
		if (f->position!=position)
		{	if (fseek (f->file,(long)position,SEEK_SET)!=0)
			{	fprintf(stderr,"file_getS: seek failed\n");
				exit_ioserver();
				return -1;
			}
			f->position=position;
		}
		
		length=host_read_string(f->file,string,length);
		f->position+=length;
		return length;
	}
}

static VOID file_putC (WORD c,WORD file_number,WORD position)
{
	CFILE *f;
	
	f=&file_table[file_number];
	
	if (file_number<FIRST_REAL_FILE)
	{	switch (file_number)
		{	case StdIn:
				fprintf(stderr,"file_putC: can not write to StdIn\n");
				exit_ioserver();
				return;
			case StdErr:
				putc(c,stderr);
				++f->position;
				break;
			default:
			
				/* file position isn't checked for stdout */
			
				putc(c,stdout);
				++f->position;
		}
	} else
	{	if (f->mode==READ_FILE)
		{	fprintf(stderr,"file_putC: write to an input file\n");
			exit_ioserver();
			return;
		}
		if (f->position!=position)
		{	fprintf(stderr,"file_putC: invalid position\n");
			exit_ioserver();
			return;
		}
		putc(c,f->file);
		++f->position;
	}
}

static VOID file_putS (WORD file_number,WORD position,CHAR *string,WORD length)
{
	CFILE *f;	
	
	f=&file_table[file_number];
	
	if (file_number<FIRST_REAL_FILE)
	{	switch (file_number)
		{	case StdIn:
				fprintf(stderr,"file_putS: can not write to StdIn\n");
				exit_ioserver();
				return;
			
			case StdErr:
				host_write_string(stderr,string,length);
				fflush(stderr);
				f->position+=length;
				break;
				
			default:
				/* file position isn't checked for stdout... */
				
				host_write_string(stdout,string,length);				
				fflush(stdout);
				f->position+=length;
		}
	} else
	{	if (f->mode==READ_FILE)
		{	fprintf(stderr,"file_putS: write to an input file\n");
			exit_ioserver();
			return;
		}
		if (f->position!=position)
		{	fprintf(stderr,"file_putS: invalid position\n");
			exit_ioserver();
			return;
		}
		
		host_write_string(f->file,string,length);
		f->position+=length;
	}
}

/********************************************************************
 *																	*
 *		exported packet handling routines							*
 *																	*
 ********************************************************************/

static VOID return_integer(WORD source, VFILE_PACKET *packet,int integer)
{
	VREDUCER_PACKET *return_packet;

	return_packet=alloc_packet(False);
	
	return_packet->reducer_id				=packet->source_id;
	return_packet->local_source				=packet->file_number;
	*((WORD *)return_packet->contents)		=integer;

	send_packet_from(MOutInteger,source,ToHost,
					vsizeof(VREDUCER_PACKET)+sizeof(WORD),return_packet,False);
}

VOID host_open_file(WORD kind,WORD source,WORD size,VFILE_PACKET *packet)
{
	VCLEAN_STRING *file_name;
	VCLEAN_STRING *mode;
		
	mode=(VCLEAN_STRING *)packet->contents;
	file_name=(VCLEAN_STRING *)(packet->contents+((sizeof(WORD)+mode->length+3)&~3));
	
	return_integer(source,packet,open_file(file_name,mode));
}

VOID host_end_file(WORD kind,WORD source,WORD size,VFILE_PACKET *packet)
{
	return_integer(source,packet,end_file(packet->file_number,packet->position));
}

VOID host_output_character(WORD kind,WORD source,WORD size,VFILE_PACKET *packet)
{	
	file_putC(*(WORD *)packet->contents,packet->file_number,packet->position);
}

VOID host_input_character(WORD kind,WORD source,WORD size,VFILE_PACKET *packet)
{	
	return_integer(source,packet,file_getC(packet->file_number,packet->position));
}

VOID host_input_string(WORD kind,WORD source,WORD size,VFILE_PACKET *packet)
{
	VREDUCER_PACKET *return_packet;
	VCLEAN_STRING *string;
	WORD length;
	
	length=*(WORD *)packet->contents;
		
	if (length>StringLength)
		length=StringLength;

	return_packet=alloc_packet(False);
	
	string=(VCLEAN_STRING *)return_packet->contents;	
	length=file_getS(packet->file_number,packet->position,string->characters,length);
	
	return_packet->reducer_id			=packet->source_id;
	return_packet->local_source			=packet->file_number;
	string->length						=length;
	
	send_packet_from(MOutString,source,ToHost,
					vsizeof(VREDUCER_PACKET)+sizeof(WORD)+length,return_packet,False);
}

VOID host_output_string(WORD kind,WORD source,WORD size,VLONG_FILE_PACKET *packet)
{
	CHAR *string;
	WORD length;

	string=packet->contents;
	length=size-vsizeof(VLONG_FILE_PACKET);
	file_putS(packet->file_number,packet->file_position+packet->message_position,string,length);
}

VOID host_output_c_string(WORD kind,WORD source,WORD size,VFILE_PACKET *packet)
{
	VCLEAN_STRING *string;
	
	string=(VCLEAN_STRING *)packet->contents;
	
	switch (packet->file_number)
	{	case StdOut:
		case StdInOut:
			host_write_string(stdout,string->characters,string->length);
			break;
		case StdErr:
			host_write_string(stderr,string->characters,string->length);
			break;
		default:
			fprintf(stderr,"output_c_string: invalid file number\n");
			exit_ioserver();
	}
}

VOID host_output_c_character(WORD kind,WORD source,WORD size,VFILE_PACKET *packet)
{
	switch (packet->file_number)
	{	case StdOut:
		case StdInOut:
			host_write_character(stdout,packet->contents);
			break;
		case StdErr:
			host_write_character(stderr,packet->contents);
			break;
		default:
			fprintf(stderr,"output_c_character: invalid file number\n");
			exit_ioserver();
	}
}

VOID host_output_c_boolean(WORD kind,WORD source,WORD size,VFILE_PACKET *packet)
{
	switch (packet->file_number)
	{	case StdOut:
		case StdInOut:
			host_write_boolean(stdout,(BOOL *)packet->contents);
			break;
		case StdErr:
			host_write_boolean(stderr,(BOOL *)packet->contents);
			break;
		default:
			fprintf(stderr,"output_c_boolean: invalid file number\n");
			exit_ioserver();
	}
}

VOID host_output_c_integer(WORD kind,WORD source,WORD size,VFILE_PACKET *packet)
{
	switch (packet->file_number)
	{	case StdOut:
		case StdInOut:
			host_write_integer(stdout,(WORD *)packet->contents);
			break;
		case StdErr:
			host_write_integer(stderr,(WORD *)packet->contents);
			break;
		default:
			fprintf(stderr,"output_c_integer: invalid file number\n");
			exit_ioserver();
	}
}

VOID host_output_c_real(WORD kind,WORD source,WORD size,VFILE_PACKET *packet)
{
	switch (packet->file_number)
	{	case StdOut:
		case StdInOut:
			host_write_real(stdout,(FLOAT *)packet->contents);
			break;
		case StdErr:
			host_write_real(stderr,(FLOAT *)packet->contents);
			break;
		default:
			fprintf(stderr,"output_c_real: invalid file number\n");
			exit_ioserver();
	}
}

VOID host_input_c_string(WORD kind,WORD source,WORD size,VFILE_PACKET *packet)
{
	VCHANNEL_PACKET *return_packet;
	VCLEAN_STRING *string;
	WORD length;
	
	return_packet=alloc_packet(False);
	
	string=(VCLEAN_STRING *)return_packet->contents;
	switch (packet->file_number)
	{	case StdIn:
		case StdInOut:
			length=host_read_string(stdin,string->characters,StringLength);
			break;
		default:
			fprintf(stderr,"host_read_string: invalid file number\n");
			exit_ioserver();
			return;
	}
	
	return_packet->destination			=(ADDRESS)packet->source_id;
	return_packet->local_source			=packet->file_number;
	return_packet->chan					=(CHANNEL *)packet->position;
	string->length=length;
	
	send_packet_from(MOutCString,source,ToHost,
				vsizeof(VCHANNEL_PACKET)+sizeof(WORD)+length,return_packet,False);
}

VOID host_fatal_error(WORD kind,WORD source,WORD size,VFILE_PACKET *packet)
{
	VCLEAN_STRING *string;
	
	string=(VCLEAN_STRING *)packet->contents;
	
	fflush(stdout);
	fflush(stderr);
	
	fprintf(stderr,"FATAL ERROR at transputer %d: ",source);
	host_write_string(stderr,string->characters,string->length);
	fprintf(stderr,"\n");
	
	exit_ioserver();
}

VOID host_errno(WORD kind,WORD source,WORD size,VFILE_PACKET *packet)
{		
	fflush(stdout);
	fflush(stderr);
	
	fprintf(stderr,"ERROR %d at transputer %d\n",packet->position, source);
	
	exit_ioserver();
}


VOID host_runtime_error(WORD kind,WORD source,WORD size,VFILE_PACKET *packet)
{
	VCLEAN_STRING *string;
	
	string=(VCLEAN_STRING *)packet->contents;
	
	fflush(stdout);
	fflush(stderr);
	
	fprintf(stderr,"RUNTIME ERROR at transputer %d: ",source);
	host_write_string(stderr,string->characters,string->length);
	fprintf(stderr,"\nthe process that caused it has been aborted\n");
	
	/* fprintf(stderr,"proceed? (y/n)\n"); */
}

VOID host_runtime_warning(WORD kind,WORD source,WORD size,VFILE_PACKET *packet)
{
	VCLEAN_STRING *string;
	
	string=(VCLEAN_STRING *)packet->contents;
	
	fflush(stdout);
	fflush(stderr);
	
	fprintf(stderr,"RUNTIME WARNING from transputer %d: ",source);
	host_write_string(stderr,string->characters,string->length);
	fprintf(stderr,"\n");
}

/********************************************************************************
 *																				*
 *		invalid messages for host												*
 *																				*
 ********************************************************************************/

VOID host_handle_reserve(WORD kind,WORD source,WORD size,ALLOCATION_PACKET *packet)
{
	fprintf(stderr,"reserve %d bytes at id %d (allocate reducer: %d)\n",packet->size,packet->id.dest,packet->reduce);
	
	if (kind==MReserveGraph)
		fprintf(stderr,"expect graph ");
	else
		fprintf(stderr,"expect block ");
	
	if (packet->id.source < 0)
		fprintf(stderr,"from transputer %d, system id %d\n",source, -packet->id.source);
	else
		fprintf(stderr,"from transputer %d, process id %d\n",source, packet->id.source);
	
	exit_ioserver();
}

VOID host_default_reception(WORD kind,WORD source,WORD size,VOID *packet)
{
	fprintf(stderr,"WARNING: no server action specified for packet\n");
	fprintf(stderr,"kind: %d\n",kind);
	fprintf(stderr,"size: %d\n",size);	
}

VOID host_close_files(VOID)
{
	WORD n;
	
	/*	close all real files that are found in the array, the dynamically
		allocated memory will be freed implicitly by the loader */
	
	for (n=FIRST_REAL_FILE; n<number_of_files; ++n)
	{
		CFILE *f;
		
		f=&file_table[n];
		
		if (f->file!=NULL && fclose (f->file)!=0)
		{
			if (f->mode==WRITE_FILE)
			{	fprintf(stderr,"Write error\n");
				exit(1);
			}
			else
			{	fprintf(stderr,"Read error\n");
				exit(1);
			}
		}
	}	
}

/********************************************************************************
 *																				*
 *		initialisation															*
 *																				*
 ********************************************************************************/

BOOL init_hostio(VOID)
{
	number_of_files					=FIRST_REAL_FILE;
	file_names						=NULL;
	
	file_table[StdIn].position		=0;
	file_table[StdOut].position		=0;
	file_table[StdInOut].position	=0;
	file_table[StdErr].position		=0;
	
	return True;
}
