#include "instructions.h"

#include "macros.h"
#include "io.h"
#include "system.h"
#include "memory.h"

#ifdef __INCLUDE_STANDARD__

#include <stddef.h>
#include <string.h>

#endif

extern NODE_INFO	node_info;
extern SYS			sys;

/********************************************************************************
 *																				*
 *		globals																	*
 *																				*
 ********************************************************************************/

typedef struct INTERNAL_INSTRUCTION {

	WORD		code;
	
	BOOL		(*step_function) (PROCESSOR_STATE *processor_state, INSTRUCTION *instruction);
	
	OPERATOR	operator;
	WORD		arity;
	
	CHAR		*symbol;

}INTERNAL_INSTRUCTION;

static BOOL step_breaks;

#define NOperateInstr 256		/* for each possible opcode in Oreg */

static INTERNAL_INSTRUCTION	*operate_table[NOperateInstr];

/********************************************************************************
 *																				*
 *		instructions for memory access											*
 *																				*
 ********************************************************************************/

#define invalid_address(a) ((ADDRESS)(a) < node_info.low_valid_address || (ADDRESS)(a) > node_info.high_valid_address-3)
BOOL test_word_address(VOID *address)
{
	if (invalid_address(address))
	{	file_print(StdErr,"bus error: invalid address %08X\n",address);
		return False;
	}
	
	if (((UWORD)address & ByteSelectMask)!=0)
	{	file_print(StdErr,"bus error: unaligned address %08X\n",address);
		return False;
	}
	
	return True;
}

static BOOL get_word (VOID *address, WORD *word)
{
	if (invalid_address(address))
	{	file_print(StdErr,"bus error: load word from invalid address %08X\n",address);
		return False;
	}
	
	if (((UWORD)address & ByteSelectMask)!=0)
	{	file_print(StdErr,"bus error: load word from unaligned address %08X\n",address);
		return False;
	}
	
	*word=*(WORD *)address;
	return True;
}

static BOOL put_word (VOID *address, WORD word)
{
	if (invalid_address(address))
	{	file_print(StdErr,"bus error: store word %08X at invalid address %08X\n",word,address);
		return False;
	}
	
	if (((UWORD)address & ByteSelectMask)!=0)
	{	file_print(StdErr,"bus error: store word %08X at unaligned address %08X\n",word,address);
		return False;
	}
	
	*(WORD *)address=word;
	return True;
}

BOOL test_byte_address(VOID *address)
{
	if (invalid_address(address))
	{	file_print(StdErr,"bus error: invalid byte address %08X\n",address);
		return False;
	}
	
	return True;
}

static BOOL get_byte (VOID *address, BYTE *byte)
{
	if (invalid_address(address))
	{	file_print(StdErr,"bus error: load byte from invalid address %08X\n",address);
		return False;
	}
		
	*byte=*(BYTE *)address;
	return True;
}

static BOOL put_byte (VOID *address, BYTE byte)
{
	if (invalid_address(address))
	{	file_print(StdErr,"bus error: store byte %08X at invalid address %08X\n",(UWORD)byte,address);
		return False;
	}
		
	*(BYTE *)address=byte;
	return True;
}

/********************************************************************************
 *																				*
 *		step functions for undefined or unknown instructions					*
 *																				*
 ********************************************************************************/
 
static BOOL step_unknown(PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	file_print(StdErr,"unknown instruction: opr %08X\n",instr->operand);
	return False;
}

static BOOL step_unimplemented(PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	file_print(StdErr,"not implemented: %s\n",instr->symbol);
	return False;
}

/********************************************************************************
 *																				*
 *		direct step functions													*
 *																				*
 ********************************************************************************/

static BOOL step_j (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{	
	ADDRESS dest;
	
	dest=processor_state->Iptr+instr->size+instr->operand;
	
	if (test_byte_address(dest))
	{	processor_state->Iptr=dest;
		return True;
	}
	return False;
}

static BOOL step_ldlp (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	ADDRESS address;
	
	address=(ADDRESS)(processor_state->Wptr+(instr->operand << ByteSelectLength));
	
	if (test_word_address(address))
	{
		processor_state->Creg=processor_state->Breg;
		processor_state->Breg=processor_state->Areg;
		processor_state->Areg=(WORD)address;
		
		processor_state->Iptr+=instr->size;
		
		return True;
	}
	return False;
}

static BOOL step_ldnl (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	WORD result;
	
	if (get_word((ADDRESS)processor_state->Areg + (instr->operand << ByteSelectLength),&result))	
	{
		processor_state->Areg=result;
		processor_state->Iptr+=instr->size;
		return True;
	}
	return False;
}

static BOOL step_ldc (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{	
	processor_state->Creg=processor_state->Breg;
	processor_state->Breg=processor_state->Areg;
	processor_state->Areg=instr->operand;
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_ldnlp (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	ADDRESS address;
	
	address=(ADDRESS)(processor_state->Areg + (instr->operand << ByteSelectLength));
	
	/* doesn't have to be valid address (yet) */
	
	if (((UWORD)address & ByteSelectMask)==0)
	{
		if(instr->operand==0)
		{	file_print(StdErr,"cannot execute ldnlp 0 (used for marks)\n");
			return False;
		}
	
		processor_state->Areg=(WORD)address;
		processor_state->Iptr+=instr->size;
		return True;
	}
	file_print(StdErr,"ldnlp: Areg & byteselectmask != 0\n");
	return False;
}

static BOOL step_ldl (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	WORD result;
	
	if(get_word(processor_state->Wptr + (instr->operand << ByteSelectLength),&result))
	{
		processor_state->Creg=processor_state->Breg;
		processor_state->Breg=processor_state->Areg;
		processor_state->Areg=result;
		
		processor_state->Iptr+=instr->size;
				
		return True;
	}
	return False;
}

static BOOL step_adc (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Areg+=instr->operand;
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_call (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	file_print(StdErr,"call not yet implemented\n");
	return False;
}

static BOOL step_cj (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	ADDRESS dest;
	
	dest=processor_state->Iptr+instr->size+instr->operand;
	
	if (test_byte_address(dest))
	{	if (processor_state->Areg)
		{	processor_state->Areg=processor_state->Breg;
			processor_state->Breg=processor_state->Creg;
			processor_state->Iptr+=instr->size;
			return True;
		}
		processor_state->Iptr=dest;
		return True;
	}
	return False;
}

static BOOL step_ajw (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	file_print(StdErr,"ajw not yet implemented\n");
	return False;
}

static BOOL step_eqc (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	if (processor_state->Areg==instr->operand)
	{	processor_state->Areg=1;
		processor_state->Iptr+=instr->size;
		return True;
	}
	processor_state->Areg=0;
	processor_state->Iptr+=instr->size;
	return True;
}

static BOOL step_stl (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{	
	if (instr->operand<0)
		file_print(StdErr,"warning: stl %d\n",instr->operand);
	
	if (put_word(processor_state->Wptr+(instr->operand<<ByteSelectLength),processor_state->Areg))
	{	
		processor_state->Areg=processor_state->Breg;
		processor_state->Breg=processor_state->Creg;
	
		processor_state->Iptr+=instr->size;
		return True;
	}
	return False;
}

static BOOL step_stnl (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{	
	if (put_word((ADDRESS)processor_state->Areg+(instr->operand<<ByteSelectLength),processor_state->Breg))
	{		
		processor_state->Breg = processor_state->Creg;
		processor_state->Areg = processor_state->Creg;
		processor_state->Iptr+=instr->size;
		return True;
	}
	return False;
}

static BOOL step_pfix (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	file_print(StdErr,"pfix cannot be stepped explicitly\n");
	return False;
}

static BOOL step_nfix (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	file_print(StdErr,"nfix cannot be stepped explicitly\n");
	return False;
}

static BOOL step_opr (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	INTERNAL_INSTRUCTION *internal_instr;
	
	internal_instr=operate_table[instr->operand];
	
	instr->step_function	=internal_instr->step_function;
	instr->symbol			=internal_instr->symbol;
	
	/* arity, operand, size, and address should have been set already */
	
	return (*instr->step_function)(processor_state,instr);
}

/********************************************************************************
 *																				*
 *		indirect step functions													*
 *																				*
 ********************************************************************************/

static BOOL step_rev (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	WORD intermediate;
	
	intermediate=processor_state->Areg;
	processor_state->Areg=processor_state->Breg;
	processor_state->Breg=intermediate;
	
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_add (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Areg=processor_state->Areg+processor_state->Breg;
	processor_state->Breg=processor_state->Creg;
	
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_sub (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Areg=processor_state->Breg-processor_state->Areg;
	processor_state->Breg=processor_state->Creg;
	
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_mul (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Areg=processor_state->Areg*processor_state->Breg;
	processor_state->Breg=processor_state->Creg;
	
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_div (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Areg=processor_state->Breg/processor_state->Areg;
	processor_state->Breg=processor_state->Creg;
	
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_rem (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Areg=processor_state->Breg%processor_state->Areg;
	processor_state->Breg=processor_state->Creg;
	
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_sum (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Areg=processor_state->Areg+processor_state->Breg;
	processor_state->Breg=processor_state->Creg;
	
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_diff (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Areg=processor_state->Breg-processor_state->Areg;
	processor_state->Breg=processor_state->Creg;
	
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_prod (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Areg=processor_state->Areg*processor_state->Breg;
	processor_state->Breg=processor_state->Creg;
	
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_and (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Areg=processor_state->Areg&processor_state->Breg;
	processor_state->Breg=processor_state->Creg;
	
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_or (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Areg=processor_state->Areg|processor_state->Breg;
	processor_state->Breg=processor_state->Creg;
	
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_xor (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Areg=processor_state->Areg^processor_state->Breg;
	processor_state->Breg=processor_state->Creg;
	
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_not (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Areg=~processor_state->Areg;
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_shl (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	if (processor_state->Areg>=0 && processor_state->Areg<=BPW)
	{	processor_state->Areg=(WORD)((UWORD)processor_state->Breg<<processor_state->Areg);
		processor_state->Breg=processor_state->Creg;
	
		processor_state->Iptr+=instr->size;
	
		return True;
	}
	file_print(StdErr,"invalid shift (shift %d bits left)\n",processor_state->Areg);
	return False;
}

static BOOL step_shr (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	if (processor_state->Areg>=0 && processor_state->Areg<=BPW)
	{	processor_state->Areg=(WORD)((UWORD)processor_state->Breg>>processor_state->Areg);
		processor_state->Breg=processor_state->Creg;
	
		processor_state->Iptr+=instr->size;
	
		return True;
	}
	file_print(StdErr,"invalid shift (shift %d bits right)\n",processor_state->Areg);
	return False;
}

static BOOL step_gt (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	if (processor_state->Breg>processor_state->Areg)
	{	processor_state->Areg=1;
		processor_state->Breg=processor_state->Creg;
		processor_state->Iptr+=instr->size;
		return True;
	} else
	{	processor_state->Areg=0;
		processor_state->Breg=processor_state->Creg;
		processor_state->Iptr+=instr->size;
		return True;
	}
}

static BOOL step_bcnt (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Areg=(WORD)((UWORD)processor_state->Areg<<ByteSelectLength);
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_wcnt (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Creg=processor_state->Breg;
	processor_state->Breg=(processor_state->Areg&ByteSelectMask);
	processor_state->Areg=processor_state->Areg>>ByteSelectLength;
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_ldpi (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	ADDRESS address;
	
	address=processor_state->Iptr+instr->size+processor_state->Areg;

	if (test_byte_address(address))
	{	processor_state->Areg=(WORD)address;
		processor_state->Iptr+=instr->size;
		return True;
	}
	return False;
}

static BOOL step_mint (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Creg=processor_state->Breg;
	processor_state->Breg=processor_state->Areg;
	processor_state->Areg=(WORD)mint();
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_bsub (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Areg=processor_state->Areg+processor_state->Breg;
	processor_state->Breg=processor_state->Creg;
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_wsub (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Areg=processor_state->Areg+(processor_state->Breg<<ByteSelectLength);
	processor_state->Breg=processor_state->Creg;
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_move (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	if (!test_byte_address((ADDRESS)processor_state->Creg))
	{	file_print(StdErr,"move: invalid source address\n");
		return False;
	}
	if (!test_byte_address((ADDRESS)processor_state->Breg))
	{	file_print(StdErr,"move: invalid destination address\n");
		return False;
	}
	if(processor_state->Areg<0)
	{	file_print(StdErr,"move: negative size\n");
		return False;
	}
	
	/* test for overlapping blocks */
	
	if ((processor_state->Creg<=processor_state->Breg && processor_state->Creg+processor_state->Areg>processor_state->Breg) ||
		(processor_state->Breg<=processor_state->Creg && processor_state->Breg+processor_state->Areg>processor_state->Creg))
	{
		file_print(StdErr,"move: overlapping blocks\n");
		return False;
	}
	
	if (processor_state->Areg>1024)
		file_print(StdErr,"warning: move of many (%d) bytes\n",processor_state->Areg);
	
	move((ADDRESS)processor_state->Breg,(ADDRESS)processor_state->Creg,processor_state->Areg);

	processor_state->Iptr+=instr->size;
	return True;
}

static BOOL step_lb (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	BYTE byte;
	
	if (get_byte((ADDRESS)processor_state->Areg,&byte))
	{
		*(UWORD *)&processor_state->Areg=(UBYTE)byte;
		processor_state->Iptr+=instr->size;
		return True;
	}
	return False;
}

static BOOL step_sb (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	BYTE byte;
	
	byte=(BYTE)(processor_state->Breg);
	
	if (put_byte((ADDRESS)processor_state->Areg,byte))
	{
		processor_state->Areg = processor_state->Creg;
		processor_state->Breg = processor_state->Creg;
		processor_state->Iptr+=instr->size;
	
		return True;
	}
	return False;
}

static BOOL step_gcall (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	ADDRESS return_address;
	
	return_address=processor_state->Iptr+instr->size;

	if(test_byte_address((ADDRESS)processor_state->Areg))
	{	processor_state->Iptr=(ADDRESS)processor_state->Areg;
		processor_state->Areg=(WORD)return_address;
		return True;
	}
	return False;
}

static BOOL step_gajw (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	ADDRESS new_workspace;
	
	new_workspace=(ADDRESS)processor_state->Areg;
	if(test_word_address(new_workspace))
	{
		processor_state->Areg=(WORD)processor_state->Wptr;
		processor_state->Wptr=new_workspace;
		processor_state->Iptr+=instr->size;
		return True;
	}
	return False;
}

static BOOL step_ret (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	file_print(StdErr,"ret not yet implemented\n");
	return False;
}

static BOOL step_dup (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Creg = processor_state->Breg;
	processor_state->Breg = processor_state->Areg;
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_xdble (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Creg=processor_state->Breg;
	if(processor_state->Areg<0)
		processor_state->Breg=-1;
	else
		processor_state->Breg=0;
	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_lshl (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	WORD nbits;
	
	nbits=processor_state->Areg;
	
	if (nbits>=0 && nbits <= 2*BPW)
	{
		if (nbits>=BPW)
		{	processor_state->Areg=0;
			processor_state->Breg=processor_state->Breg<<(nbits-BPW);
			processor_state->Iptr+=instr->size;

			return True;
		}
		processor_state->Areg=processor_state->Breg<<nbits;
		processor_state->Breg=(WORD)((UWORD)processor_state->Breg>>(BPW-nbits));
		processor_state->Breg|=(processor_state->Creg<<nbits);
		processor_state->Iptr+=instr->size;
		
		return True;
	}
	file_print(StdErr,"invalid long shift (shift %d bits left)\n",nbits);
	return False;
}

static BOOL step_lshr (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	WORD nbits;
	
	nbits=processor_state->Areg;
	
	if (nbits>=0 && nbits <= 2*BPW)
	{
		if (nbits>=BPW)
		{	processor_state->Breg=0;
			processor_state->Areg=(WORD)((UWORD)processor_state->Creg>>(nbits-BPW));
			processor_state->Iptr+=instr->size;

			return True;
		}
		processor_state->Areg=(WORD)((UWORD)processor_state->Breg>>nbits);
		processor_state->Areg|=(processor_state->Creg<<(BPW-nbits));
		processor_state->Breg=(WORD)((UWORD)processor_state->Creg>>nbits);
		processor_state->Iptr+=instr->size;
		
		return True;
	}
	file_print(StdErr,"invalid long shift (shift %d bits right)\n",nbits);
	return False;
}

static BOOL step_ldpri (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Creg=processor_state->Breg;
	processor_state->Breg=processor_state->Areg;
	processor_state->Areg=processor_state->prio;
	
	processor_state->Iptr+=instr->size;

	return True;
}

/********************************************************************************
 *																				*
 *		floating point routines													*
 *																				*
 ********************************************************************************/

static BOOL step_fpldnldb(PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	FLOAT *float64_ptr;
	
	float64_ptr=(FLOAT *)processor_state->Areg;
	
	if (test_word_address((ADDRESS)float64_ptr))
	{
		processor_state->Areg=processor_state->Breg;
		processor_state->Breg=processor_state->Creg;
		
		processor_state->fCreg=processor_state->fBreg;
		processor_state->fBreg=processor_state->fAreg;
		processor_state->fAreg=*float64_ptr;
		
		processor_state->round_mode=ToNearest;
		
		processor_state->Iptr+=instr->size;

		return True;
	}
	return False;
}

static BOOL step_fpstnldb(PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	FLOAT *float64_ptr;
	
	float64_ptr=(FLOAT *)processor_state->Areg;
	
	if (test_word_address((ADDRESS)float64_ptr))
	{
		processor_state->Areg=processor_state->Breg;
		processor_state->Breg=processor_state->Creg;
		
		*float64_ptr=processor_state->fAreg;
		processor_state->fAreg=processor_state->fBreg;
		processor_state->fBreg=processor_state->fCreg;
		
		processor_state->round_mode=ToNearest;
		
		processor_state->Iptr+=instr->size;

		return True;
	}
	return False;
}

static BOOL step_fpi32tor64(PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	WORD *int32_ptr;
	
	int32_ptr=(WORD *)processor_state->Areg;
	
	if(test_word_address((ADDRESS)int32_ptr))
	{
		processor_state->Areg=processor_state->Breg;
		processor_state->Breg=processor_state->Creg;
		
		processor_state->fCreg=processor_state->fBreg;
		processor_state->fBreg=processor_state->fAreg;
		processor_state->fAreg=(FLOAT)(*int32_ptr);
				
		processor_state->round_mode=ToNearest;
		
		processor_state->Iptr+=instr->size;

		return True;
	}
	return False;
}

static BOOL step_fpadd (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->fAreg=processor_state->fAreg+processor_state->fBreg;
	processor_state->fBreg=processor_state->fCreg;
	
	processor_state->round_mode=ToNearest;

	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_fpsub (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->fAreg=processor_state->fBreg-processor_state->fAreg;
	processor_state->fBreg=processor_state->fCreg;
	
	processor_state->round_mode=ToNearest;

	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_fpmul (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->fAreg=processor_state->fAreg*processor_state->fBreg;
	processor_state->fBreg=processor_state->fCreg;
	
	processor_state->round_mode=ToNearest;

	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_fpdiv (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->fAreg=processor_state->fBreg/processor_state->fAreg;
	processor_state->fBreg=processor_state->fCreg;
	
	processor_state->round_mode=ToNearest;

	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_fpdup (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->fCreg=processor_state->fBreg;
	processor_state->fBreg=processor_state->fAreg;
	
	processor_state->round_mode=ToNearest;

	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_fprev (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	FLOAT intermediate;
	
	intermediate=processor_state->fAreg;
	processor_state->fAreg=processor_state->fBreg;
	processor_state->fBreg=intermediate;
	
	processor_state->round_mode=ToNearest;

	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_fpldzerodb(PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->fCreg=processor_state->fBreg;
	processor_state->fBreg=processor_state->fAreg;
	processor_state->fAreg=0.0;
	
	processor_state->round_mode=ToNearest;

	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_fpgt(PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Creg=processor_state->Breg;
	processor_state->Breg=processor_state->Areg;
	
	if (processor_state->fBreg > processor_state->fAreg)
		processor_state->Areg=1;
	else
		processor_state->Areg=0;
	
	processor_state->fAreg=processor_state->fCreg;
	processor_state->fBreg=processor_state->fCreg;
	
	processor_state->round_mode=ToNearest;

	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_fpeq(PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	processor_state->Creg=processor_state->Breg;
	processor_state->Breg=processor_state->Areg;
	
	if (processor_state->fBreg == processor_state->fAreg)
		processor_state->Areg=1;
	else
		processor_state->Areg=0;
	
	processor_state->fAreg=processor_state->fCreg;
	processor_state->fBreg=processor_state->fCreg;
	
	processor_state->round_mode=ToNearest;

	processor_state->Iptr+=instr->size;
	
	return True;
}

static BOOL step_fpint(PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	switch(processor_state->round_mode)
	{
		case ToNearest:
			fpint_nearest(&processor_state->fAreg);
			break;
		case ToZero:
			fpint_zero(&processor_state->fAreg);
			break;
		case ToPlusInfinity:
			fpint_plus_inf(&processor_state->fAreg);
			break;
		case ToMinusInfinity:
			fpint_min_inf(&processor_state->fAreg);
			break;
		default:
			file_print(StdErr,"(internal error) unknown round mode\n");
			return False;
	}
	processor_state->round_mode=ToNearest;
	processor_state->Iptr+=instr->size;

	return True;
}

static BOOL step_fpentry (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	switch(processor_state->Areg)
	{
		case 0x0B:
			fpuabs(&processor_state->fAreg);
			processor_state->round_mode=ToNearest;
			break;

		case 0x12:
			processor_state->fAreg=processor_state->fAreg*2.0;
			processor_state->round_mode=ToNearest;
			break;
		case 0x11:
			processor_state->fAreg=processor_state->fAreg/2.0;
			processor_state->round_mode=ToNearest;
			break;
		
		case 0x22:
			processor_state->round_mode=ToNearest;
			break;
		case 0x06:
			processor_state->round_mode=ToZero;
			break;
		case 0x04:
			processor_state->round_mode=ToPlusInfinity;
			break;
		case 0x05:
			processor_state->round_mode=ToMinusInfinity;
			break;
		
		default:
			file_print(StdErr,"unknown fpentry %08X\n",processor_state->Areg);
			return False;
	}
	processor_state->Areg=processor_state->Breg;
	processor_state->Breg=processor_state->Creg;
	
	processor_state->Iptr+=instr->size;

	return True;
}

/********************************************************************************
 *																				*
 *		timer instructions														*
 *																				*
 ********************************************************************************/

static BOOL step_ldtimer (PROCESSOR_STATE *processor_state,INSTRUCTION *instr)
{
	/* loads time that debugger maintains during stepping */
	
	processor_state->Creg=processor_state->Breg;
	processor_state->Breg=processor_state->Areg;
	processor_state->Areg=processor_state->time;
	processor_state->Iptr+=instr->size;
	return True;
}

/********************************************************************************
 *																				*
 *		'parallel' instructions													*
 *																				*
 ********************************************************************************/

/********************************************************************************
 *																				*
 *		INSTRUCTION tables														*
 *																				*
 ********************************************************************************/

static INTERNAL_INSTRUCTION break_instruction={0x00,step_j,Instr_jump,0,"break"};

static INTERNAL_INSTRUCTION	instruction_table[]=
{
	{0x00,		step_j,				Instr_jump,			1,	"j       "},
	{0x01,		step_ldlp,			Instr_ldlp,			1,	"ldlp    "},
	{0x02,		step_pfix,			Instr_pfix,			1,	"pfix    "},
	{0x03,		step_ldnl,			Instr_ldnl,			1,	"ldnl    "},
	{0x04,		step_ldc,			Instr_ldc,			1,	"ldc     "},
	{0x05,		step_ldnlp,			Instr_ldnlp,		1,	"ldnlp   "},
	{0x06,		step_nfix,			Instr_nfix,			1,	"nfix    "},
	{0x07,		step_ldl,			Instr_ldl,			1,	"ldl     "},
	{0x08,		step_adc,			Instr_adc,			1,	"adc     "},
	{0x09,		step_call,			Instr_call,			1,	"call    "},
	{0x0A,		step_cj,			Instr_cjump,		1,	"cj      "},
	{0x0B,		step_ajw,			Instr_ajw,			1,	"ajw     "},
	{0x0C,		step_eqc,			Instr_eqc,			1,	"eqc     "},
	{0x0D,		step_stl,			Instr_stl,			1,	"stl     "},
	{0x0E,		step_stnl,			Instr_stnl,			1,	"stnl    "},
	{0x0F,		step_opr,			Instr_opr,			1,	"opr     "}
};

static INTERNAL_INSTRUCTION operation_table[]=
{
	{0x00,		step_unknown,		Instr_unknown,		0,	"unknown"},
	
	{0x00,		step_rev,			Instr_rev,			0,	"rev"},
	{0x05,		step_add,			Instr_add,			0,	"add"},
	{0x0C,		step_sub,			Instr_sub,			0,	"sub"},
	{0x53,		step_mul,			Instr_mul,			0,	"mul"},
	{0x2C,		step_div,			Instr_div,			0,	"div"},
	{0x1F,		step_rem,			Instr_rem,			0,	"rem"},
	{0x52,		step_sum,			Instr_sum,			0,	"sum"},
	{0x04,		step_diff,			Instr_diff,			0,	"diff"},
	{0x08,		step_prod,			Instr_prod,			0,	"prod"},
	{0x46,		step_and,			Instr_and,			0,	"and"},
	{0x4B,		step_or,			Instr_or,			0,	"or"},
	{0x33,		step_xor,			Instr_xor,			0,	"xor"},
	{0x32,		step_not,			Instr_not,			0,	"not"},
	{0x41,		step_shl,			Instr_shl,			0,	"shl"},
	{0x40,		step_shr,			Instr_shr,			0,	"shr"},
	{0x09,		step_gt,			Instr_gt,			0,	"gt"},
	{0x34,		step_bcnt,			Instr_bcnt,			0,	"bcnt"},
	{0x3F,		step_wcnt,			Instr_wcnt,			0,	"wcnt"},
	{0x1B,		step_ldpi,			Instr_ldpi,			0,	"ldpi"},
	{0x42,		step_mint,			Instr_mint,			0,	"mint"},
	{0x02,		step_bsub,			Instr_bsub,			0,	"bsub"},
	{0x0A,		step_wsub,			Instr_wsub,			0,	"wsub"},
	{0x4A,		step_move,			Instr_move,			0,	"move"},
	{0x07,		step_unimplemented,	Instr_in,			0,	"in"},
	{0x0B,		step_unimplemented,	Instr_out,			0,	"out"},
	{0x01,		step_lb,			Instr_lb,			0,	"lb"},
	{0x3B,		step_sb,			Instr_sb,			0,	"sb"},
	{0x06,		step_gcall,			Instr_gcall,		0,	"gcall"},
	{0x3C,		step_gajw,			Instr_gajw,			0,	"gajw"},
	{0x21,		step_unimplemented,	Instr_lend,			0,	"lend"},
	{0x20,		step_ret,			Instr_ret,			0,	"ret"},
	{0x39,		step_unimplemented,	Instr_runp,			0,	"runp"},
	{0x15,		step_unimplemented,	Instr_stopp,		0,	"stopp"},
	{0x22,		step_ldtimer,		Instr_ldtimer,		0,	"ldtimer"},
	{0x54,		step_unimplemented,	Instr_sttimer,		0,	"sttimer"},
	{0x5A,		step_dup,			Instr_dup,			0,	"dup"},
	{0x1D,		step_xdble,			Instr_xdble,		0,	"xdble"},
	{0x3A,		step_unimplemented,	Instr_xword,		0,	"xword"},
	{0x1A,		step_unimplemented,	Instr_ldiv,			0,	"ldiv"},
	{0x35,		step_lshr,			Instr_lshr,			0,	"lshr"},
	{0x36,		step_lshl,			Instr_lshl,			0,	"lshl"},
	{0x1E,		step_ldpri,			Instr_ldpri,		0,	"ldpri"},
	{0x12,		step_unimplemented,	Instr_resetch,		0,	"resetch"},
	{0x29,		step_unimplemented, Instr_testerr,		0,	"testerr"},
	{0x9C,		step_unimplemented, Instr_fptesterr,	0,	"fptesterr"},
	{0x0F,		step_unimplemented, Instr_outword,		0,	"outword"},
	{0x0E,		step_unimplemented, Instr_outbyte,		0,	"outbyte"},
	{0x1C,		step_unimplemented, Instr_stlf,			0,	"stlf"},
	{0x18,		step_unimplemented, Instr_sthf,			0,	"sthf"},
	{0x50,		step_unimplemented, Instr_sthb,			0,	"sthb"},
	{0x17,		step_unimplemented, Instr_stlb,			0,	"stlb"},
	{0x58,		step_unimplemented, Instr_sethalterr,	0,	"sethalterr"},

	/* floating point operations */

	{0x8A,		step_fpldnldb,		Instr_fpldnldb,		0,	"fpldnldb"},
	{0x84,		step_fpstnldb,		Instr_fpstnldb,		0,	"fpstnldb"},
	{0x98,		step_fpi32tor64,	Instr_fpi32tor64,	0,	"fpi32tor64"},
	{0x87,		step_fpadd,			Instr_fpadd,		0,	"fpadd"},
	{0x89,		step_fpsub,			Instr_fpsub,		0,	"fpsub"},
	{0x8B,		step_fpmul,			Instr_fpmul,		0,	"fpmul"},
	{0x8C,		step_fpdiv,			Instr_fpdiv,		0,	"fpdiv"},
	{0xA3,		step_fpdup,			Instr_fpdup,		0,	"fpdup"},
	{0x94,		step_fpgt,			Instr_fpgt,			0,	"fpgt"},
	{0xA0,		step_fpldzerodb,	Instr_fpldzerodb,	0,	"fpldzerodb"},
	{0xA1,		step_fpint,			Instr_fpint,		0,	"fpint"},
	{0x95,		step_fpeq,			Instr_fpeq,			0,	"fpeq"},
	{0xA4,		step_fprev,			Instr_fprev,		0,	"fprev"},
	{0xAB,		step_fpentry,		Instr_fpentry,		0,	"fpentry"}
};

static VOID initialise_operate_table(VOID)
{
	WORD i;
	INTERNAL_INSTRUCTION *instr;
	
	/* initialise unknown operations */
	
	for (i=0; i<NOperateInstr; ++i)
		operate_table[i]=&operation_table[0];
	
	/* initialise known operations */
	
	for (i=1,instr=&operation_table[1]; i<sizeof(operation_table)/sizeof(INTERNAL_INSTRUCTION); ++i,++instr)
		operate_table[instr->code]=instr;
}


BOOL initialise_instructions(VOID)
{
	step_breaks=True;

	initialise_operate_table();
	
	return True;
}

VOID set_round_mode(ROUND_MODE mode)
{
	sys.processor_state.round_mode=mode;
}

/********************************************************************************
 *																				*
 *		low level calculations on addresses	(data areas are not considered)		*
 *																				*
 ********************************************************************************/

/* valid instruction addresses point to 'the' first prefix instruction of an
   instruction (or the instruction itself if it hasn't got any prefices). If
   an instruction starts with some pfix 0 instructions, any pointer to these
   instructions, or the following one are valid instruction addresses for this
   instruction. All instructions except 'get_instr_address' and 'show_...')
   expect a valid instruction address.
*/

ADDRESS get_instr_address(ADDRESS address)
{
	/* returns a correct start address of instruction 'at' this address. Note that this
	   routine returns the first leading nop (if any) */
	
	UWORD icode;
	ADDRESS iptr;
	
	iptr=address-1;						/* prev instruction or prefix */
	icode=((UWORD)(*iptr))>>4;
		
	while (icode==2 || icode==6)
		icode=((UWORD)(*(--iptr)))>>4;
	
	/* code of prev instruction found */
		
	return iptr+1;
}

ADDRESS get_next_instr_address(ADDRESS address)
{
	/* returns correct start address of instruction at 'next' address */
	
	UWORD icode;
	ADDRESS iptr;
	
	iptr=address;
	icode=((UWORD)(*iptr))>>4;
	
	while (icode==2 || icode==6)
		icode=((UWORD)(*(++iptr)))>>4;
	return iptr+1;
}

ADDRESS get_prev_instr_address(ADDRESS address)
{
	/* returns correct start address of instruction at 'previous' address */
	
	UWORD icode;
	ADDRESS iptr;
	
	iptr=address-1;						/* prev instruction or prefix */
	icode=((UWORD)(*iptr))>>4;
	
	while (icode==2 || icode==6)
		icode=((UWORD)(*(--iptr)))>>4;
	
	/* code of prev instruction found, find first prefix */
	
	icode=((UWORD)(*(--iptr)))>>4;
	while (icode==2 || icode==6)
		icode=((UWORD)(*(--iptr)))>>4;
	
	return iptr+1;
}

ADDRESS skip_nops(ADDRESS address)
{
	/* skips leading nops of instruction address, address must be a valid start address */
	
	UWORD instruction;
	UWORD icode;
	WORD operand;
	ADDRESS iptr;
	
	iptr=address;
	
	instruction		=(UWORD)(*iptr);
	icode			=instruction >> 4;
	operand			=instruction & 0xF;
	
	while (icode==2 && operand == 0)
	{
		instruction		=(UWORD)(*(++iptr));
		icode			=instruction >> 4;
		operand			=instruction & 0xF;
	}
	return iptr;
}

/************************************************************************************
 *																					*
 *		low level operations on instructions (data not considered)					*
 *																					*
 ************************************************************************************/

VOID get_instr(ADDRESS address, INSTRUCTION *instruction)

	/* assumes the address is a valid instruction address */
{
	UWORD		ibyte;
	UWORD		icode;
	WORD		operand;
	ADDRESS		iptr;
	WORD		size;
	
	INTERNAL_INSTRUCTION *internal_instr;
	
	size		=1;
	iptr		=address;
	ibyte		=(UWORD)(*iptr);
	
	if (ibyte != 0)
	{
		/* this is not a break */
		
		icode		=ibyte >> 4;		/* first four bits */
		operand		=ibyte & 0xF;		/* last four bits */
		
		/* read prefixes */
		
		for(;;)
		{	if(icode==2)
			{	++size;
				ibyte		=(UWORD)(*(++iptr));
				icode		=ibyte>>4;
				operand		=(operand<<4)|(ibyte & 0xF);
				continue;
			}
			if(icode==6)
			{	++size;
				ibyte		=(UWORD)(*(++iptr));
				icode		=ibyte>>4;
				operand		=((~operand)<<4)|(ibyte & 0xF);
				continue;
			}
			break;
		}
		
		if (icode==15)		
			internal_instr=operate_table[operand];
		else
			internal_instr=&instruction_table[icode];
	} else
	{
		/* jump 0, this is a break, act as jump, print break */
		
		operand=0;
		internal_instr=&break_instruction;
	}
	
	instruction->step_function	=internal_instr->step_function;
	instruction->operator		=internal_instr->operator;
	instruction->arity			=internal_instr->arity;
	instruction->symbol			=internal_instr->symbol;
	
	instruction->operand		=operand;
	instruction->size			=size;
	instruction->address		=address;
}

VOID get_next_instr(INSTRUCTION *instruction, INSTRUCTION *next_instruction)
{
	/* returns fills next instruction with instruction after 'instruction', source
	   and dest may be the same */
	
	get_instr(instruction->address+instruction->size,next_instruction);
}

CHAR *get_instr_symbol(INSTRUCTION *instruction)
{
	return instruction->symbol;
}

BOOL step(INSTRUCTION *instruction)
{	
	/* this routine steps one instruction according to the processor state. The
	   stepped instruction is return in instruction */
		
	get_instr(sys.processor_state.Iptr,instruction);
	return(*instruction->step_function)(&sys.processor_state,instruction);
}

/****************************************************************************
 *																			*
 *		low level instructions on marks										*
 *																			*
 ****************************************************************************/

ADDRESS next_mark(ADDRESS address, WORD max)
{
	/* finds the next mark. address must be a valid instruction address
	   note: if the current address is marked, this address is returned.
	   The exact address of the mark is returned (ignoring leading nops) */
	
	UWORD instruction;
	UWORD icode;
	WORD operand;

	WORD n;
	
	address=skip_nops(address);
	
	for(n=max;n>0;--n)
	{
		instruction	=(UWORD)(*address);
		icode		=instruction>>4;
		operand		=instruction & 0xF;

		if (icode == 5 && operand == 0)
			return address;
		
		address=get_next_instr_address(address);
		address=skip_nops(address);
	}

	return NULL;
}

ADDRESS prev_mark(ADDRESS address, WORD max)
{
	/* finds previous mark, address must be a valid instruction address
	   note: the search starts BEFORE the current address */
	
	UWORD instruction;
	UWORD icode;
	WORD operand;

	WORD n;
	
	address=get_prev_instr_address(address);
	address=skip_nops(address);
	
	for(n=max;n>0;--n)
	{
		instruction	=(UWORD)(*address);
		icode		=instruction>>4;
		operand		=instruction & 0xF;

		if (icode == 5 && operand == 0)
			return address;
		
		address=get_prev_instr_address(address);
		address=skip_nops(address);
	}

	return NULL;
}

/********************************************************************************
 *																				*
 *		instructions to extract data from data area								*
 *																				*
 ********************************************************************************/

ADDRESS get_data_area(ADDRESS address)
{
	/* returns the data area after the current sequence of instructions (if any)
	   the address must be a valid instruction address */
	
	address=next_mark(address,10000);
	
	if (!address)
		return NULL;
	
	return get_next_instr_address(address);
}

ADDRESS get_symbols(ADDRESS data_area)
{
	/* returns a pointer to the first symbol in the data area (if any) */
	
	INSTRUCTION instruction;
	
	get_instr(data_area,&instruction);
	if (instruction.operand<0)
		return data_area;
	return NULL;
}

ADDRESS get_symbols_and_data(ADDRESS data_area, ADDRESS *symbols, ADDRESS *data, WORD *data_size)
{
	/* returns pointers to symbols and data in the data area, NULL if these are
	   not available. Additionally the size of the data is delivered. The address
	   that is returned is the first address after the data area */
	
	INSTRUCTION instruction;
	ADDRESS		symbol;
	WORD		sym_def_offset;
	
	symbol			=data_area;

	get_instr(symbol,&instruction);
	sym_def_offset	=instruction.operand;
		
	if (sym_def_offset<0)
		*symbols=symbol;
	else
		*symbols=NULL;
	
	while (sym_def_offset<0)				/* symbol found, skip */
	{
		/* symbol found, skip it */
		
		symbol+=instruction.size;
		while (*symbol)
			++symbol;
		++symbol;
		
		get_instr(symbol,&instruction);
		sym_def_offset=instruction.operand;		
	}
	/* last symbol found, the operand of the last instruction contains the size
	   if the data area, which starts immediately after this instruction */
	
	symbol+=instruction.size;
	
	if (sym_def_offset>0)
		*data=symbol;
	else
		*data=NULL;
		
	*data_size=sym_def_offset;
	
	/* return address after mark */
	
	return symbol+sym_def_offset+1;
}

ADDRESS get_next_symbol(ADDRESS symbol)
{
	/* delivers the next symbol, NULL if none available */
	
	INSTRUCTION instruction;
	
	if (symbol==NULL)
		return NULL;
	
	get_instr(symbol,&instruction);
	
	symbol+=instruction.size;
	while(*symbol)
		++symbol;
	++symbol;
	
	get_instr(symbol,&instruction);
	if(instruction.operand>=0)
		return NULL;
	return symbol;
}

ADDRESS get_symbol_definition(ADDRESS data_area,ADDRESS symbol, CHAR **string)

/* delivers the definition address and the string of the parameter symbol */
{
	INSTRUCTION instruction;
	
	get_instr(symbol,&instruction);
		
	*string=(CHAR *)(symbol+instruction.size);
	
	return data_area+instruction.operand;
}

ADDRESS get_applied_symbol(ADDRESS address, CHAR **string)
{
	/* returns the symbol that is defined at the given address. If no symbol can be
	   found for this address, NULL is delivered */
	
	ADDRESS data_area;
	ADDRESS symbol;
	ADDRESS definition;
	
	if (!(data_area=get_data_area(address)))
		return NULL;
	
	if (!(symbol=get_symbols(data_area)))
		return NULL;
	
	definition=get_symbol_definition(data_area,symbol,string);
	while(definition!=address)
	{	if (!(symbol=get_next_symbol(symbol)))
			return NULL;
		definition=get_symbol_definition(data_area,symbol,string);
	}
	
	/* found a matching definition */
	
	return symbol;	
}

ADDRESS get_data_area_end(ADDRESS data_area)
{
	ADDRESS symbols;
	ADDRESS data;
	WORD data_size;
	
	return get_symbols_and_data(data_area,&symbols,&data,&data_size);	
}

/********************************************************************************
 *																				*
 *		'high level' address calculations, use these in preference to low		*
 *		level ones																*
 *																				*
 ********************************************************************************/

static ADDRESS get_Poffset_instr_address(ADDRESS address, WORD offset)
{
	UWORD instruction;
	UWORD icode;
	WORD operand;
	
	/* if the current address points to a mark, skip data area */
	
	address=skip_nops(address);
	
	instruction	=(UWORD)(*address);
	icode		=instruction>>4;
	operand		=instruction & 0xF;

	while(icode == 5 && operand == 0)
	{
		/* this is a mark, set address to first address after data area */
		
		address=get_data_area(address);
		address=get_data_area_end(address);
		address=skip_nops(address);
	
		instruction	=(UWORD)(*address);
		icode		=instruction>>4;
		operand		=instruction & 0xF;
	}

	while (offset>0)
	{
		address=get_next_instr_address(address);
		address=skip_nops(address);
		
		instruction	=(UWORD)(*address);
		icode		=instruction>>4;
		operand		=instruction & 0xF;

		while(icode == 5 && operand == 0)
		{
			/* this is a mark, set address to first address after data area */

			address=get_data_area(address);
			address=get_data_area_end(address);
			address=skip_nops(address);
	
			instruction	=(UWORD)(*address);
			icode		=instruction>>4;
			operand		=instruction & 0xF;
		}
		--offset;
	}
	return address;
}

static ADDRESS get_Noffset_instr_address(ADDRESS address, WORD offset)

/* this routine cannot look past symbol tables yet */
{
	UWORD instruction;
	UWORD icode;
	WORD operand;

	/* get previous instruction */
	
	address=get_prev_instr_address(address);
	address=skip_nops(address);
	
	instruction	=(UWORD)(*address);
	icode		=instruction>>4;
	operand		=instruction & 0xF;

	if (icode == 5 && operand == 0)
	{
		/* this is a mark, return next instruction (the original one) */
		
		return get_next_instr_address(address);
	}

	--offset;
	while (offset>0)	
	{
		address=get_prev_instr_address(address);
		address=skip_nops(address);
		
		instruction	=(UWORD)(*address);
		icode		=instruction>>4;
		operand		=instruction & 0xF;

		if (icode == 5 && operand == 0)
		{
			/* this is a mark, return next instruction */

			return get_next_instr_address(address);
		}
		--offset;
	}
	return address;
}

ADDRESS get_offset_instr_address(ADDRESS address, WORD offset)
{
	/* the address must be a valid instruction address. The address that is returned
	   id the address that is 'offset' instructions past the parameter address.
	   data areas are skipped */
		
	if (offset>=0)
		return get_Poffset_instr_address(address,offset);
	
	return get_Noffset_instr_address(address,-offset);
}

/********************************************************************************
 *																				*
 *		routines to show instructions											*
 *																				*
 ********************************************************************************/

VOID show_basic_instruction(WORD file_nr, INSTRUCTION *instruction)
{
	/* shows an instruction that has been obtained by get_instruction or a related
	   routine. Symbols aren't printed. This routine is supposed to be 'fast' and
	   simple (e.g. used during stepping) */

	if (instruction->arity==1)
		file_print(file_nr,"%08X %s %-10d\n",instruction->address,instruction->symbol,instruction->operand);
	else
		file_print(file_nr,"%08X %s\n",instruction->address,instruction->symbol);
}

ADDRESS show_instruction(WORD file_nr, ADDRESS iptr, ADDRESS address, ADDRESS data_area, ADDRESS *definition_symbol)
{
	/* address can be any address 'within' instruction sequence, the result is the
	   address of the next instruction. The symbol field contains the symbols that
	   yet have to be defined. fpentry instructions are combined with ldc instructions
	   if possible. Note that while stepping this combination doesn't take place */
	
	INSTRUCTION	instruction;
	ADDRESS		leading_address;
	ADDRESS		instruction_address;
	ADDRESS		jump_address;
	CHAR		*jump_symbol;
	
	ADDRESS definition;
	CHAR *symbol;
	
	ADDRESS next_address;
	INSTRUCTION next_instruction;

	leading_address=get_instr_address(address);
	instruction_address=skip_nops(leading_address);
	
	if (*definition_symbol)
	{
		definition=get_symbol_definition(data_area,*definition_symbol,&symbol);
	
		/* skip symbols defined before this instruction */
		
		while(*definition_symbol && definition<leading_address)
		{			
			*definition_symbol=get_next_symbol(*definition_symbol);
			if (*definition_symbol)
				definition=get_symbol_definition(data_area,*definition_symbol,&symbol);
		}
				
		/* show all symbols defined for this instruction address */
		
		while(*definition_symbol && definition>=leading_address && definition <= instruction_address)
		{
			file_print(file_nr,"%08X %s\n",definition,symbol);
			
			*definition_symbol=get_next_symbol(*definition_symbol);
			if (*definition_symbol)
				definition=get_symbol_definition(data_area,*definition_symbol,&symbol);
		}
	}
		
	get_instr(instruction_address,&instruction);
	next_address=get_offset_instr_address(instruction_address,1);

	if (instruction.operator!=Instr_ldc)
	{
		/* a ldc may be the start of an fpentry */
		
		if(leading_address<=iptr && iptr<=instruction_address)
			file_print (file_nr,"%08X      ->",instruction_address);
		else
			file_print (file_nr,"%08X        ",instruction_address);
	}

	switch(instruction.operator)
	{
		case Instr_jump:
		case Instr_cjump:
		case Instr_call:
		{
			if (instruction.arity!=0)
			{
				jump_address=instruction_address+instruction.size+instruction.operand;
				file_print(file_nr,"%s %-10d ->%08X",instruction.symbol,instruction.operand,jump_address);
				
				if (get_applied_symbol(jump_address,&jump_symbol))
					file_print(file_nr,": %s\n",jump_symbol);
				else
					file_print(file_nr,"\n");
			} else
				file_print(file_nr,"%s\n",instruction.symbol);
				
			break;
		}
		case Instr_ldc:
		{
			/* see if next instruction is fpentry and combine if so */
			
			get_instr(next_address,&next_instruction);
			
			if(next_instruction.operator==Instr_fpentry)
			{
				if(leading_address<=iptr && iptr<=instruction_address)
					file_print (file_nr,"%08X      ->",instruction_address);
				else
				{	leading_address=get_instr_address(next_address);
					instruction_address=skip_nops(next_address);
					
					if(leading_address<=iptr && iptr<=instruction_address)
						file_print (file_nr,"%08X      ->",instruction_address);
					else
						file_print (file_nr,"%08X        ",instruction_address);
				}

				switch (instruction.operand)
				{
					case 0x22:
						file_print(file_nr,"fpurn\n");
						break;
					case 0x06:
						file_print(file_nr,"fpurz\n");
						break;
					case 0x04:
						file_print(file_nr,"fpurp\n");
						break;
					case 0x05:
						file_print(file_nr,"fpurm\n");
						break;
					
					case 0x01:
						file_print(file_nr,"fpusqrtfirst\n");
						break;
					case 0x02:
						file_print(file_nr,"fpusqrtstep\n");
						break;
					case 0x03:
						file_print(file_nr,"fpusqrtlast\n");
						break;
						
					case 0x12:
						file_print(file_nr,"fpumulby2\n");
						break;
					case 0x11:
						file_print(file_nr,"fpudivby2\n");
						break;
					case 0x0A:
						file_print(file_nr,"fpuexpinc32\n");
						break;
					case 0x09:
						file_print(file_nr,"fpuexpdec32\n");
						break;
					
					case 0x0B:
						file_print(file_nr,"fpuabs\n");
						break;
					
					case 0x23:
						file_print(file_nr,"fpuseterr\n");
						break;
					case 0x9C:
						file_print(file_nr,"fpuclrerr\n");
						break;
				
					case 0x07:
						file_print(file_nr,"fpur32tor64\n");
						break;
					case 0x08:
						file_print(file_nr,"fpur64tor32\n");
						break;
						
					case 0x0E:
						file_print(file_nr,"fpuchki32\n");
						break;
					case 0x0F:
						file_print(file_nr,"fpuchki64\n");
						break;
						
					case 0x0D:
						file_print(file_nr,"fpunoround\n");
						break;
						
					default:
						file_print(file_nr,"unknown fpentry %08X\n",instruction.operand);
				}
				next_address=get_offset_instr_address(next_address,1);
			
			} else
			{	if(leading_address<=iptr && iptr<=instruction_address)
					file_print (file_nr,"%08X      ->",instruction_address);
				else
					file_print (file_nr,"%08X        ",instruction_address);

				file_print(file_nr,"%s %-10d\n",instruction.symbol,instruction.operand);
			}
			break;
		}
		case Instr_unknown:
			file_print(file_nr,"unknown opr: %08X\n",instruction.operand);
			break;

		default:
		{
			if (instruction.arity==1)
				file_print(file_nr,"%s %-10d\n",instruction.symbol,instruction.operand);
			else
				file_print(file_nr,"%s\n",instruction.symbol);
		}
	}
	return next_address;
}

VOID show_data(WORD file_nr, ADDRESS data, WORD size)
{
	WORD i;
	
	while (size>0)
	{	file_print(file_nr,"DATA ");
		if (size>16)
		{	for(i=0;i<16;++i)
			{	file_print(file_nr,"%02X ",*data++);
				--size;
			}
		} else
		{	while (size>0)
			{	file_print(file_nr,"%02X ",*data++);
				--size;
			}
		}
		file_print(file_nr,"\n");
	}
}

VOID show_code(WORD file_nr, ADDRESS iptr, ADDRESS address, WORD count)
{
	ADDRESS data_area;
	ADDRESS symbols;	
	ADDRESS data;
	WORD size;
	
	if (count<=0)
		return;

	address=get_instr_address(address);
	address=get_offset_instr_address(address,0);	/* address points certainly to instr */
	
	while (data_area=get_data_area(address))
	{
		get_symbols_and_data(data_area,&symbols,&data,&size);
				
		while(address<data_area && count>0)
		{	address=show_instruction(file_nr,iptr,address,data_area,&symbols);
			--count;
		}
		if (count==0)
			return;
		
		if (data)
			show_data(file_nr,data,size);
	}
	
	symbols=NULL;
	
	while(count>0)
	{	address=show_instruction(file_nr,iptr,address,(ADDRESS)NULL,&symbols);
		--count;
	}
}