#include "io.h"

#include "macros.h"
#include "messageno.h"
#include "packets.h"
#include "kernel.h"
#include "router.h"
#include "memory.h"
#include "processes.h"

#ifdef __INCLUDE_STANDARD__

#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#endif

/********************************************************************************
 *																				*
 *		import system vars														*
 *																				*
 ********************************************************************************/

extern NODE_INFO		node_info;
extern PROCESSOR_INFO	processor_info;

extern SYS				sys;

extern VOID debug_wake_up(WSP *wsp);

/********************************************************************
 *																	*
 *		message handling routines									*
 *																	*
 ********************************************************************/

static VOID handle_out_string(WORD kind,WORD source,WORD size,VREDUCER_PACKET *packet)
{
	WSP *regs;
	VCLEAN_STRING *source_string;
	
	regs			=pid_to_regs(packet->reducer_id);
	source_string	=(VCLEAN_STRING *)packet->contents;
	
	move((VCLEAN_STRING *)regs->a1->string_node.string,source_string,source_string->length+sizeof(WORD));
	
	run_reducer(packet->reducer_id);
}

static VOID handle_out_integer(WORD kind,WORD source,WORD size,VREDUCER_PACKET *packet)
{
	WSP *regs;

	regs=pid_to_regs(packet->reducer_id);
	regs->b1=*(WORD *)packet->contents;
	
	run_reducer(packet->reducer_id);
}

static VOID handle_out_c_string(WORD kind,WORD source,WORD size,VCHANNEL_PACKET *packet)
{
	WORD				|*chan;
	VCLEAN_STRING	*source_string;
	CHAR				*dest_string;
	
	chan=packet->chan;
	source_string=(VCLEAN_STRING *)packet->contents;
	
	dest_string=(CHAR *)packet->destination;
	move(dest_string,source_string->characters,source_string->length);
	dest_string[source_string->length]='\0';
	
	|*chan=source_string->length;
}

/************************************************************************
 *																		*
 *		initialisation													*
 *																		*
 ************************************************************************/

static RECEIVE_ACTION_TABLE io_actions[]=
{
	{MOutString,	{handle_out_string,		ignore_reception}},
	{MOutInteger,	{handle_out_integer,	ignore_reception}},
	{MOutCString,	{handle_out_c_string,	handle_out_c_string}}
};

BOOL initialise_io(VOID)
{
	set_packet_actions(io_actions,sizeof(io_actions)/sizeof(RECEIVE_ACTION_TABLE));
	
	return True;
}

/****************************************************************************
 *																			*
 *		errors/exceptions													*
 *																			*
 ****************************************************************************/

VOID fatal_error(BOOL sync, CHAR *string, ...)
{
	/* This routine is used to print error messages when some irrecoverable
	   error has occured. After the error message has been printed the
	   whole system will be aborted. Amongst others errors that occur
	   within the router are reported via this routine. In most cases
	   these aren't that serious that messages can't be routed anymore.
	   Still, programming errors might foul up the router badly (as the
	   T800 has no protection hardware). If this is the case it's not
	   certain errors can be routed: better to have lost in routing
	   than never to have routed at all.
	*/

	va_list args;
	WORD size;
	VFILE_PACKET *packet;
	
	if(processor_info.fatal_error_flag)
	{
		/* some fatal error occured earlier */
		
		if (sync)
			procstop();
		return;
	}
	
	processor_info.fatal_error_flag=True;
	packet=alloc_fatal_packet();
	
	va_start(args,string);
	size = vsprintf(((VCLEAN_STRING *)packet->contents)->characters,string,args);
	va_end(args);
	
	packet->file_number			=StdErr;
	packet->source_id			=0;
	packet->position			=0;
	((VCLEAN_STRING *)packet->contents)->length=size;
	
	send_fatal_packet(MFatalError,vsizeof(VFILE_PACKET)+sizeof(WORD)+size,sync);
	
	if (sync)
		procstop();
}

VOID tell_error(BOOL sync, WORD number)
{
	VFILE_PACKET *packet;

	/* This routine is used to get error number to the host */
	
	if(processor_info.fatal_error_flag)
	{	if (sync)
			procstop();
		return;
	}
		
	processor_info.fatal_error_flag=True;
	packet=alloc_fatal_packet();
	
	packet->file_number			=StdErr;
	packet->source_id			=0;
	packet->position			=number;
	
	send_fatal_packet(MErrno,vsizeof(VFILE_PACKET),sync);
	
	if (sync)
		procstop();	
}

VOID tell_action(BOOL sync, WORD number)
{
	/* This routine is used to get action number to the host */

	VFILE_PACKET *new_packet;
	
	if(!(new_packet=alloc_packet(sync)))
		return;
	
	new_packet->file_number			=StdErr;
	new_packet->source_id			=0;
	new_packet->position			=number;
	
	send_packet(MAction,unique_rout(ToHost),
				vsizeof(VFILE_PACKET),new_packet,sync);
}

VOID runtime_error (BOOL sync, CHAR *string, ...)
{
	va_list args;
	WORD size;
	VFILE_PACKET *new_packet;
	
	processor_info.error_flag=True;
	
	if(processor_info.fatal_error_flag)
	{	if (sync)
			procstop();
		return;
	}

	if(!(new_packet=alloc_packet(sync)))
		return;

	va_start(args,string);
	size = vsprintf(((VCLEAN_STRING *)new_packet->contents)->characters,string,args);
	va_end(args);

	new_packet->file_number			=StdErr;
	new_packet->source_id			=0;
	new_packet->position			=0;
	((VCLEAN_STRING *)new_packet->contents)->length=size;
	
	send_packet(MRuntimeError,unique_rout(ToHost),
				vsizeof(VFILE_PACKET)+sizeof(WORD)+size,new_packet,sync);
	
	if (sync)
		procstop();
}

VOID runtime_warning (BOOL sync, CHAR *string, ...)
{
	va_list args;
	WORD size;
	VFILE_PACKET *new_packet;
	
	processor_info.warning_flag=True;
	
	if(processor_info.fatal_error_flag)
	{	if (sync)
			procstop();
		return;
	}

	if(!(new_packet=alloc_packet(sync)))
		return;

	va_start(args,string);
	size = vsprintf(((VCLEAN_STRING *)new_packet->contents)->characters,string,args);
	va_end(args);

	new_packet->file_number			=StdErr;
	new_packet->source_id			=0;
	new_packet->position			=0;
	((VCLEAN_STRING *)new_packet->contents)->length=size;
	
	send_packet(MRuntimeWarning,unique_rout(ToHost),
				vsizeof(VFILE_PACKET)+sizeof(WORD)+size,new_packet,sync);
}

VOID no_crash(VOID)
{
	VFILE_PACKET *new_packet;
	
	if(!(new_packet=alloc_packet(True)))
		return;

	move(((VCLEAN_STRING *)new_packet->contents)->characters,node_info.machine_name,node_info.machine_name_length+1);
	
	new_packet->file_number			=StdErr;
	new_packet->source_id			=0;
	new_packet->position			=0;
	((VCLEAN_STRING *)new_packet->contents)->length=node_info.machine_name_length;
	
	send_packet(MOutCString,unique_rout(ToHost),
				vsizeof(VFILE_PACKET)+sizeof(WORD)+node_info.machine_name_length,new_packet,True);
}

/********************************************************************************
 *																				*
 *		synchronous file IO routines											*
 *																				*
 ********************************************************************************/

VOID hexit(WORD n)
{
	VFILE_PACKET *new_packet;
	
	if(!(new_packet=alloc_packet(True)))
		return;
	
	new_packet->position=n;
	
	send_packet(MExit,unique_rout(ToHost),vsizeof(VFILE_PACKET),new_packet,True);
}
	
VOID output_string(WORD destination, WORD file_nr, WORD position, CHAR *string, ...)
{
	va_list args;
	VFILE_PACKET *new_packet;
	WORD size;
	
	if(!(new_packet=alloc_packet(True)))
		return;
	
	va_start(args,string);
	size = vsprintf(((VCLEAN_STRING *)new_packet->contents)->characters,string,args);
	va_end(args);

	new_packet->file_number			=file_nr;
	new_packet->source_id			=0;
	new_packet->position			=position;
	((VCLEAN_STRING *)new_packet->contents)->length=size;
	
	send_packet(MOutString,destination,
				vsizeof(VFILE_PACKET)+sizeof(WORD)+size,new_packet,True);
}

VOID va_file_print(WORD file_nr, CHAR *string, va_list args)
{
	VFILE_PACKET *new_packet;
	WORD size;
	
	if(!(new_packet=alloc_packet(True)))
		return;
	
	size = vsprintf(((VCLEAN_STRING *)new_packet->contents)->characters,string,args);

	new_packet->file_number			=file_nr;
	new_packet->source_id			=0;
	new_packet->position			=0;
	((VCLEAN_STRING *)new_packet->contents)->length=size;
		
	send_packet(MOutCString,unique_rout(ToHost),
				vsizeof(VFILE_PACKET)+sizeof(WORD)+size,new_packet,True);
}

WORD file_read(WORD file_nr, CHAR *string)
{
	VFILE_PACKET *new_packet;
	WORD |sync_channel;
	
	if(!(new_packet=alloc_packet(True)))
		return 0;
	
	new_packet->file_number			=file_nr;
	new_packet->source_id			=(WORD)string;
	new_packet->position			=(WORD)&sync_channel;

	send_packet(MInCString,unique_rout(ToHost),
				vsizeof(VFILE_PACKET),new_packet,True);

	return |sync_channel;
}

VOID file_print(WORD file_nr, CHAR *string, ...)
{
	va_list args;
		
	va_start(args,string);
	va_file_print(file_nr,string,args);
	va_end(args);
}

VOID print(CHAR *string, ...)
{
	va_list args;
	
	va_start(args,string);
	va_file_print(StdOut,string,args);
	va_end(args);
}