#include "kernel.h"

#ifndef _TYPES_
#include "types.h"
#endif

#ifndef _SYSTEM_
#include "system.h"
#endif

#include "io.h"
#include "debugger.h"
#include "memory.h"
#include "hardware.h"
#include "depend.h"
#include "packets.h"
#include "graphs.h"
#include "shell.h"
#include "processes.h"
#include "errorno.h"
#include "messageno.h"
#include "macros.h"

#include "heapalloc.h"
#include "heapcollect.h"
#include "heapcopy.h"

#ifdef __INCLUDE_STANDARD__

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>

#endif

/********************************************************************************
 *																				*
 *		globally known data structures:											*
 *		-	options																*
 *		-	node_info: static info about node (e.g. neighbours)					*
 *		-	net_info: static info about network (e.g. diameter)					*
 *		-	processor_info: variable info (e.g. timings)						*
 *																				*
 ********************************************************************************/

OPTIONS			options;
NODE_INFO		node_info;
NET_INFO		net_info;
PROCESSOR_INFO	processor_info;

SYS				sys;

extern VOID wake_virtual_link(WORD linkno);

/********************************************************************************
 *																				*
 *		(VERY) low level routines												*
 *																				*
 ********************************************************************************/

#if __CACHE_ROUTINES__

LABEL *routine_on_chip (LABEL *routine, WORD size)
{
	/*	This routine must be executed at high priority after saving
		enough on-chip registers. Care should be taken that the on-chip
		regs of reducers are not referred in any way...
		
		The result of this function is the on-chip address where the indicated
		routine has been moved.
	*/
	
	move((ADDRESS)sys.reducer_info.on_chip_first_regs,(ADDRESS)routine,size);
	return (LABEL *)sys.reducer_info.on_chip_first_regs;
}

#endif

/********************************************************************************
 *																				*
 *		event channel monitor													*
 *																				*
 ********************************************************************************/

static VOID event_monitor(VOID)
{
	/*	this process monitors the event channel and invokes the debugger if
		something comes in*/
	
	for (;;)
	{
		BYTE e;
		chread(event_channel(),&e,1);
		
		/*
		runtime_warning(False,"event");
		BREAK();
		*/
	}
}

/********************************************************************************
 *																				*
 *		signal handler															*
 *																				*
 ********************************************************************************/

static VOID sigint_handler(int sig)
{
	if (sig == SIGINT)
		abort();
	
	/* reinstall signal handler */
	
	signal(sig,sigint_handler);
}

/************************************************************************************
 *																					*
 *		signal handler: this works in a straightforward way. just execute the		*
 *		function that is stored in the signal table defined below					*
 *																					*
 ************************************************************************************/

typedef VOID (*SIG_FUNCT) (WSP *wsp, WORD prio);

static VOID ignore_signal(WSP *wsp, WORD prio)
{
	runp(wsp,prio);
}

static VOID async_message(WSP *wsp, WORD prio)
{
	wake_virtual_link(NLinks);
	runp(wsp,prio);
}

static VOID sync_message(WSP *wsp, WORD prio)
{
	wake_virtual_link(NLinks);
}

static VOID signal_collect(WSP *wsp, WORD prio)
{
	call_collect();
	runp(wsp,prio);
}

static VOID get_ids(WORD proc, WORD n)
{
	/*	send message to retrieve n ids from the remote processor */
	
	GETID_PACKET *out;
	NODE *message;

	message=alloc_hpacket(MInIds,proc,sizeof(GETID_PACKET),&out);	
	out->number=n;
	send_heap_message(message);
}

static VOID put_wait(WORD id, WORD pid)
{
	/*	the heap should be in a consistent state */
	
	RELEASE_NODE		*new, *list;
	
	new=alloc_bytes(sizeof(RELEASE_NODE));
	
	list=table_idwait(id);
	
	if (list == NULL)
		new->code=sys.code_info.release_proc;
	else
		new->code=sys.code_info.release_proc1;
		
	new->wait	=list;
	new->pid	=pid;
	
	/*	adminster the reducer suspended */

	set_inactive(pid_to_env(pid));		

	table_idwait(id)=new;
}

BOOL channel_from_buffer(WORD pid,WORD proc)
{
	WSP				*wsp;
	WORD			*buffer;
	CHANNEL_NODE	*channel;
	
	/* buffers available, try to get id from it */
	
	buffer=(WORD *)(&table_idbuff1(proc)->string_node.string->characters);
	
	if (*buffer == 0)
		return False;
	
	/* buffer contains id, allocate a channel node and fill it */
	
	channel						=alloc_bytes(sizeof(CHANNEL_NODE));
	channel->code				=sys.code_info.channel_code;
	channel->address.processor	=proc;
	channel->address.weight		=MaxWeight;
	channel->address.id			=*(buffer + *buffer);

	/* link new channel node */
	
	channel->prev	=NULL;
	channel->next	=sys.comm_info.active_channels;
	
	if (sys.comm_info.active_channels != NULL)
		sys.comm_info.active_channels->prev=channel;
	
	sys.comm_info.active_channels=channel;

	wsp=pid_to_regs(pid);
	
	wsp->extra_local	=1;
	wsp->a1				=(NODE *)channel;

	/* adjust buffer administration */
	
	--(*buffer);
	
	/*	if this buffer has become empty, exchange buffers and fill
		the empty one again */
	
	if (*buffer == 0)
	{
		NODE *temp;
		
		temp=table_idbuff1(proc);
		table_idbuff1(proc)=table_idbuff2(proc);
		table_idbuff2(proc)=temp;
		
		get_ids(proc,options.idbuffsize);
	}
	return True;
}

static VOID create_channel(WSP *wsp, WORD prio)
{
	WORD proc, pid;
	
	/*	-> b1:	processor id
		<- a1:	channel node, possibly after suspending the reducer a
				while
	*/
	
	proc	=wsp->b1;
	pid		=get_pid(get_process_env(wsp));
	
	if (table_idbuff1(proc)!=NULL)
	{
		/*	buffers are present, try to get an id from it */
		
		if (channel_from_buffer(pid,proc))
		{
			runp(wsp,prio);
			return;
		}

		/*	buffer not yet filled, do nothing, it will become filled,
			put process in waiting list
		*/
			
	} else
	{
		STRING_NODE *string;
		WORD		*buffer;
		WORD		size;
		
		/*	no buffers allocated yet, allocate and start filling */
	
		size=(4+options.idbuffsize)*sizeof(WORD);
		
		string=alloc_words(size);
		buffer=(WORD *)(string+1);
		fill_string(string,buffer);
		*buffer++=(1+options.idbuffsize)*sizeof(WORD);		/* buff size plus number field */
		*buffer=0;											/* number field */
		
		table_idbuff1(proc)=(NODE *)string;

		string=alloc_words(size);
		buffer=(WORD *)(string+1);
		fill_string(string,buffer);
		*buffer++=(1+options.idbuffsize)*sizeof(WORD);
		*buffer=0;
		
		table_idbuff2(proc)=(NODE *)string;

		get_ids(proc,2*options.idbuffsize);
	}
		
	if (prio != LowPrio)
	{
		fatal_error(False,"create channel");
		return;
	}
	
	/*	suspend reducer */

	put_wait(proc,pid);
}

static VOID send_graph(WSP *wsp, WORD prio)
{
	/*
		The root of the graph is a closure
		
		-> b1: code of reducer
		-> a1: channel node to another processor
		-> a2: non-hnf graph that must be sent
	*/
	
	WORD copy, size;

	/*	build a copy of the graph and put it in a temporary global table entry */

	copy_graph(wsp->a2,copy);
	size=table_graph(copy)->string_node.string->length;
	graph_message(copy,wsp->a1->channel_node.address.processor,wsp->a1->channel_node.address.id,(ADDRESS)wsp->b1);

	transmit(copy,MReserveGraph,size,True,
			wsp->a1->channel_node.address.processor,
			wsp->a1->channel_node.address.id);
	
	/* and restart reducer */
	
	runp(wsp,prio);
}

static VOID global_node(WSP *wsp, WORD prio)
{
	/*	make the node in local 5 of the reducer globally known and allocate
		the amount of memory passed in b1 */
	
	WORD index;
		
	for (;;)
	{
		index=NewGlobalAddress(wsp->a1,RefMax);
	
		if (index <= 0)
		{
			/* couldn't get index, call garbage collector */

			collect(-index);
			continue;
		}
		
		if (wsp->b1 > 0)
			alloc_gblock(index,wsp->b1);
		
		/* return index in local 2... */
	
		wsp->b2=index;
		runp(wsp,prio);
		
		return;
	}
}

static VOID remove_global_pointer(WSP *wsp, WORD prio)
{
	/* decrement ref count table entry
	   -> b1: table index
	   -> b2: weight
	*/
	
	decrement_reference_count(wsp->b1,wsp->b2);
	runp(wsp,prio);
}

WSP *init_reducer (WSP *new_regs)
{
	/* allocate a reducer with regs 'new_regs'. If new_regs is not NULL, these
	   registers are supposed to be allocated already. Otherwise the registers
	   will be allocated as well. On succes the regs are returned, the amount of
	   needed words otherwise.
	*/

	WORD				size;
	ADDRESS				new_heap_ptr;
	REDUCER_ENVIRONMENT	*new_env;
	WORD				env_id;
	WORD				*local_heap_ptr;
   		
	/* allocate regs and environments */
	
	if (new_regs == NULL)
	{
		size=sys.reducer_info.init_stack_size+sys.reducer_info.init_heap_size+EnvironSize+ReducerSize;
		new_regs=(WSP *)(sys_root->heap_free_ptr+sizeof(HARD_WSP));
		new_env=(REDUCER_ENVIRONMENT *)(sys_root->heap_free_ptr+sizeof(REDUCER_LOCALS));
		
	} else
	{
		size=sys.reducer_info.init_stack_size+sys.reducer_info.init_heap_size+EnvironSize;
		new_env=(REDUCER_ENVIRONMENT *)(sys_root->heap_free_ptr);		
	}
	
	new_heap_ptr=(ADDRESS)((WORD *)sys_root->heap_free_ptr+size);
	
	if(new_heap_ptr>sys_root->heap_end_ptr)
		return size;
	
	sys_root->heap_free_ptr=new_heap_ptr;
	
	/* regs and environent allocated, make globally known */
	
	env_id=NewGlobalAddress(new_env,RefProcess);
	
	if (env_id <= 0)
		return -env_id;
	
	/* globally known environment allocated, initialise fields */
	
	local_heap_ptr=(WORD *)new_env+EnvironSize;
	
	new_regs->heap_start=local_heap_ptr;
	local_heap_ptr+=sys.reducer_info.init_heap_size;

	new_regs->heap_end		=local_heap_ptr;
	new_env->a_stack_start	=(NODE **)local_heap_ptr;
	new_regs->a_stack_ptr	=(NODE **)local_heap_ptr;

	local_heap_ptr+=sys.reducer_info.init_stack_size;

	new_env->b_stack_start		=local_heap_ptr;
	new_regs->b_stack_ptr		=local_heap_ptr-1;	/* make room for initial return address */
	
	new_env->registers			=new_regs;
	new_env->pid				=env_id;			/* active, highest bit clear */

	/* return allocated regs */
	
	return new_regs;
}

static VOID init_reducer1(WSP *wsp, WORD prio)
{
	/*	initialise a new reducer with on-chip regs. It has already been taken out
		of the free list. It has no environment, so this must be allocated.
		The garbage collector will not touch the on-chip registers.
			
		- b1 contains the startup code
		- b2 contains the workspace
		- a1 contains the node to update
		- a2 contains the node to evaluate

		leave values above intact and return the address of the new environment in
		b3
	*/

	WSP *new_regs;
   	
   	new_regs=init_reducer((WSP *)wsp->b2);

   	while ((WORD)new_regs >= 0)
   	{
   		/* note that the registers stay put */
   		
		collect((WORD)new_regs);
	   	new_regs=init_reducer((WSP *)wsp->b2);
	}
	
	/* return ptr to environment, so that reducer can access it */
	
	wsp->b3=(WORD)(new_regs->heap_start);

	runp(wsp,prio);
}

static VOID init_reducer2(WSP *wsp, WORD prio)
{
	/* initialise a new reducer with off-chip regs.
	
	   - b1 contains the startup code
	   - a1 contains the node to update
	   - a2 contains the node to evaluate

	   leave values above intact and return the address of the new registers in
	   b2
	*/

	WSP *new_regs;
	
	new_regs=init_reducer(NULL);
   	
   	while ((WORD)new_regs >= 0)
   	{   	
		collect((WORD)new_regs);
	   	new_regs=init_reducer(NULL);
	}

	/* return ptr to regs, so that reducer can access it */
	
	wsp->b2=(WORD)new_regs;
	
	runp(wsp,prio);
}

static VOID release_process(WSP *wsp, WORD prio)
{
	/*	release the local process of which the id is in the first element
		of the waiting list in a1. The rest of the waiting list will
		be returned in a1, unless it is empty
	*/
	
	RELEASE_NODE *release_node;
	
	release_node=&wsp->a1->release_node;
	
	if (release_node->wait != NULL)
		wsp->a1=(NODE *)release_node->wait;

	run_reducer(release_node->pid);
	runp(wsp,prio);
}

extern VOID return_hnf(WORD id,REQUEST *request);

static VOID release_request(WSP *wsp, WORD prio)
{
	/*	answer the request of which the id is in the first element of
		the waiting list in a1. The rest of the waiting list will be
		returned in a1, unless it is empty
	*/
	
	WORD			id;
	REQUEST_NODE	*request_node;
	RELEASE_NODE	*release_node;
	
	release_node=&wsp->a1->release_node;
	
	if (release_node->wait != NULL)
		wsp->a1=(NODE *)release_node->wait;

	id				=release_node->pid;
	request_node	=table_graph_requests(id);
	
	while (request_node != NULL)
	{
		REQUEST	request;

		request			=request_node->request;
		request_node	=request_node->wait;
		
		table_graph_requests(id)=request_node;

		return_hnf(id,&request);

		if (request_node != NULL)
			request_node=table_graph_requests(id);
	}

	runp(wsp,prio);
}

static VOID proc_at_distance(WSP *wsp, WORD prio)
{
	/* get some processor at the given distance (in b1) */

	WORD distance,first_index,last_index,set_size;
	DESTINATION_INFO *set,*match;
	
	distance=wsp->b1;
	
	/*	check (positive) distance, if it's too large for this network, the largest
		distance possible should be taken */

	if (distance < 0)
		distance = -distance;
	
	if (distance >= sys.comm_info.n_transputers)
		distance = sys.comm_info.n_transputers-1;
	
	/*	get info for this distance (there are entries for distances upto and
		including n_transputers */
	
	first_index	=sys.comm_info.dest_info[distance].distance_info;
	last_index	=sys.comm_info.dest_info[distance+1].distance_info;
	
	/*	if processors at this distance do not exist, one with the higest distance
		smaller than the given one (the least upperbound) should be found. Note that
		there is always a processor with distance 0 */
	
	while (first_index == last_index)
	{
		last_index=first_index;
		first_index=sys.comm_info.dest_info[--distance].distance_info;
	}
	
	/* found (non-empty) matching set of processors, determine bounds */
	
	set_size=last_index-first_index;
	set=&sys.comm_info.dest_info[first_index];
	
	/* get random processor out of this set */
	
	match=&set[rand() % set_size];
	
	if (match->proc <= 0 || match->proc > sys.comm_info.n_transputers || match->distance != distance)
	{
		fatal_error(False,"proc_at_distance delivered wrong proc");
		
		wsp->b1=sys.comm_info.transputer_id;
	
	} else
	
		wsp->b1=match->proc;
	
	runp(wsp,prio);
}

static VOID signal_error(WSP *wsp, WORD prio)
{
	tell_error(False,wsp->b1);
}

static SIG_FUNCT signal_table[]=
{
	ignore_signal,
	signal_collect,
	ignore_signal,
	async_message,
	sync_message,
	send_graph,
	global_node,
	signal_error,
	remove_global_pointer,
	init_reducer1,
	init_reducer2,
	release_process,
	proc_at_distance,
	release_request,
	create_channel
};

static VOID signal_handler(VOID)
{
	SIGNUM sig;
		
	for(;;)
	{
		chread(sys_signal_address,&sig,sizeof(SIGNUM));
		
		if (sig.number >= 0 && sig.number <= SLastSignal)
		{
			/* call handling function with workspace and prio */
			
			(*signal_table[sig.number])((WSP *)((WORD)sig.wsp&~1),(WORD)sig.wsp&1);
			continue;
			
		} else
		{
			processor_info.error_number=ESignalUnknown;
			fatal_error(False,"unknown signal: %d",sig.number);
			
			/*	actually, if there is a low proc interrupted, this must
				be killed as well */
			
			break;
		}
	}
	
	/*	signal handler has been quit, in order to keep all processes that
		might stil be running happy, just read all signals without processing
		them (leave calling processes pending) */
			
	for (;;)
		chread(sys_signal_address,&sig,sizeof(SIGNUM));
}

VOID kernel(VOID)
{		
	signal(SIGINT,sigint_handler);
	
	/*
		start up high prio system processes
		- event channel monitor
		- scheduler
		- collector
		- request handler
		- debugger
		- shell (root only)
	*/
	
	if (!push_process	(request_process,	2*KiloByte,	0))		return;
	if (!push_process	(debug_process,		4*KiloByte,	0))		return;
	
	if (!push_process	(event_monitor,		KiloByte,	0))		return;
	if (!push_process	(scheduler,			2*KiloByte,	0))		return;
	if (!push_process	(collector,			2*KiloByte,	0))		return;

	if (node_info.id==options.root_processor && !push_process(shell,4*KiloByte,0))
		return;
	
	procresch();
	
	signal_handler();
}
