#include "loader.h"

#include "messageno.h"
#include "errorno.h"
#include "io.h"
#include "packets.h"
#include "kernel.h"
#include "router.h"
#include "processes.h"
#include "heapalloc.h"
#include "memory.h"
#include "depend.h"
#include "hardware.h"

#ifdef __INCLUDE_STANDARD__

#include <stdio.h>
#include <string.h>

#endif

/********************************************************************************
 *																				*
 *		import system vars														*
 *																				*
 ********************************************************************************/

extern NODE_INFO		node_info;
extern PROCESSOR_INFO	processor_info;
extern NET_INFO			net_info;

extern OPTIONS			options;

extern SYS				sys;

/********************************************************************************
 *																				*
 *		global variables														*
 *																				*
 ********************************************************************************/

static CHANNEL	loader_channel;

/********************************************************************************
 *																				*
 *		startup routines														*
 *																				*
 ********************************************************************************/

static VOID synchronise_processors(VOID)
{
	VPACKET *ready;
	
	if (node_info.id!=options.root_processor)
	{
		/* this is not the root processor, send a message to root */
		
		if (!(ready=alloc_packet(False)))
			return;
				
		send_packet(MProgramInit,unique_rout(options.root_processor),0,ready,False);
		
	} else
	{
		if (processor_info.n_procs_initialised >= net_info.size)
			signal_channel(&loader_channel);
	}
}

static BOOL initialise_program(VOID)
{
	/* allocate heap */
		
	if ((sys_root->heap_free_ptr=push_mem(options.heap_size))==NULL)
	{
		processor_info.error_number=EAllocHeap;
		return False;
	}

	invalid_heap();
	valid_heap();
	
	/* execute prelude of system module to get start address etc. This must
	   be done at every transputer. The prelude is a plain c function that
	   is placed at the very beginning of the loaded code. To check possible
	   problems, first check if the system stack contains valid data. */

	if (sys_root != &sys.root)
		return False;

	if ((*(UBYTE *)processor_info.program_start)==0xD0)
	{
		/* program starts with 'stl 0', assume this is valid prelude... */

		VOID(* prelude)(VOID);
		
		prelude=(VOID *)processor_info.program_start;
		(*prelude)();
		
	} else
		return False;
	
	/* reset processor info */
	
	processor_info.ticks_collecting.lo		= 0;
	processor_info.ticks_collecting.hi		= 0;
	processor_info.ticks_running.lo			= 0;
	processor_info.ticks_running.hi			= 0;
	processor_info.ticks_copying.lo			= 0;
	processor_info.ticks_copying.hi			= 0;
	
	/* enable reducer messages */
	
	switch_action_set(ActivePacketSet);
	
	synchronise_processors();

	return True;
}

static VOID reset_clock(VOID)
{
	VPACKET *ready;

	if (!(ready=alloc_packet(False)))
		return;
	
	send_packet(MResetClock,unique_rout(ToHost),0,ready,False);
}

/********************************************************************
 *																	*
 *		message handling routines									*
 *																	*
 ********************************************************************/

static VOID handle_code_segment(WORD kind,WORD source,WORD size,VCODE_SEGMENT *packet)
{
	WORD length;
	UCHAR *code_segment;
	
	length=size-vsizeof(VCODE_SEGMENT);
	
	code_segment=user_allocate(length);
	
	if (code_segment == NULL)
		processor_info.error_number=EAllocProgram;
	else
	{
		move(code_segment,packet->code,length);
		processor_info.program_size+=length;
	}
	
	if (packet->position == -1 || packet->position == -3)
	{
		if (processor_info.error_number != ENoErr)
		{
			fatal_error(False,"load error");
			return;
		}
		
		processor_info.program_start=code_segment+length-processor_info.program_size;
		
		set_debug_run(packet->position==-3);

		if (!initialise_program())
		{
			runtime_warning(False,"could not initialise program");
			signal_channel(&loader_channel);
		}
	}
	
	if (packet->position == -2)
	{
		/* an error has occured, throw away program and wake up loader */
		
		user_free_all();

		processor_info.program_size=0;
		signal_channel(&loader_channel);
	}
}

static VOID initialised_action(WORD kind,WORD source,WORD size,VOID *packet)
{
	++processor_info.n_procs_initialised;
	
	if (processor_info.n_procs_initialised >= net_info.size)
		signal_channel(&loader_channel);
}

static VOID handle_halt(WORD kind,WORD source,WORD size,VOID *packet)
{
	PROCESSOR_INFO *i;
	
	/* get timing info and other status to return to root, use standard io
	   path to ensure ordered delivery of this packet: outstanding IO should
	   be sent away first... */
	
	i=alloc_packet(False);
	*i=processor_info;

	set_restart_routine(invalid_run_reducer);

	/*	kill all local reducers after ensuring safe status */
	
	low1();
	high();
	kill_reducers();

	/* swap packet handler to ignore reducer messages in transit (basically stuff
	   used for graph reduction: the heap should be considered invalid from
	   now on: it may be inspected, but reference counts, etc. may not be
	   correct... To avoid any damage, dangerous messages are simply ignored
	   by this processor from now after switching the action set (this 'freezes'
	   the heap in the state it was when a halt was performed). Note that there
	   may still be 'ignored' packets in the router queues, but we will not bother
	   removing them: if they are meant for this processor they will not do any
	   harm, and if they are meant for another, there is no reason why they should
	   not be sent */
	
	switch_action_set(InactivePacketSet);
	
	/* remove user allocated space (e.g. program) */
	
	user_free_all();
	processor_info.program_size=0;
	
	/* remove heap, but only after heap messages have been sent by the router */
	
	while(!(is_nil(sys.comm_info.first_message)))
		procresch();
	
	invalid_heap();
	pop_mem(options.heap_size);
	
	/* return exit message */
			
	send_packet(MProgramHalted,unique_rout(ToHost),sizeof(PROCESSOR_INFO),i,False);
}

static VOID wake_loader(WORD kind,WORD source,WORD size,VOID *packet)
{
	signal_channel(&loader_channel);
}

/************************************************************************
 *																		*
 *		exported routines												*
 *																		*
 ************************************************************************/

VOID load_program(CHAR *name, BOOL debug)
{
	processor_info.n_procs_initialised=1;
	
	if (debug)
		put_command(ToHost,MProgramDebugLoad,name);
	else
		put_command(ToHost,MProgramLoad,name);
	
	/* wait for load completion, on error the program size will be zero */
	
	wait_channel(&loader_channel);
}

VOID run_program(VOID)
{
	/* executed at root after synchronisation */
	
	if (processor_info.program_size!=0)
	{
		WORD pid;

		pid=halloc_reducer(sys.code_info.start_address);
		
		if (pid==0)
		{
			fatal_error(False,"couldn't allocate root reducer");
			return;
		}
	
		reset_clock();
		run_reducer(pid);
	
		/* wait for completion */
	
		wait_channel(&loader_channel);
	}
}

/************************************************************************
 *																		*
 *		initialisation													*
 *																		*
 ************************************************************************/

static RECEIVE_ACTION_TABLE loader_actions[]=
{
	{MProgramHalt,		{handle_halt,			handle_halt}},
	{MProgramSegment,	{handle_code_segment,	handle_code_segment}},	
	{MProgramInit,		{initialised_action,	initialised_action}},
	{MProgramReady, 	{wake_loader,			wake_loader}}
};

BOOL initialise_loader(VOID)
{
	set_packet_actions(loader_actions,sizeof(loader_actions)/sizeof(RECEIVE_ACTION_TABLE));
	
	return True;
}