/********************************************************************************
 *																				*
 *	Memory manager used by the runtime system to prevent difficulties with		*
 *	C malloc/free routines. The runtime	system usually has no need to			*
 *	perform a free, except for some initialising routines, such as the ones		*
 *	used by the router. The memory manager below is very simple. It allocates	*
 *	all available memory, except for some space that is probably needed by		*
 *	the C runtime system. memory space is regared as a large stack.				*
 *																				*
 *	this module should be the last module that is linked and that contains		*
 *	code that is needed after startup: unless specified otherwise, this			*
 *	routine will allocate all memory that is found after some initialisation	*
 *	routines (else it will use 'malloc', in the hope that the C compiler is		*
 *	able to copy with boards of different sizes).								*
 *																				*
 *	in addion this routine implements startup of C routines, as malloc would	*
 *	be used for this otherwise.													*
 *																				*
 ********************************************************************************/

#include "memory.h"
#include "processes.h"
#include "depend.h"
#include "macros.h"

#ifdef __INCLUDE_STANDARD__

#include <stdlib.h>
#include <stdarg.h>

#endif

extern NODE_INFO	node_info;

/************************************************************************************
 *																					*
 *		typedefs																	*
 *																					*
 ************************************************************************************/

typedef struct ALLOCATION_RECORD
{
	ADDRESS	start;
	ADDRESS	end;
	
	ADDRESS	first_free;
	ADDRESS	last_free;
	
	WORD	free_bytes;

} ALLOCATION_RECORD;

typedef struct PROCESS_PARS
{
	/*	process parameters, these will be preceded by the rest of the
		parameters and the stack space */
	
	ADDRESS	ret;
	ADDRESS	global_data;
	ADDRESS	process_record;
	
} PROCESS_PARS;

typedef struct PROCESS_RECORD
{
	/*	record for C processes, this will be preceded by the stack space
		and parameters */

	ADDRESS	stack_bottom;
	ADDRESS	prev_process_record;	
	ADDRESS	prev_wsp;
	WORD	error_number;
	
} PROCESS_RECORD;

/************************************************************************************
 *																					*
 *		globals																		*
 *																					*
 ************************************************************************************/

static ALLOCATION_RECORD int_mem;
static ALLOCATION_RECORD ext_mem;

static WORD user_reserved;

/************************************************************************************
 *																					*
 *		system memory allocation/deallocation										*
 *																					*
 ************************************************************************************/

VOID *push_int_mem (WORD size)
{
	VOID *result;
	
	/* always allocate a multiple of wordlength, from low to high */
	
	size=word_up(size);
	
	if (size > int_mem.free_bytes)
		return NULL;
	
	result=int_mem.first_free;
	
	int_mem.first_free+=size;
	int_mem.free_bytes-=size;
	
	return result;
}

VOID *push_mem (WORD size)
{
	/* always allocate a multiple of wordlength, from high to low */
	
	size=word_up(size);
	
	if (size > ext_mem.free_bytes)
		return NULL;
	
	ext_mem.last_free-=size;
	ext_mem.free_bytes-=size;
	
	return ext_mem.last_free;
}

VOID pop_mem (WORD size)
{
	/* always free a multiple of wordlength */
	
	size=word_up(size);

	ext_mem.last_free+=size;
	ext_mem.free_bytes+=size;
}

/************************************************************************************
 *																					*
 *		user allocation/deallocation (from other end of memory)						*
 *																					*
 ************************************************************************************/

VOID *user_allocate (WORD size)
{
	/* allocates from low end of memory	
	   always allocate a multiple of wordlength */
	
	VOID *result;
	
	size=word_up(size);
	
	if (size > ext_mem.free_bytes)
		return NULL;
	
	result=ext_mem.first_free;

	ext_mem.first_free+=size;
	ext_mem.free_bytes-=size;

	user_reserved+=size;

	return result;
}

VOID user_free_all (VOID)
{
	ext_mem.first_free-=user_reserved;
	ext_mem.free_bytes+=user_reserved;
	
	user_reserved=0;
}

/************************************************************************************
 *																					*
 *		allocation and start of C processes											*
 *																					*
 ************************************************************************************/

static VOID *get_global_data();
static VOID *get_process_record();

#pragma PACT linkage occam get_global_data
#pragma PACT linkage occam get_process_record

static VOID *get_global_data(global_data)
	VOID *global_data;
{
	return global_data;
}

static VOID *get_process_record(global_data,process_record)
	VOID *global_data,*process_record;
{
	return process_record;
}

#pragma PACT linkage pact get_global_data
#pragma PACT linkage pact get_process_record

VOID *push_process(VOID (* func) (), WORD stack_size, WORD npars, ...)
{
	/* allocate process space and start up */
	
	ADDRESS			stack;
	PROCESS_RECORD	*record;
	PROCESS_PARS	*pars;
	WORD			parameter_block_size;

	stack_size=word_up(stack_size);

	parameter_block_size=sizeof(PROCESS_PARS)+npars*sizeof(WORD);
	
	if ((stack=push_mem(sizeof(PROCESS_RECORD)+parameter_block_size+stack_size))==NULL)
		return NULL;
	
	pars=(PROCESS_PARS *)(stack+stack_size);
	record=(PROCESS_RECORD *)((ADDRESS)pars+parameter_block_size);
	
	record->error_number			=0;
	record->prev_wsp				=NULL;
	record->prev_process_record		=NULL;
	record->stack_bottom			=stack;
	
	pars->process_record			=(ADDRESS)(&record->stack_bottom);
	pars->global_data				=get_global_data();
	pars->ret						=(ADDRESS)func;

	if (npars > 0)
	{
		va_list			argp;
	
		va_start(argp, npars);
		move((ADDRESS)(pars+1),(ADDRESS)argp,npars*sizeof(WORD));
		va_end(argp);
	}
	
	set_iptr(pars,start__process);
	runp(pars,HighPrio);
}

/************************************************************************************
 *																					*
 *		determination of external memory size										*
 *																					*
 ************************************************************************************/

static BOOL valid_address(ADDRESS address)
{
	UBYTE contents,inverse;
	
	/* save byte */
	
	contents=*address;
	inverse=~contents;
	
	/* write inverted word */
	
	*address=inverse;
	
	/* read back and test */
		
	if (*address==inverse)
	{
		*address=contents;
		return True;
	}
	return False;
}

WORD ext_mem_size(VOID)
{
	/*	This routine tries to determine the size of external memory by writing
		values to memory and reading them back. This will not work if the
		highest bits of an address are ignored by your hardware. If this is
		the case you should set the external memory size explicitly with the
		-M option. Note: this routine starts searching after 'main' so that it will
		not overwrite its own code. It starts in steps of 1 Meg and then
		uses binary search.
	*/
	
	ADDRESS min, max, between;
	
	min=(ADDRESS)mint();
	max=(ADDRESS)reserve_off_chip_memory;
	
	while(valid_address(max))
	{
		min=max;
		max+=MegaByte;		
	}
	
	/* real bound somewhere in between min and max */
	
	between=max;
	
	while (min < max-1)
	{
		/*	beware for overflows, compute unsigned max-min and add half of
			it to min, note that when min=max-1, between will start to
			point to min and max will point to the first invalid address */
		
		between=(ADDRESS)((UWORD)min+(((UWORD)max-(UWORD)min)/2));
		
		if (valid_address(between))
			min=between;
		else
			max=between;		
	}
	
	/* max will point to the first invalid address */
	
	return (ADDRESS)((WORD)max-(WORD)mint());
}

/************************************************************************************
 *																					*
 *		initial memory allocation													*
 *																					*
 ************************************************************************************/

BOOL reserve_on_chip_memory (VOID)
{
	int_mem.start		=(ADDRESS)node_info.int_mem_start;	
	int_mem.end			=int_mem.start+node_info.int_mem_size;
	
	int_mem.first_free	=int_mem.start;
	int_mem.last_free	=int_mem.end;
	int_mem.free_bytes	=node_info.int_mem_size;
	
	return True;
}

VOID reset_external_memory(VOID)
{
	ext_mem.first_free	=ext_mem.start;
	ext_mem.last_free	=ext_mem.end;
	ext_mem.free_bytes	=ext_mem.end-ext_mem.start;

	user_reserved=0;
}

BOOL reserve_off_chip_memory (VOID)
{
	WORD amount;
	
	/*	allocate all memory, except for a 'little' that is used by C code and
		runtime system (well 100K would be allright, but it is not possible
		to allocate some 200k by means of malloc, so perhaps we have to take
		a closer look at the booters of the C compiler (use of out own network
		booter already has delivered us 80K + parallel booting) */
	
	amount=node_info.ext_mem_size-300*KiloByte;
	
	ext_mem.start=malloc(amount);

	if (ext_mem.start==NULL)	
		return False;
	
	ext_mem.end=ext_mem.start+amount;
		
	reset_external_memory();
		
	return True;
}
