
#include "types.h"
#include "hardware.h"
#include "depend.h"
#include "netboot.h"
#include "connect.h"
#include "memory.h"
#include "errorno.h"
#include "messageno.h"

#ifndef _SYSTEM_
#include "system.h"
#endif

#ifdef __INCLUDE_STANDARD__

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <node.h>
#include <server/linkio.h>

#endif

/********************************************************************************
 *																				*
 *		global data structures													*
 *																				*
 ********************************************************************************/

extern NODE_INFO		node_info;
extern NET_INFO			net_info;
extern OPTIONS			options;
extern PROCESSOR_INFO	processor_info;

/********************************************************************************
 *																				*
 *		global vars/constants													*
 *																				*
 ********************************************************************************/

static CHAR unimplemented[]	="\2\0\1";				/* size: 2, SP_UNIMPLEMENTED, NULL */
static CHAR boardsize[]		="TBOARDSIZE";
static CHAR bootpathfile[]	="TBOOTPATH";
static CHAR	read_binary[]	="rb";

#define probe_wait_time		10000					/* ten milliseconds (should be ample)	*/
#define request_wait_time	1000000					/* one second							*/
#define reset_wait_time		1000					/* one millisecond						*/
#define reset_address		0xC0

/************************************************************************************
 *																					*
 *		routines to parse options (root only)										*
 *																					*
 ************************************************************************************/

static VOID error_options(VOID)
{ 
	fprintf(stderr,"usage: server clear [options]\n\n");
	fprintf(stderr,"current options available are:\n");
	
	fprintf(stderr,"-h<size>        :set heap size\n");
	fprintf(stderr,"-p<size>        :set packet size\n");
	fprintf(stderr,"-i<size>        :set id buffer size\n");
	fprintf(stderr,"-s<size>        :set initial size of stacks\n");
	fprintf(stderr,"-st<number>     :set swap time of reducers\n");
	fprintf(stderr,"-c              :extra checks\n");
	
	fprintf(stderr,"-v              :verbose\n");
	fprintf(stderr,"-ev             :extra verbose\n");
	fprintf(stderr,"-g              :show garbage collections\n");
	
	fprintf(stderr,"-f              :use fixed seed for pseudo random number generator\n");
	fprintf(stderr,"-B              :show basic value only\n");
	fprintf(stderr,"-R<number>      :transputer used as root\n");
	fprintf(stderr,"-M<size>        :external memory size of transputers\n");
	
	exit(1);
}

static VOID set_options(WORD argc, CHAR *argv[])
{
	WORD argument;

	/* this routine should be called at the root transputer, first set default
	   options */	
	
	options.ext_mem_size			=0;				/* use local test */

	options.heap_size				=2*MegaByte;	/* size in bytes */
	options.init_heap_size			=25;			/* nr of words */
	options.init_stack_size			=4;				/* nr of words */
	options.increase_heap_size		=1000;			/* nr of words */
	options.init_id_table_size		=40;			/* nr of processors will be added */
	options.swap_time				=5;
	
	options.verbose					=False;
	options.extra_verbose			=False;
	options.show_garbage_collection	=False;
	options.check					=False;
	options.fixed_random			=False;
	options.basic_value_only		=False;
	options.line_buffering			=False;
	options.root_processor			=1;
	
	options.n_buffers				=10;
	options.packet_size				=1024;			/* size in bytes */
	
	options.idbuffsize				=5;
			
	if (argc<1)
		error_options();

	for (argument=1; argument<argc; argument++)
	{
		CHAR *string;
		
		string=argv[argument];
		
		if (*string++=='-')
		{	switch(*string++)
			{
				case 'h':
					
					if (!scan_value(&string,&options.heap_size) ||
						options.heap_size < 4)
						error_options();
					break;

				case 'p':
					
					if (!scan_value(&string,&options.packet_size) ||
						options.packet_size < 128)
						error_options();
					break;
				
				case 'i':
				
					if (!scan_value(&string,&options.idbuffsize) ||
						options.idbuffsize < 5)
						error_options();
					break;

				case 's':
				
					switch (*string)
					{
						case 't':
						{
							++string;
							if (!scan_value(&string,&options.swap_time) ||
								options.swap_time < 1)
								error_options();
							break;
						}
						default:
							if (!scan_value(&string,&options.init_stack_size) ||
								options.init_stack_size < 1)
								error_options();
					}
					break;
											
				case 'v':
					options.verbose=True;
					break;
				
				case 'c':
					options.check=True;
					break;
				
				case 'e':
				
					switch (*string++)
					{
						case 'v':
							options.verbose=True;
							options.extra_verbose=True;
							break;
						
						default:
							error_options();
					}
					break;
				
				case 'g':
					options.show_garbage_collection=True;
					break;
					
				case 'f':
					options.fixed_random=True;
					break;
					
				case 'B':
					options.basic_value_only=True;
					break;
				
				case 'L':
					options.line_buffering=True;
					break;
				
				case 'R':
				
					if (!scan_value(&string,&options.root_processor) ||
						options.root_processor < 1)
						error_options();
					break;
				
				case 'M':
					if (!scan_value(&string,&options.ext_mem_size) ||
						options.ext_mem_size < 500*KiloByte)
						error_options();
					break;
				
				default:
					error_options();
			}
		} else
			error_options();
	}
	
	if (options.heap_size < 2*(2*sizeof(WORD)+sizeof(REDUCER_LOCALS)+options.init_id_table_size*sizeof(ENTRY)))
	{
		fprintf(stderr,"ERROR: invalid heap size\n");
		exit(1);
	}
}

/********************************************************************************
 *																				*
 *		initialise external memory (as early as possible: after reception		*
 *		of options).															*
 *																				*
 ********************************************************************************/

static BOOL initialise_memory(VOID)
{
	WORD free;
	
	node_info.int_mem_start	=(VRED_ON_CHIP_DATA *)_Node.memstart;
	node_info.int_mem_size	=(ADDRESS)ExtMemStart-(ADDRESS)node_info.int_mem_start;

	if (options.ext_mem_size != 0)
		node_info.ext_mem_size=options.ext_mem_size;
	else
		node_info.ext_mem_size=ext_mem_size();

#if (VariableArraySize != 0)
	free=node_info.int_mem_size-(sizeof(VRED_ON_CHIP_DATA)-sizeof(REDUCER_LOCALS));
#else
	free=node_info.int_mem_size-sizeof(VRED_ON_CHIP_DATA);
#endif

	if (free < sizeof(REDUCER_LOCALS))
		node_info.max_reducers=0;
	else
		node_info.max_reducers=free/sizeof(REDUCER_LOCALS);

	if (node_info.max_reducers <= 0)
		return False;
	
	if (!reserve_on_chip_memory())
		return False;

	return reserve_off_chip_memory();
}

/********************************************************************************
 *																				*
 *			hardreset															*
 *																				*
 *	This function is used to reset the transputer directly connected via		*
 *	the link number passed as parameter, using the nearest neighbour reset		*
 *	for Parsytec hardware.														*
 *																				*
 ********************************************************************************/

static VOID hard_reset (WORD link)
{
	WORD *r;
	
	/* set r to (positive) address that is reserved for reset */
	
	r=(WORD *)reset_address;
	
	*r = 0;
	*r = 1;
	*r = 2;
	*r = 3;

	*r = (1<<link);
	
	wait(reset_wait_time);

	*r = 0;
}

/********************************************************************************
 *																				*
 *		get image file from host/parent											*
 *																				*
 ********************************************************************************/

static VOID *read_image(CHAR *name, WORD *size, WORD **bootpath, WORD *bootpath_size)
{
	/*	read bootable with given name. the file should be located in the
		current directory as the name is argv[0]. */

	VOID	*image;
	WORD	image_size,n;
	FILE	*file;
	CHAR	*path_filename;
	
	if ((file=fopen(name,read_binary))==NULL)
		return NULL;
	
	/* file opened, determine image size */
	
	fseek(file,0,SEEK_END);
	image_size=ftell(file);
	rewind(file);
	
	/* allocate buffer for image */

	if (!initialise_memory() ||  (image=push_mem(image_size))==NULL)
	{
		fclose(file);
		return NULL;
	}
	
	/* image buffer allocated and file open, read it */
	
	n=fread(image,sizeof(BYTE),image_size,file);
	fclose(file);
	
	if (n != image_size)
		return NULL;

	/*	if reading of the image file succeeds, we will start reading a
		bootpath file. If this does not succeed NULL is returned in the
		bootpath pointer. The booter may then proceed to boot without
		bootpath
	*/
	
	path_filename=getenv(bootpathfile);
	
	if (path_filename==NULL || (file=fopen(path_filename,read_binary))==NULL)
	{
		*bootpath=NULL;
		
		if (options.extra_verbose)
			fprintf(stderr,"booting reset network\n");
		
	} else
	{		
		/*	try read bootpath, first determine size and then read in all entries,
			note that the bootpath is stored in binary format */
	
		if (options.extra_verbose)
			fprintf(stderr,"booting network according to \"%s\"\n",path_filename);

		fseek(file,0,SEEK_END);
		*bootpath_size=ftell(file);
		rewind(file);
		
		/* allocate bootpath array */
		
		if ((*bootpath=push_mem(*bootpath_size))==NULL)
			fclose(file);
		else
		{
			/* read bootpath data */
		
			n=fread(*bootpath,1,*bootpath_size,file);
			fclose(file);
			
			if (n != *bootpath_size)
				*bootpath=NULL;
			
			/* bootpath read */
		}
	}

	if (*bootpath==NULL)
		*bootpath_size=0;

	/* image file read, return */
	
	*size=image_size;
	return image;
}

static VOID *chread_image(WORD *size, WORD **bootpath, WORD *bootpath_size)
{
	/*	read from link: options, size of image and image itself, the node id must be
		set to zero for a while; after reading of the image file to bootpath
		should be read from the link */

	VOID	*image;
	WORD	image_size;
	
	VOID	|*in;
	VOID	|*out;
	
	_Node.id	=0;
	in			=_Node.fromserver;
	out			=_Node.toserver;

	chread(in,&options,sizeof(OPTIONS));	
	chread(in,&image_size,sizeof(WORD));
	
	/* allocate buffer for image */
	
	if (!initialise_memory() || (image=push_mem(image_size))==NULL)
		image_size=0;
	
	/* return image size to parent to indicate failure/success */
	
	chwrite(out,&image_size,sizeof(WORD));
	
	if (image_size == 0)
	{
		/* no image, return NULL */
		
		*size=0;
		return NULL;
	}
	
	/* read image */
	
	chread(in,image,image_size);
	
	/* read bootpath size */
	
	chread(in,bootpath_size,sizeof(WORD));
	
	if (*bootpath_size > 0)
	{
		/* allocate a buffer for the bootpath data */
		
		if ((*bootpath=push_mem(*bootpath_size))==NULL)
			*bootpath_size=0;
	}
	
	/* return bootpath size to indicate failure/success */
	
	chwrite(out,bootpath_size,sizeof(WORD));
	
	/* and read bootpath if possible */
	
	if (*bootpath_size > 0)
		chread(in,*bootpath,*bootpath_size);
	else
		*bootpath=NULL;
	
	*size=image_size;
	return image;
}

/********************************************************************************
 *																				*
 *		probelink: test for reset processors (to be run at high prio)			*
 *																				*
 ********************************************************************************/

static BOOL probelink(VOID |*out,VOID *booter)
{
	/*	try to write a pre-booter to internal memory of reset neighbour,
		if this succeeds, the pre-booter wil simulate the reset state
		(for one link only, so that a parallel boot is possible) */
	
	return (chtwrite(out,booter,(*(UBYTE *)booter)+1,probe_wait_time)==0);
}

/********************************************************************************
 *																				*
 *		chboot																	*
 *																				*
 ********************************************************************************/

static BOOL handle_initial_requests(VOID |*out, VOID |*in)
{
	UBYTE	request[MINPACKSZ+16];
	UWORD	request_size;
	
	for (;;)
	{
		/*	time-out on basic request read: the booted processor may not
			have been able to initialise itself */
		
		if (chtread(in,request,MINPACKSZ,request_wait_time)!=0)
			return False;
		
		request_size=VLEN(request);
		
		if (request_size > MINMSGSZ+16)
			return False;
		
		if (request_size > MINMSGSZ)
			chread(in,&request[MINPACKSZ],request_size-MINMSGSZ);
		
		/* request totally read, handle it */
				
		switch (request[2])
		{
			case SP_GETENV:
			{
				/*	getenv: only boardsize environment variable can be handled:
					the memory size is returned
				*/
	
				UWORD	env_size;
				CHAR	*env;
				
				env_size	=VLEN(&request[3]);
				env			=(CHAR *)&request[5];
									
				if (env_size > sizeof(boardsize) || strncmp(env,boardsize,env_size)!=0)
					return False;
	
				/* set up return message with memory size in bytes */
				
				SETVLEN(request,14);
				request[2]=SP_SUCCESS;
				
				SETVLEN(&request[3],10);
				env=(CHAR *)&request[5];
				sprintf(env,"0x%08x",(_Node.process->processor->memsize)*sizeof(WORD));
				
				chwrite(out,request,16);				
				break;
			}
			case SP_DBG_INIT:
			case SP_ID:
			case SP_COMMAND:
			{
				/* return 'unimplemented' message */
									
				chwrite(out,unimplemented,sizeof(unimplemented));
				
				if (request[2]==SP_COMMAND)
					return True;
				
				break;
			}
			default:
				return False;
		}
	}
}

static BOOL	chboot(VOID |*out,VOID |*in,VOID *image, WORD image_size,WORD *bootpath,WORD bootpath_size)
{
	WORD result;
	
	/*	boot reset transputer with image (assuming probelink has determined
		that this processor has enough memory for booting (not yet tested!)) */
	
	chwrite(out,image,image_size);
		
	/*	handle initial requests from booted transputer */
	
	if (!handle_initial_requests(out,in))
		return False;
	
	/*	initial requests have been handled, proceed with boot, write options
		first */

	chwrite(out,&options,sizeof(OPTIONS));	
	chwrite(out,&image_size,sizeof(WORD));
	
	chread(in,&result,sizeof(WORD));
	
	if (result==0)
		return False;
	
	/*	other side ready to accept image */

	chwrite(out,image,image_size);
	
	/* forward new bootpath (if any) */
		
	chwrite(out,&bootpath_size,sizeof(WORD));
	chread(in,&result,sizeof(WORD));
	
	if (result!=0)
		chwrite(out,bootpath,bootpath_size);

	return True;
}

/********************************************************************************
 *																				*
 *		bootlink (process)														*
 *																				*
 ********************************************************************************/

static VOID bootlink(WORD linkno,VOID *image, WORD image_size,WORD *bootpath,WORD bootpath_size, WORD |*result)
{
	/*	boot processors as indicated by the bootpath */
	
	VOID |*out;
	VOID |*in;

	node_info.nbooted[linkno]=0;
	node_info.neighbour[linkno]=LUnknown;

	out		=link_to_chanout(linkno);
	in		=link_to_chanin(linkno);
	
	if (out!=_Node.toserver)
	{
		if (bootpath!=NULL)
		{
			/*	neighbour should be booted over this link, but this might fail
				if it does not have enough memory. */
			
			hard_reset(linkno);
			
			if (chboot(out,in,image,image_size,bootpath,bootpath_size))			
				chread(in,&node_info.nbooted[linkno],sizeof(WORD));
		}
	}
	
	/* return number booted over this link */
	
	chwrite(result,&node_info.nbooted[linkno],sizeof(WORD));
}

static VOID trybootlink(WORD linkno,VOID *image, WORD image_size, WORD |*result)
{
	/*	boot as many processors as possible over this link and return the
		amount booted over the channel.
	*/
	
	VOID |*out;
	VOID |*in;
	VOID *booter;

	node_info.nbooted[linkno]=0;
	node_info.neighbour[linkno]=LUnknown;

	out		=link_to_chanout(linkno);
	in		=link_to_chanin(linkno);
	booter	=pre__booter;

	if (out!=_Node.toserver)
		if (probelink(out,booter))
			if (chboot(out,in,image,image_size,NULL,0))			
				chread(in,&node_info.nbooted[linkno],sizeof(WORD));
	
	/* return number booted over this link */

	chwrite(result,&node_info.nbooted[linkno],sizeof(WORD));
}

/********************************************************************************
 *																				*
 *		netboot																	*
 *																				*
 ********************************************************************************/

static VOID set_default_node_info(VOID)
{
	/*	initialise clean parameters (which are partly a local copy of already
		known data). */

	node_info.id			=_Node.id;
	node_info.to_host		=_Node.toserver;
	node_info.from_host		=_Node.fromserver;
		
	/* set memory bounds for debugger (protect special RAM */
	
	node_info.low_valid_address		=(ADDRESS)node_info.int_mem_start;
	node_info.high_valid_address	=((ADDRESS)mint()+node_info.ext_mem_size);

	processor_info.error_number		=ENoErr;
	processor_info.fatal_error_flag	=False;
	processor_info.error_flag		=False;
	processor_info.warning_flag		=False;

	processor_info.program_size	=0;
	
	sprintf(node_info.machine_name,"T%d ",node_info.id);
	node_info.machine_name_length=strlen(node_info.machine_name);	
}

VOID netboot (WORD argc,CHAR *argv[])
{
	VOID	*image;
	WORD	image_size;
		
	WORD	*bootpath;
	WORD	bootpath_size;
	
	WORD	*sub_bootpath[NLinks];
	WORD	sub_bootpath_size[NLinks];

	WORD	i,nbooted,new_id;
	WORD	|linkresult[NLinks];
	
	/****************************************************************************
	 *																			*
	 *		read image that must be used for netboot							*
	 *																			*
	 ****************************************************************************/
	
	if (argv[0] != NULL && *argv[0] != '\0')
	{
		/*	first parse options, so that we are not going to start an
		expensive boot procedure if something is wrong. 
		*/
	
		set_options(argc,argv);
		
		if ((image=read_image(argv[0],&image_size,&bootpath,&bootpath_size))==NULL)
		{
			fprintf(stderr,"cannot read image\n");
			exit(1);
		}
		
	} else
	{
		if ((image=chread_image(&image_size,&bootpath,&bootpath_size))==NULL)
			procstop();
		
		/* the parent will not boot this processor, aborted locally */
	}
	
	/************************************************************************
	 *																		*
	 *		memory has been 'reserved' already,	be careful not to use		*
	 *		(initialise) internal memory in this routine as					*
	 *		the stack is in it.												*
	 *																		*
	 *	options and image have been read, id indicates to some extend		*
	 *	whether the processor is the root or not: if it is zero, this		*
	 *	processor is not the root. The final id will be distributed later	*
	 *																		*
	 ************************************************************************/

	nbooted=1;			/* this processor has been booted */

	if (bootpath==NULL)
	{
		/************************************************************************
		 *																		*
		 *		boot all reset links											*
		 *																		*
		 ************************************************************************/
		 
		 /* start booting over all reset links in parallel */
	
		for(i=0;i<NLinks;++i)
		{
			chreset(&linkresult[i]);
			push_process(trybootlink,KiloByte,4,i,image,image_size,&linkresult[i]);
		}

	} else
	{
		/************************************************************************
		 *																		*
		 *		determine sub-bootpaths											*
		 *																		*
		 ************************************************************************/
	
		WORD *p;

		p=&bootpath[1];
		
		for (i=0; i<NLinks; ++i)
		{
			if (*p==0)
			{
				/* no one connected to this link */
				
				sub_bootpath[i]			=NULL;
				sub_bootpath_size[i]	=0;
				
				++p;
				
			} else
			{
				/* someone connected to this link, extract sub-path */
				
				int n_sub_entries;
				
				n_sub_entries=1+(*p)*NLinks;
				
				sub_bootpath[i]			=p;
				sub_bootpath_size[i]	=n_sub_entries*sizeof(WORD);
				
				p+=n_sub_entries;
			}
		}
		
		/* start booting over all links in parallel */
		
		for(i=0;i<NLinks;++i)
		{
			chreset(&linkresult[i]);
			push_process(bootlink,KiloByte,6,i,image,image_size,sub_bootpath[i],sub_bootpath_size[i],&linkresult[i]);
		}
	}
	
	for (i=0;i<NLinks;++i)
	{
		WORD result;
		
		chread(&linkresult[i],&result,sizeof(WORD));
		nbooted+=result;		
	}
	
	/* give link booters time to die */
	
	procresch();
	
	/* all links have been booted, reset external memory */
	
	reset_external_memory();
	
	/************************************************************************
	 *																		*
	 *		Pass no of procs booted from here upwards and then				*
	 *		distribute netsize, id, parent id and options					*
	 *																		*
	 ************************************************************************/
	
	if (_Node.id)
	{
		net_info.size	=nbooted;
		_Node.id		=RootId;
		
		node_info.neighbour[chanin_to_link(_Node.fromserver)]=ToHost;

		/* options already set at root */
		
	} else
	{
		chwrite(_Node.toserver,&nbooted,sizeof(WORD));
		
		chread(_Node.fromserver,&net_info.size,sizeof(WORD));
		chread(_Node.fromserver,&_Node.id,sizeof(WORD));
		chread(_Node.fromserver,&node_info.neighbour[chanin_to_link(_Node.fromserver)],sizeof(WORD));
	}

	/* forward network size and id's */
	
	new_id=_Node.id+1;
	
	for (i=0; i<NLinks; ++i)
	{
		if (node_info.nbooted[i]>0)
		{
			sendlink(i,&net_info.size,sizeof(WORD));		/* pass netsize		*/
			sendlink(i,&new_id,sizeof(WORD));				/* pass id			*/
			sendlink(i,&_Node.id,sizeof(WORD));				/* pass parent id	*/
			
			new_id+=node_info.nbooted[i];
		}
	}
	
	set_default_node_info();
}
