#include "packets.h"

#ifndef _HARDWARE_
#include "hardware.h"
#endif

#ifndef _SYSTEM_
#include "system.h"
#endif

#include "messageno.h"
#include "router.h"
#include "nodes.h"
#include "memory.h"
#include "heapalloc.h"
#include "io.h"
#include "kernel.h"
#include "graphs.h"

#ifdef __INCLUDE_STANDARD__
#include <stdio.h>
#include <string.h>
#include <asm.h>
#endif

/************************************************************************************
 *																					*
 *		imports from system															*
 *																					*
 ************************************************************************************/

extern NODE_INFO	node_info;
extern NET_INFO		net_info;
extern OPTIONS		options;

extern SYS 			sys;

extern BOOL			collector_running;

/************************************************************************************
 *																					*
 *		global variables															*
 *																					*
 ************************************************************************************/

static VQUEUE_PACKET *first_out_packet;			/* first of system packets to be sent */
static VQUEUE_PACKET *last_out_packet;			/* last of system packets to be sent */

extern VQUEUE_PACKET *fatal_error_packet;		/* preallocated fatal error packet */

static VQUEUE_PACKET *free_packets;				/* list of free packets */

static WORD max_body_size;						/* packet length + source field */
static WORD packet_position;					/* position of current packet to be sent */
static BOOL out_heap_lock;

/********************************************************************
 *																	*
 *		table constants and vars									*
 *																	*
 ********************************************************************/

static RECEIVE_ACTION	*current_table;
static RECEIVE_ACTION	**action_table;

/************************************************************************************
 *																					*
 *		settings																	*
 *																					*
 ************************************************************************************/

VOID block_heap_messages(VOID)
{
	out_heap_lock=True;
}

VOID unblock_heap_messages(VOID)
{
	out_heap_lock=False;
	wake_virtual_link(NLinks);
}

/********************************************************************
 *																	*
 *		system packets (from global finite pool)					*
 *																	*
 ********************************************************************/

/* there's a small amount of buffers available to perform some specific
   tasks like debugging and printing of status information by the
   runtime system. If this amount isn't enough this results in a
   fatal error (which can always be printed). */

static BOOL init_system_packets(WORD size, WORD amount)
{
	WORD i;
	
	if(!(free_packets=(VQUEUE_PACKET *)push_mem(vsizeof(VQUEUE_PACKET)+size)))
		return False;

	free_packets->admin.next=NULL;
	
	for (i=0;i<amount;++i)
	{
		VQUEUE_PACKET *new_packet;

		if(!(new_packet=(VQUEUE_PACKET *)push_mem(vsizeof(VQUEUE_PACKET)+size)))
			return False;
		
		new_packet->admin.next=free_packets;
		free_packets=new_packet;
	}
	return True;
}

VOID *alloc_packet(BOOL sync)
{
	/* only valid for high priority */
	
	VQUEUE_PACKET *new_packet;

	new_packet=free_packets;
	
	if (!new_packet)
	{	fatal_error(sync,"system packet pool empty");
		return NULL;
	}
	
	free_packets=free_packets->admin.next;
	new_packet->admin.system_packet=True;
	
	return &new_packet->packet;
}

VOID *alloc_fatal_packet(VOID)
{
	/* indicate that the fatal error packet must not be freed */
	
	fatal_error_packet->admin.system_packet=False;
	
	return &fatal_error_packet->packet;
}

VOID send_packet_from(WORD kind, WORD dest, WORD source, WORD size, VOID *packet, BOOL sync)
{
	/* depending on the value of 'sync' this routine will either
	   return immediately, or only after the message has been accepted
	   by the router.
	*/
	
	VQUEUE_PACKET *queue_packet;
	
	queue_packet=queue_address(packet);
	
	queue_packet->admin.next=NULL;						/* no next packet		*/
	chreset(&(queue_packet->admin.ready));
	
	queue_packet->header.size			=size+sizeof(WORD);
	queue_packet->header.kind			=kind;
	queue_packet->header.destination	=dest;
	
	queue_packet->source				=source;
	
	/* priority should be high */
		
	if (first_out_packet)
		last_out_packet->admin.next=queue_packet;
	else
		first_out_packet=queue_packet;
	last_out_packet=queue_packet;
		
	/* the packet is in the list, wake up router process */
	
	wake_virtual_link(NLinks);
	
	if (sync)
		wait_channel(&queue_packet->admin.ready);	
}

VOID send_fatal_packet(WORD kind,WORD size,BOOL sync)
{
	/* block_heap_messages(); */
	
	/* put packet in list of outgoing messages */

	fatal_error_packet->admin.next=NULL;				/* no next packet */
	chreset(&(fatal_error_packet->admin.ready));
	
	fatal_error_packet->header.size			=size+sizeof(WORD);
	fatal_error_packet->header.kind			=kind;
	fatal_error_packet->header.destination	=ToHost;
	
	fatal_error_packet->source				=node_info.id;

	first_out_packet=fatal_error_packet;							/* discard all others */
	last_out_packet=fatal_error_packet;

	/* the packet is in the list, wake up router process */
	
	wake_virtual_link(NLinks);
	
	if (sync)
		wait_channel(&fatal_error_packet->admin.ready);	
}

/************************************************************************
 *																		*
 *		heap messages													*
 *																		*
 ************************************************************************/

NODE *alloc_hpacket_from(WORD kind,WORD dest,WORD source,WORD size,VOID **packet_pointer)
{
	NODE			*cons;
	STRING_NODE 	*string;
	VBYTE_HPACKET	*header;
	
	/* allocate space for cons, string and packet */
	
	cons=alloc_words(sizeof(HNF2_NODE)+sizeof(STRING_NODE)+vsizeof(VBYTE_HPACKET)+size);
	
	string		=(STRING_NODE *)(((HNF2_NODE *)cons)+1);
	header		=(VBYTE_HPACKET *)(string+1);

	fill_cons(cons,string,string);
	fill_string(string,header);

	header->header.size			=vsizeof(VBYTE_HPACKET)-sizeof(WORD)+size;
	header->header.kind			=kind;
	header->header.destination	=dest;
	header->source				=source;
	
	*packet_pointer=&header->data;
	
	return cons;
}

NODE *alloc_hpacket2_from(WORD kind,WORD dest,WORD source,WORD size,WORD body,VOID **header_pointer)
{
	NODE			*cons, *tuple;
	STRING_NODE 	*string;
	VBYTE_HPACKET	*header;
	
	/* allocate space for cons, tuple, string and packet */
	
	cons=alloc_words(2*sizeof(HNF2_NODE)+sizeof(STRING_NODE)+vsizeof(VBYTE_HPACKET)+size);
	
	tuple		=(NODE *)(((HNF2_NODE *)cons)+1);
	string		=(STRING_NODE *)(((HNF2_NODE *)tuple)+1);
	header		=(VBYTE_HPACKET *)(string+1);

	fill_cons(cons,tuple,tuple);
	fill_tuple2(tuple,string,table_graph(body));
	fill_string(string,header);

	header->header.size			=vsizeof(VBYTE_HPACKET)-sizeof(WORD)+size;
	header->header.kind			=kind;
	header->header.destination	=dest;
	header->source				=source;
	
	*header_pointer=&header->data;
	
	return cons;
}

#define alloc_hpacket(kind,dest,size,packet)	alloc_hpacket_from(kind,dest,node_info.id,size,(VOID **)packet)
#define alloc_hpacket2(kind,dest,size,body,header)	alloc_hpacket2_from(kind,dest,node_info.id,size,body,(VOID **)header)

VOID send_heap_message(NODE *message)
{
	/* this routine assumes it's argument is a dyadic hnf node node that contains the
	   message in its first argument. The second argument will be filled in by this routine */
	
	if(is_nil(sys.comm_info.first_message))
	{
		message->hnf2_node.arg2=sys.comm_info.first_message;
		sys.comm_info.first_message=message;
		
	}else
	{
		message->hnf2_node.arg2=sys.comm_info.last_message->hnf2_node.arg2;
		sys.comm_info.last_message->hnf2_node.arg2=message;
	}
	sys.comm_info.last_message=message;

	wake_virtual_link(NLinks);
}

/************************************************************************************
 *																					*
 *		administration of actions to be performed on reception of packet			*
 *																					*
 ************************************************************************************/

VOID ignore_reception (WORD kind,WORD source,WORD size,VOID *packet)
{
	/*	used for 'known' packets that should be ignored (for instance, after
		a halt has been executed
	*/
}

static VOID default_reception(WORD kind,WORD source,WORD size,VOID *packet)
{
	fatal_error(False,"no action specified for packet kind %d",kind);
}

VOID set_packet_actions(RECEIVE_ACTION_TABLE *action,WORD amount)
{
	WORD i;
	
	for (i=0; i<amount; ++i,++action)
	{
		action_table[ActivePacketSet][action->code]		=action->actions[ActivePacketSet];
		action_table[InactivePacketSet][action->code]	=action->actions[InactivePacketSet];
	}
}

RECEIVE_ACTION get_packet_action(WORD kind)
{
	return current_table[kind];
}

VOID switch_action_set(WORD set)
{
	if (set < 0 || set >= NPacketSets)
	{
		fatal_error(False,"switch to invalid action set");
		return;
	}
	
	current_table=action_table[set];
}

static BOOL init_receive_actions(WORD n_kinds)
{
	WORD i,j;
	
	if (!(action_table=(RECEIVE_ACTION **)push_mem(NPacketSets*sizeof(RECEIVE_ACTION *))))
		return False;
	
	for (i=0; i<NPacketSets;i++)
	{
		if(!(action_table[i]=(RECEIVE_ACTION *)push_mem(n_kinds*sizeof(RECEIVE_ACTION))))
			return False;
	
		for(j=0;j<n_kinds;j++)
			action_table[i][j]=default_reception;
	}
	
	return True;
}

/************************************************************************************
 *																					*
 *		routines used by router	(at high prio)										*
 *																					*
 ************************************************************************************/

static BOOL packet_available(VOID)
{
	if(first_out_packet)
		return True;
	
	if(out_heap_lock)
		return False;

	if (!is_nil(sys.comm_info.first_message))
		return True;

	return sys.comm_info.first_inactive_channel != NULL;
}

static VOID in_system_packet(VPACKET_BUFFER *buffer)
{
	/* stores system packet in supplied buffer */
	
	VQUEUE_PACKET *queue_element;
	
	queue_element=first_out_packet;
	
	/* the packet length, kind and destination is put in the header */
	
	move(buffer_header_address(buffer),(ADDRESS)&queue_element->header,queue_element->header.size+sizeof(PACKET_HEADER));
	
	/* the buffer is filled, remove packet from list */
	
	first_out_packet=queue_element->admin.next;
	if (!first_out_packet)
		last_out_packet=NULL;
	
	/* and unblock anyone waiting for the packet to be accepted */
	
	signal_channel(&queue_element->admin.ready);
	
	if (queue_element->admin.system_packet)
	{	
		queue_element->admin.next=free_packets;
		free_packets=queue_element;
	}
}

static VOID remove_first_message(VOID)
{	
	NODE *cons_node;
	
	cons_node=sys.comm_info.first_message;
	
	/* remove first cons node from message list */
	
	sys.comm_info.first_message=cons_node->hnf2_node.arg2;		
	if(is_nil(sys.comm_info.first_message))
		sys.comm_info.last_message=sys.comm_info.first_message;

	/*	set both fields of cons node to first arg (it may be reused later, but the
		rest of the list must not be referred by it to avoid space leaks!! */
	
	cons_node->hnf2_node.arg2=cons_node->hnf2_node.arg1;
}

static VOID in_decrement_refcount_packet(VPACKET_BUFFER *buffer)
{
	/* builds decrement reference count messages in the supplied buffer
	   (it doesn't combine decrement messages yet...) */
	
	CHANNEL_NODE *garbage_channel;
	DECREF_PACKET *message;
	
	garbage_channel	=sys.comm_info.first_inactive_channel;
	message			=buffer_body_address(buffer);
	
	buffer->header.length		=sizeof(DECREF_PACKET)+sizeof(WORD);
	buffer->header.kind			=MDecRef;
	buffer->header.destination	=garbage_channel->address.processor;
	
	buffer->body.source			=node_info.id;
	
	message->id.dest			=garbage_channel->address.id;
	message->weight				=garbage_channel->address.weight;
	
	/* message built; remove the garbage channel from the list */
	
	sys.comm_info.first_inactive_channel=garbage_channel->next;

	if (garbage_channel->next == NULL)
		sys.comm_info.last_inactive_channel=NULL;
	else
		sys.comm_info.first_inactive_channel->prev=NULL;
}

static VOID in_reducer_packet(VPACKET_BUFFER *buffer)
{
	/*	stores (part of) a reducer message in the supplied buffer */
	
	NODE			*head;
	WORD			head_arity;
	ADDRESS			destination;
	VCLEAN_STRING	*message;
	
	head=sys.comm_info.first_message->hnf2_node.arg1;
	head_arity=get_node_arity(head);
	
	destination=(ADDRESS)&buffer->header.kind;
	
	if (head_arity==0)
	{
		/* the head is supposed to be a string node that may not be bigger than
		   the packet size: this is the way to send small messages. The length of
		   the message is smaller than the size of the string as the kind and the
		   destination are put in the header. The source is put in the body. */
		
		message=head->string_node.string;
		buffer->header.length=message->length-2*sizeof(WORD);
		
		if(buffer->header.length<=max_body_size)
		{
			move((ADDRESS)destination,(ADDRESS)message->characters,message->length);
			remove_first_message();
			return;
		}
		fatal_error(False,"string contains too large a message");
		in_system_packet(buffer);
		remove_first_message();
		
		return;
	}
	
	if (head_arity==2)
	{
		WORD packet_length;
		WORD buffer_free;
			
		/*	the head consists of a dyadic tuple. The first element contains
			a string with a 'header' and the second one contains the message
			itself (also a string) The message will be split into packets if
			it's too long. every packet is prefixed with the header (which may
			not be greater than the packet size) and a number that indicates
			the position of the packet in the original message
		*/
			
		message=head->hnf2_node.arg1->string_node.string;

		buffer->header.length=message->length-sizeof(WORD);

		/*
			we have used the temporary 'packet_length' here as use of
			message->length results in a compiler error (core stack overflow,
			which is probably caused by duplicating the length on the stack)
		*/
		
		packet_length=message->length;		
		move(destination,(ADDRESS)message->characters,packet_length);
		
		destination+=message->length;
		*((WORD *)destination)=packet_position;
		destination+=sizeof(WORD);
		
		/* header and position prefixed, now append rest of message */			
		
		message=head->hnf2_node.arg2->string_node.string;
		
		buffer_free=max_body_size-buffer->header.length;
		packet_length=message->length-packet_position;
		
		if(packet_length > buffer_free)
		{	
			packet_length=buffer_free;
		
			buffer->header.length+=packet_length;				
			move((ADDRESS)destination,(ADDRESS)message->characters+packet_position,packet_length);
			packet_position+=packet_length;
			return;
		}
		buffer->header.length+=packet_length;
		move((ADDRESS)destination,(ADDRESS)message->characters+packet_position,packet_length);
		packet_position=0;
		remove_first_message();
		return;
	}
	
	fatal_error(False,"message consists of %d parts (code %08X)",head_arity,get_code(head));
	in_system_packet(buffer);
	remove_first_message();
}

static VOID in_packet(VPACKET_BUFFER *buffer)
	
	/* this routine finds a packet to store in the indicated buffer.
	   Packets that are supplied by the runtime system are handled
	   first. The process that provides the packet is released. */
{
	/* decrement ref count messages have highest prio... */
	
	if (sys.comm_info.first_inactive_channel != NULL)
	{
		in_decrement_refcount_packet(buffer);
		return;
	}

	if(first_out_packet)
	{	in_system_packet(buffer);
		return;
	}
	
	in_reducer_packet(buffer);
}

static VOID out_packet(VPACKET_BUFFER *buffer)
{
	/* call appropriate action with packet for parameter */

	(*current_table[buffer->header.kind])
		(buffer->header.kind,
		buffer->body.source,
		buffer->header.length-sizeof(WORD),
		(VPACKET *)&buffer->body.bytes);
}

/****************************************************************************
 *																			*
 *		initialisation														*
 *																			*
 ****************************************************************************/

BOOL initialise_packets(WORD n_kinds)
{
	first_out_packet	=NULL;
	last_out_packet		=NULL;
	
	packet_position	=0;
	max_body_size	=options.packet_size+sizeof(WORD);
	
	if (!initialise_router())
		return False;
	
	if (!init_system_packets(256,32))
		return False;
	
	if (!init_receive_actions(n_kinds))
		return False;

	switch_action_set(InactivePacketSet);
	
	out_heap_lock			=True;
	
	/* initialise fatal error packet */
	
	fatal_error_packet				=free_packets;		
	free_packets					=free_packets->admin.next;
	fatal_error_packet->admin.next	=NULL;
			
	return True;
}

BOOL start_packets(VOID)
{
	if (!(start_virtual_link(NLinks,out_packet,3*KiloByte,in_packet,3*KiloByte,packet_available)))
		return False;

	return start_router();
}