#include "processes.h"

#include "system.h"
#include "hardware.h"
#include "depend.h"

#ifdef __INCLUDE_STANDARD__
#include <string.h>
#endif

extern NODE_INFO		node_info;
extern PROCESSOR_INFO	processor_info;
extern OPTIONS			options;

extern SYS				sys;

#define CHECK_DEADLOCK	0

/************************************************************************************
 *																					*
 *		global pointer to routine take wakes up processes							*
 *																					*
 ************************************************************************************/

REDUCER_START_ROUTINE reducer_start_routine;

/************************************************************************************
 *																					*
 *		workspace of debugger process												*
 *																					*
 ************************************************************************************/

typedef struct DWSP
{
	WORD	front_ptr;			/* saved low prio front ptr	*/
	WORD	back_ptr;			/* saved low prio back ptr	*/
	
	WSP		*wsp;				/* saved wsp				*/
	WORD	Areg;				/* saved Areg				*/
	WORD	Error;				/* saved error flag			*/
	
	FLOAT	fAreg;
	FLOAT	fBreg;
	FLOAT	fCreg;
	WORD	fError;
	
	PROCESSOR_STATE	*processor_state;
	SIGNUM			sig;

} DWSP;

static struct
{
	HARD_WSP	system;
	DWSP		local;
	
} dproc;

#define debugging_process	((VOID *)(&dproc.local))

/************************************************************************************
 *																					*
 *		basic routines (all this is very low level)									*
 *																					*
 ************************************************************************************/

VOID off_chip_suspend (WSP *regs)
{
	WSP *first_suspended;
		
	first_suspended=sys.reducer_info.off_chip_suspended_regs;

	sys.reducer_info.off_chip_suspended_regs = regs;
	
	if (first_suspended != NULL)
		set_prev_process(first_suspended,regs);

	set_prev_process(regs,NULL);
	set_next_process(regs,first_suspended);	
}

VOID off_chip_unsuspend (WSP *regs)
{
	WSP *prev, *next;
	
	prev=get_prev_process(regs);
	next=get_next_process(regs);
	
	if (prev != NULL)
		set_next_process(prev,next);	
	else
		sys.reducer_info.off_chip_suspended_regs=next;
	
	if (next != NULL)
		set_prev_process(next,prev);
}

VOID off_chip_activate (WSP *regs)
{
	/* deactivation of off-chip reducers is not needed, as they can only
	   deactivate themselves, i.e. when they are running in on-chip
	   memory
	*/
	
	if (sys.reducer_info.off_chip_first_regs == NULL)
		sys.reducer_info.off_chip_first_regs = regs;
	else
		set_next_process(sys.reducer_info.off_chip_last_regs,regs);
		
	sys.reducer_info.off_chip_last_regs=regs;
}

/********************************************************************************
 *																				*
 *		routines to start reducers												*
 *																				*
 ********************************************************************************/

/*	There are different routines needed while running a debugging process, as
	adding a process at the back of the scheduling list could destroy the back
	ptr as found by the debugging process. runp1 and low1 do not suffer from
	this problem as they keep the back pointer (the scheduling list is not empty,
	unless you are in the debugger, in which case the low prio debugging process
	does not run).
*/

VOID invalid_run_reducer(WORD pid)
{
	/* used in inactive state */
	
	fatal_error(False,"invalid_run_reducer");
}

static VOID active_run_reducer (WORD pid)
{
	WSP *regs;
	REDUCER_ENVIRONMENT *env;

	env=pid_to_env(pid);		
	set_active(env);
	
	regs=env_to_regs(env);
	
	if (regs <= sys.reducer_info.on_chip_last_regs)
	{
		/* this reducer can be started immediately */
		
		runp(regs,LowPrio);
	
	} else
	{
		off_chip_unsuspend(regs);
		off_chip_activate(regs);		
	}
}

static VOID adjust_debug_process(VOID)
{
	VOID *iptr;
	
	/* check iptr of (interrupted) debugging process */

	iptr=hard_IntIptr;

	if (iptr >= (VOID *)DebugProcess && iptr < (VOID *)DebugProcess4)
	{
		/*	it is not possible to encounter an empty low priority
			scheduling list as at least the scheduler will be in it.
			(When the scheduler has increased priority, its low prio
			counterpart will be the interrupted process (not the case
			here), so we can assume a non-empty scheduling list
		
			save special values for debugging process and change
			its iptr
		*/

		if (iptr == DebugProcess)
		{
			/* debug process still has to save Areg, do it for process */
			
			dproc.local.Areg		=hard_IntAreg;
			*hard_AregIntSaveLoc	=hard_IntBreg;
			*hard_BregIntSaveLoc	=hard_IntCreg;
		}
		
		if (iptr == DebugProcess3)
		{
			/* Areg has been saved but regs are incorrect */
			
			*hard_AregIntSaveLoc	=hard_IntBreg;
			*hard_BregIntSaveLoc	=hard_IntCreg;
		}
		
		/*	Areg has been saved, iptr should be 'DebugProcess2', save
			wsp of last process for debugging process. Note that there
			is no danger of overwriting a previous wsp, as the iptr would
			have been forwarded already in this case. Take care that the
			field before the wsp field does not contain valid information.
		*/

		savel(((WORD *)&dproc.local.wsp)-1);
		*hard_IptrIntSaveLoc=DebugProcess4;
	}
}

static VOID debug_run_reducer (WORD pid)
{
	WSP *regs;
	REDUCER_ENVIRONMENT *env;
	
	env=pid_to_env(pid);
	set_active(env);
	
	regs=env_to_regs(env);
	
	if (regs <= sys.reducer_info.on_chip_last_regs)
	{
		/*	see routine above. If the debugging process has been interrupted
			special measures must be taken. If it was not running, the scheduler
			will have been 'primed' already, else the debugging process will
			do this before it allows a context switch */

		if (hard_IntWsp==debugging_process)
			adjust_debug_process();
	
		/* no dangerous situation anymore, start reducer */
		
		runp(regs,LowPrio);
	
	} else
	{
		off_chip_unsuspend(regs);
		off_chip_activate(regs);
	}	
}

VOID simulate_run_reducer (WORD pid)
{
	WSP *regs;
	REDUCER_ENVIRONMENT *env;

	env=pid_to_env(pid);
	set_active(env);
	
	regs=env_to_regs(env);
	
	if (regs <= sys.reducer_info.on_chip_last_regs)
	{
		/*	see routine above, use simulated processor regs */

		if(sys.processor_state.low_front==NotProcessP)
			sys.processor_state.low_front=(ADDRESS)regs;
		else
			set_next_process(sys.processor_state.low_back,regs);
	
		sys.processor_state.low_back=(ADDRESS)regs;
	
	} else
	{
		off_chip_unsuspend(regs);
		off_chip_activate(regs);		
	}	
}

/********************************************************************************
 *																				*
 *		scheduler and idle time monitor											*
 *																				*
 ********************************************************************************/

static VOID schedule_reducers(VOID)
{
	/* assume high priority and at least one off-chip runnable reducer
	
	   to swap out processes, first look for stopped procs, then look for
	   suspended ones, and finally swap out running ones (if necessary).
	*/
	
	WSP						*in, *out;
	REDUCER_ENVIRONMENT		*in_env, *out_env;
	WSP						*schedlist[2];
	
	/* save current scheduling list and empty it */

	savel(schedlist);	
	stlf(NotProcessP);
	
	/* set next field of last active off-chip reducer right, there should be
	   at least one off-chip active reducer */

	set_next_process(sys.reducer_info.off_chip_last_regs,NULL);
	
	/****************************************************************************
	 *																			*
	 *		swap reducers for stopped (and empty) ones							*
	 *																			*
	 ****************************************************************************/

	for (in=sys.reducer_info.off_chip_first_regs,out=sys.reducer_info.on_chip_free_list;
		in != NULL && out != NULL;
		in=get_next_process(in),out=get_next_process(out))
	{		
		/* get stopped reducers out of free list, test for empty ones as well */

		in_env=get_process_env(in);

		/*	if the stopped reducer is not empty, swap useful regs (stacks etc.)
			out, otherwise only swap new reducer in */
		
		if (out->heap_start != NULL)
		{
			WSP *stopped_regs;
			
			out_env=get_process_env(out);

			/*	copy valid regs into swap space:
		   		a-stack, b-stack, heap-start, and heap-end.
		   		note that reducer should still be 'Active' */
		
			stopped_regs=sys.reducer_info.swap;
			move_idle_process(stopped_regs,out);
		
			/* set env of stopped reducer to swap space */
		
			out_env->registers=stopped_regs;
		
			/* push reducer on off-chip stopped list */

			set_next_process(stopped_regs,sys.reducer_info.off_chip_free_list);
			sys.reducer_info.off_chip_free_list=stopped_regs;
		}

		/* copy regs of active reducer to on-chip memory (+iptr!) */
		
		move_process(out,in);
		
		/* set env of activated reducer to on-chip regs and start reducer */
		
		in_env->registers=out;
		runp(out,LowPrio);
		
		/* set swap space to in */
	
		sys.reducer_info.swap=in;
	}

	/* adjust on-chip free list, off-chip active list updated later */
	
	sys.reducer_info.on_chip_free_list=out;

	/****************************************************************************
	 *																			*
	 *		swap reducers for suspended ones									*
	 *																			*
	 ****************************************************************************/
	
	if (in != NULL)
	{
		/* find an on-chip suspended reducer */

		for (out=sys.reducer_info.on_chip_first_regs;out<=sys.reducer_info.on_chip_last_regs;out=(WSP *)((REDUCER_LOCALS *)out+1))
		{
			WSP *suspended_regs;
			
			if (is_active(out_env=get_process_env(out)))
				continue;
			
			/* this process is suspended, swap it out */
								
			in_env=get_process_env(in);

			/*	copy suspended reducer locals (+iptr) into swap space (rest of
				'negative' locals is NOT saved!) */

			suspended_regs=sys.reducer_info.swap;
			move_process(suspended_regs,out);
	
			/* set env of suspended reducer to swap space */
	
			out_env->registers=suspended_regs;
	
			/* put reducer in off-chip suspended list */

			off_chip_suspend (suspended_regs);

			/* copy regs of active reducer to on-chip memory (+iptr!) */
	
			move_process(out,in);
	
			/* set env of activated reducer to on-chip regs and start reducer */
	
			in_env->registers=out;
			runp(out,LowPrio);
	
			/* set swap space to in */

			sys.reducer_info.swap=in;

			/* if no more processes to swap in, break */
			
			if ((in=get_next_process(in))==NULL)
				break;
		}
	}

	if ((out=schedlist[0])!=NotProcessP)
	{
		BOOL	debugging;
		WSP		*back, *saved_front, *saved_back;
		
		/****************************************************************************
		 *																			*
		 *		shortcut if no off-chip reducers anymore							*
		 *																			*
		 ****************************************************************************/

		if (in==NULL)
		{
			/*	the old scheduling list is not empty, put it in front of new
				scheduling list (which is also NOT empty, as all off-chip reducers
				have been swapped in). Note that the debugging process (if any) will
				be put at the start of the scheduling list automatically
			*/
				
			back=schedlist[1];
			savel(schedlist);
			stlf(out);
			set_next_process(back,schedlist[0]);
		
			/* the active list has become empty (back pointer need not be updated) */
	
			sys.reducer_info.off_chip_first_regs=NULL;
			return;
		}
		
		/****************************************************************************
		 *																			*
		 *		swap reducers for active ones (there are some)						*
		 *																			*
		 ****************************************************************************/

		/*	walk down the old transputer scheduling list and replace on-chip active
			processes by off-chip active processes until the end of the list is
			reached, or there are no more off-chip processes to swap in
	
			Note that low prio system processes are normally started at the front
			of the scheduling list, after which they increase priority and stop. This
			means that they are normally not encountered by the scheduler.
			One low prio system process that IS encountered by the scheduler
			is the debugger process (if any). This process should be placed
			were it belongs: at the very start of the scheduling list so
			that it is able to check the currently running (interrupted)
			process (in casu: the scheduler itself). All other low prio
			system processes should just be restarted.
		*/

		back=schedlist[1];
		set_next_process(back,NULL);

		debugging=False;
		
		if (out==debugging_process)
		{
			debugging=True;
			out=get_next_process(out);
			
			if (out==NULL)
			{
				/* empty scheduling list, there is nothing to swap out. Reinsert
				   debugging process and adjust off-chip scheduling list */
				
				runp1(debugging_process);
				sys.reducer_info.off_chip_first_regs=in;
				return;
			}
		}
		
		/*	at this point, 'out' is certainly not NULL, and neither is 'in', so
			the loop below will be executed at least once */
		
		for(saved_front=NULL;out!=NULL;out=get_next_process(out))
		{
			WSP *running_regs;
			
			if (out > sys.reducer_info.on_chip_last_regs)
			{				
				runp(out,LowPrio);
				continue;
			}
			
			/* running reducer found, replace it by another running reducer */
	
			out_env	=get_process_env(out);								
			in_env	=get_process_env(in);
	
			/* copy reducer locals (+ iptr) into swap space */
	
			running_regs=sys.reducer_info.swap;
			move_process(running_regs,out);
	
			/* set env of saved running reducer to swap space */
	
			out_env->registers=running_regs;
	
			/* put saved reducer in temporary list, so that it will not get swapped
			   in again */
		
			if (saved_front == NULL)
				saved_front = running_regs;
			else
				set_next_process(saved_back,running_regs);
			
			saved_back=running_regs;
	
			/* copy regs of active reducer to on-chip memory (+iptr!) */
	
			move_process(out,in);
	
			/* set env of activated reducer to on-chip regs and start reducer */
	
			in_env->registers=out;
			runp(out,LowPrio);
	
			/* set swap space to in and remove in from active list*/
	
			sys.reducer_info.swap=in;
			
			if ((in=get_next_process(in))!=NULL)
				continue;
			
			/* 'in' is NULL, so we are ready swapping */
			
			if ((out=get_next_process(out))!=NULL)
			{
				/* 'in' is NULL and the old scheduling list is not empty, insert
					it before reducers that have been swapped in */
				
				savel(schedlist);
				stlf(out);
				set_next_process(back,schedlist[0]);
			}
			
			/* if you are debugging, reinsert debug process */
			
			if (debugging)
				runp1(debugging_process);

			/* old off-chip list empty, make saved one the new one */
				 
			sys.reducer_info.off_chip_first_regs=saved_front;
			sys.reducer_info.off_chip_last_regs	=saved_back;
			
			/* ready */
			
			return;
		}
		
		/*	'in' is not empty, 'out' is, i.e. the old scheduling list has been
			replaced completely. Reinsert the debugging process if necessary.
			Note that at least one process has been swapped out: the saved
			list is not empty */
		
		if (debugging)
			runp1(debugging_process);
	
		/* adjust active list, i.e. set it to 'in' and append saved list of
		   old active reducers */
	
		sys.reducer_info.off_chip_first_regs=in;		
		set_next_process(sys.reducer_info.off_chip_last_regs,saved_front);
		sys.reducer_info.off_chip_last_regs=saved_back;
		
		return;
	}
	
	/*	shortcut when old scheduling list was empty (note that there was no
		debugging process either). Adjust active list */
	
	sys.reducer_info.off_chip_first_regs=in;
}

#if CHECK_DEADLOCK

static BOOL check_deadlock(VOID)
{
	/* assume high priority and no off-chip runnable reducer */
	
	WSP						*wsp;
	WSP						*schedlist[2];
	
	if (reducer_start_routine==invalid_run_reducer)
		return False;
		
	/* save current scheduling list and check it */

	savel(schedlist);	
	
	if ((wsp=schedlist[0])==NotProcessP)
		return True;

	while (wsp != schedlist[1])
	{
		/*	this is not the last process in the list, return false if this is
			a reducer */

		if (wsp <= sys.reducer_info.on_chip_last_regs)
			return False;
		
		wsp=get_next_process(wsp);
	}
	
	/* last process */
	
	return wsp > sys.reducer_info.on_chip_last_regs;
}
#endif

VOID scheduler(VOID)
{
	WORD sched_time,bias;
	WORD times_scheduled;
	WORD swap_time;
		
	/* This process performs 2 tasks. First of all it schedules all
	   process by swapping them in and out of internal memory. Secondly it keeps
	   track of the idle time. It should be started at high priority
	
	   This process will continuously deschedule itself. It counts the time it
	   does NOT run: if it takes a 'long' time to schedule this process again,
	   it is asumed another process has been running meanwhile. This process
	   is started in the beginning, but timing ia reset when all processors are
	   synchronised. The bias should be the time (in us) that is needed to
	   schedule once on a quite transputer (including change of prio).
	
	   Scheduling of other processes is performed when this process has been
	   scheduled a number of times AND the active off-chip list is not empty.
	   In order to performs safe scheduling, the process increases its
	   priority for this purpose.
	*/
	
	swap_time				=options.swap_time;
	times_scheduled			=swap_time;
	bias					=node_info.scheduling_bias;
	reducer_start_routine	=invalid_run_reducer;
	
	/* Get time for decreasing prio, low prio timing and increasing prio.
	   Time of descheduling (macro) is assumed small and ignored. In this
	   way the high prio clock timings can be made more accurate.
	*/
	
	for (;;)
	{
		/* Timing of active time. Timing is basically done by means of the high
		   prio clock. Note that this clock cycles in about 4295 seconds (say 70 min).
		   For the sake of simplicity (and efficiency) we use signed integers (which
		   come in handy for comparison). This means that this only works as long as
		   it does not take more than 2000 sec (35 min) to schedule this timing process.
		   It is VERY unlikely scheduling takes that long (as there are running
		   about 50 reducers at most (in on-chip memory), and high prio processes
		   should not take that much time. IF it happens, use some additional timer:
		   a high prio process that uses the timer and more bits to keep track of time
		*/
		
		sched_time=ldtimer();
		low();
		high();
		sched_time=minus(ldtimer(),plus(sched_time,bias));
		
		if (sched_time > 5)
		{
			/* some other process has been running meanwhile (this is considered
			   a safe margin: it should be possible to reschedule within 5 us) */
			
			i64incu32(processor_info.ticks_running,sched_time);
		}
		
		/* schedule processes when needed */
		
		if (sys.reducer_info.off_chip_first_regs == NULL)
		{
#if CHECK_DEADLOCK

			/*	there are no off-chip runnable processes, check if the
				scheduling list has reducers. If not: note a deadlock has
				ocurred
			*/
			
			if (check_deadlock())
			{
				runtime_warning(False,"deadlock");
				BREAK();
			}
#endif
			continue;
		}
		
		if (--times_scheduled!=0)
			continue;
		
		times_scheduled=swap_time;
		
		schedule_reducers();
	}
}

/********************************************************************************
 *																				*
 *		mode of execution														*
 *																				*
 ********************************************************************************/

VOID set_debug_run(BOOL debug)
{
	if (debug)
	{
		set_restart_routine(debug_run_reducer);
		set_iptr(debugging_process,DebugProcess);
		runp(debugging_process,LowPrio);
		
	} else
		set_restart_routine(active_run_reducer);
}

/********************************************************************************
 *																				*
 *		deallocation of reducers (to do)										*
 *																				*
 ********************************************************************************/ 
 
/********************************************************************************
 *																				*
 *		killing of processes													*
 *																				*
 ********************************************************************************/

VOID kill_reducers(VOID)
{
	/* this routine should be called at high priority at a 'safe' moment (i.e.
	   after increasing prio) It kills al (active) reducers
	*/

	WSP				*schedlist[2];
	
	/* save current scheduling list and kill all on-chip reducers */

	savel(schedlist);
	stlf(NotProcessP);
	
	/* kill all active off-chip reducers, so they won't get scheduled */
	
	sys.reducer_info.off_chip_first_regs	=NULL;
	
	/*	the active on-chip reducers have been killed already as the scheduling list
		has been emptied. Non-reducers should be put back however (except debugging
		process) ...
	*/
	
	if (schedlist[0]!=NotProcessP)
	{
		set_next_process(schedlist[1],NULL);
	
		while (schedlist[0] != NULL)
		{
			if (schedlist[0] > sys.reducer_info.on_chip_last_regs && schedlist[0]!=debugging_process)
				runp(schedlist[0],LowPrio);
			
			schedlist[0]=get_next_process(schedlist[0]);
		}
	}
}
