#include "remoteio.h"

#ifndef _TYPES_
#include "types.h"
#endif

#include "hardware.h"
#include "hostio.h"
#include "memory.h"
#include "hostserver.h"	/* only for exit and host_buffer... */
#include "loader.h"
#include "packets.h"

#ifdef __INCLUDE_STANDARD__

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>
#include <errno.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#endif

extern OPTIONS	options;

#define PORT 5678

/********************************************************************************
 *																				*
 *		global buffer and socket descriptor,									*
 *		at this moment only 1 remote server is supported						*
 *																				*
 ********************************************************************************/

static int			server_descriptor;		/* file descriptor of remote server */

static UCHAR		*io_buffer;
static UCHAR		*io_buffer_end;
static UCHAR		*io_buffer_free;
static BOOL			use_io_buffer;

/*	the code-loading routines use the io_buffer as well: the routine below
	returns this buffer. This is no problem as code loading and
	ordinary io will never take place simultaneously... */

/********************************************************************************
 *																				*
 *		low level routines														*
 *																				*
 ********************************************************************************/

static int remote_server(char *server_name)
{
	/* return socket descriptor of server on success */
	
	int sd,r;
	struct sockaddr_in sname;
	unsigned long host_addr;
	
	/* clear socket structure */
	
	bzero ((char *)&sname,sizeof (sname));

	/* first try to find host inet address by interpreting the given address
	   as an internet address of the form a.b.c.d */
	
	host_addr=inet_addr(server_name);
	
	if ((long)host_addr!=-1l)
		sname.sin_addr.s_addr=host_addr;
	else
	{
		/* no inet address found, try find it by name */
	
		struct hostent *server_host_ent;

		server_host_ent=GetHostByName (server_name);

		if (server_host_ent==NULL)
			return -1;
		
		bcopy ((char *)server_host_ent->h_addr,(char *)&sname.sin_addr,server_host_ent->h_length);
	}

	sname.sin_family=AF_INET;
	sname.sin_port=htons(PORT);
	
	/*	get a socket that can be used for connection with another
		socket */
	
	sd=Socket (AF_INET,SOCK_STREAM,0);
	
	if (sd<0)
		return -1;
	
	/* connect this socket to the socket address "sname" */
	
	r=Connect (sd,(struct sockaddr *)&sname,sizeof (sname));
	
	if (r<0)
	{		
		/* connection did not succeed, close socket and return */
		
		Close(sd);
		return -1;
	}

	/* connection succeeded, return socket descriptor */
	
	return sd;
}

static void remote_command(int remote_server, WORD command)
{
	int r;
	WORD message[2];
	
	message[0]=0;
	message[1]=command;
	
	r=Write(remote_server,(char *)&message,2*sizeof(WORD));
	
	if (r<0)
	{
		fprintf (stderr,"remote_command %d\n",Errno);
		return;
	}
}

static void release_remote_server (int remote_server)
{
	remote_command(remote_server,RemoteEnd);
}

static void flush_remote_commands(int remote_server)
{
	int r;
	
	/* flush io buffer to remote host */
	
	if (io_buffer_free == io_buffer)
		return;
	
	/* remote server exists and buffer not empty: write buffer to it */

	r=Write(remote_server,(char *)io_buffer,io_buffer_free-io_buffer);
	
	/* empty buffer */
	
	io_buffer_free=io_buffer;

	if (r<0)
	{
		fprintf (stderr,"flush_remote_io %d\n",Errno);
		return;
	}
}

/************************************************************************************
 *																					*
 *		exported routines															*
 *																					*
 ************************************************************************************/

VOID remote_clear_screen(VOID)
{
	if (server_descriptor >= 0)
		remote_command(server_descriptor,RemotePlotClearScreen);	
}

BOOL remote_connect (CHAR *name)
{
	if (server_descriptor >= 0)
	{
		io_buffer_free=io_buffer;
		return False;
	}
			
	server_descriptor=remote_server(name);
	
	return (server_descriptor >= 0);
}

BOOL remote_disconnect(VOID)
{
	if (server_descriptor < 0)
	{
		io_buffer_free=io_buffer;
		return False;
	}
	
	flush_remote_commands	(server_descriptor);	
	release_remote_server	(server_descriptor);
	Close					(server_descriptor);
	
	server_descriptor=-1;
		
	return True;
}

BOOL remote_end(VOID)
{
	/* flush buffer, to be called at end of program */
		
	if (server_descriptor < 0)
	{
		io_buffer_free=io_buffer;
		return False;
	}

	flush_remote_commands	(server_descriptor);
	use_io_buffer=False;

	return True;
}

BOOL remote_begin(VOID)
{
	/* indicate the buffer may be used, to be called at start of program */
	
	if (server_descriptor < 0)
	{
		io_buffer_free=io_buffer;
		return False;
	}
	
	use_io_buffer=True;
	return True;
}

typedef struct VREMOTE_PACKET2
{
	WORD	size;							/* source field of packet		*/
	WORD	remote_command;					/* command for remote server	*/
	BYTE	contents[VariableArraySize];	/* data for remote server		*/

} VREMOTE_PACKET2;


VOID handle_remote_server(WORD kind,WORD source,WORD size,VREMOTE_PACKET *packet)
{	
	VREMOTE_PACKET2 *p;
	
	/* test if remote server exists */
	
	if (server_descriptor < 0)
	{
		CHAR	*line;
		WORD	length;
	
		/* write prompt */

		fprintf(stdout,"not connected to server, give server name: ");
		fflush(stdout);
	
		line=(CHAR *)host_buffer;
		length=host_read_string(stdin,line,StringLength);
		
		/* use (the empty!) io buffer temporarily to scan the server name */
		
		if (!scan_word(&line,io_buffer,StringLength) ||
			!remote_connect((CHAR *)io_buffer))
		{
			fprintf(stderr,"connection failed\n");
			exit_ioserver();
			return;
		}
		remote_clear_screen();
	}
	
	/*	remote server exists, write packet to buffer (flush if necessary),
		the source field will be overwritten by the (new) length of the message */

	p=(VREMOTE_PACKET2 *)((WORD *)packet-1);
	p->size=size-sizeof(WORD);							/* size of contents */
	size+=sizeof(WORD);									/* size field must be sent too */
		
	if (use_io_buffer)
	{
		if (io_buffer_free+size > io_buffer_end)
			flush_remote_commands(server_descriptor);
		
		if (io_buffer_free+size > io_buffer_end)
		{
			int r;
			
			/* you can just as well write large packet directly */
			
			r=Write(server_descriptor,(char *)(&p->size),size);
		
			if (r<0)
				fprintf (stderr,"handle_remote_server %d\n",Errno);
			
			return;
		}
		
		/* message fits in buffer */
		
		move(io_buffer_free,(char *)(&p->size),size);
		io_buffer_free+=size;
		
		/* if buffer full, flush it */
		
		if (io_buffer_free==io_buffer_end)
			flush_remote_commands(server_descriptor);
			
	} else
	{
		/* write message directly, buffering is off */

		int r;
				
		r=Write(server_descriptor,(char *)(&p->size),size);
	
		if (r<0)
			fprintf (stderr,"handle_remote_server %d\n",Errno);
		
		return;
	}
}

VCODE_SEGMENT *code_packet_address(VOID)
{
	VQUEUE_PACKET *code_buffer;
	
	code_buffer=(VQUEUE_PACKET *)io_buffer;
	code_buffer->admin.system_packet=False;
	
	return &code_buffer->packet;
}

BOOL initialise_remote_io(VOID)
{
	WORD buff_size;

	buff_size=vsizeof(VQUEUE_PACKET)+options.packet_size;
	
	if ((io_buffer=push_mem(buff_size))==NULL)
		return False;
	
	io_buffer_free=io_buffer;
	io_buffer_end=io_buffer+buff_size;
	use_io_buffer=False;

	server_descriptor=-1;

	return True;	
}