/********************************************************************
 *																	*
 *	level 1 router routines (deadlock- and starvation-free)			*
 *	Written by Marco Kesseler										*
 *	february 1992													*
 *																	*
 *	Dep. of Mathematics and Computer Science						*
 *	University of Nijmegen											*
 *	Toernooiveld 1													*
 *	6525 ED Nijmegen												*
 *	The Netherlands													*
 *																	*
 ********************************************************************/

#include "router.h"

#ifndef _TYPES_
#include "types.h"
#endif

#include "messageno.h"
#include "memory.h"
#include "packets.h"
#include "macros.h"
#include "hardware.h"
#include "connect.h"

#ifdef __INCLUDE_STANDARD__

#include <stdlib.h>
#include <stdio.h>

#endif

/********************************************************************
 *																	*
 *	if the macro below is 1 a single write is used instead of 2		*
 *	for sending header and body of a message (where possible).		*
 *	This makes use of an undocumented feature of the transputer		*
 *	links...														*
 *																	*
 ********************************************************************/
 
#define SINGLE_WRITE	1

/********************************************************************************
 *																				*
 *		global structures used during initialisation							*
 *		part of it will be initialised by the router							*
 *																				*
 ********************************************************************************/

extern NET_INFO		net_info;
extern NODE_INFO	node_info;
extern OPTIONS		options;

/************************************************
 *												*
 *	constants									*
 *												*
 ************************************************/

#define NoCancelSent		(net_info.Nclass+1)

/* cancel_sent[i] is Nclass+1 if no cancel message has been sent over link i.
   Otherwise it indicates the class that has been canceled. If a message
   comes in all links for which cancel_sent is lower than the class of
   the received packet are woken. */

/* cancel_received[i] is -2 while a packet is being requested. Otherwise
   it indicates the 'latest' class that has been canceled. If a
   buffer is freed all links for which the received cancel has a higher
   class than the one of the freed buffer are woken to send a request. */


/************************************************
 *												*
 *	types										*
 *												*
 ************************************************/

/*	NOTE: DO NOT CHANGE ORDER OF FIELDS IN STRUCTURES !! i.e. the first field
	of the packet_link structure must be the sync_channel! */

typedef struct PACKET_LINK
{
	CHANNEL					sync_channel;			/* used to unblock output process						*/
	
	union {
		CHANNEL				*channel_p;				/* used to write to hard link							*/
		CHANNEL				channel;				/* not used at the moment								*/
	} out;
	
	union {
		CHANNEL				*channel_p;				/* used to read from hard link							*/
		CHANNEL				channel;				/* used to read from soft link							*/
	} in;
	
	BOOL					do_request;				/* True if request must be sent							*/
	WORD					do_packet;				/* class of packet to be sent							*/
	BOOL					do_signal;				/* True if signal must be sent							*/
	
	WORD					cancel_sent;			/* class of sent cancel									*/
	WORD					cancel_received;		/* class of received cancel								*/
	
	WORD					*next;					/* temporal ordering of packets							*/
	WORD					*prev;					/* temporal ordering of packets							*/
	
	WORD					lowest_priviliged;		/* (lowest) class this buffer has exclusive access to	*/
	BOOL					access_zero;			/* True if link may allocate zero class buffer			*/
	
	WORD					link_number;			/* number of link 										*/
	struct PACKET_LINK		*next_link;				/* a circular list of all running links							*/
	PACKET_BUFFER_HEADER	output_request;			/* buffer used for requests (when not piggy-backed)		*/
	PACKET_BUFFER_HEADER	output_header;			/* buffer used for output of headers					*/
	PACKET_BUFFER_HEADER	input_header;			/* buffer used for input of headers						*/
	
} PACKET_LINK;

typedef struct PACKET_CLASS {

	BOOL				reservable;
	WORD				count;
	struct PACKET_LINK	*exclusive_access;

} PACKET_CLASS;

/************************************************
 *												*
 *	global variables							*
 *												*
 ************************************************/

static PACKET_LINK*			link[VLinks];
static VPACKET_BUFFER		**buffer;
static PACKET_CLASS			**class;
static UWORD				*rout;		/* used for normal routing */
static UWORD				*urout;		/* used when unique paths are required */
static UWORD				*brout;		/* used for broadcast routing */
static UWORD				arrowhead;
static WORD					freehead;
static WORD					lowest_class_free;
static WORD					reservable_zero;
static PACKET_LINK			*round_robin;

/* exclusive_access_zero indicates which link may allocate a buffer of
   class 0. For buffers of higher classes another method is used to ensure
   fairness. At some point in time a link may want to allocate a
   buffer of class N or lower. Every other link that reserves a buffer
   looks if some link L after it (modulo number of links) needs this
   buffer too. If so the buffer of class C (<=N) is designated to this link
   by setting exclusive_access[C] to L. NULL indicates no link has the
   exclusive right to this buffer. Unlike the original algorithm a
   link may only have exclusive access to one buffer only. This is
   because a link will not always request a higher class after a
   lower one is cancelled. To ensure this lowest_priviliged is used.
   It is set to Nclass if a link has no priviliges. Otherwise it indicates
   the (lowest) class a link has priviliged access to. */

static BOOL router_running;

/********************************************************************
 *																	*
 *		low level routines											*
 *																	*
 ********************************************************************/

#define match(request,class)	((class) > (request)-2)

static VOID add_to_toa(WORD index)
{
	/* used for all sorts of routing */
	
	WORD c;
	WORD dest;
	UWORD dest_vector;
	WORD i;
	VPACKET_BUFFER	*b;
	
	b		= buffer[index];
	c		= b->header.class;
	dest	= b->header.destination;

	if(dest<0)
	{
		if (dest < ToAll)
		{
			/* unique path routeing */
			
			dest_vector=urout[dest-UniqBase];
		} else
		{
			/* Broadcast routeing. Set the replicate number in the buffer
			   to the number of bits set in the brout vector. The replicate
			   vector is used for broadcast routeing only and indicates the
			   number of times a buffer should be sent away from a certain
			   processor (i.e. maximum: the number of links) */
			
			dest_vector=brout[b->body.source];
			b->replicate=bitcnt(dest_vector);
		}
	} else
	{
		/* adaptive routeing */
		
		dest_vector=rout[dest];
	}
	
	for(i=0; i<VLinks; i++)
	{	if (vector_elem(dest_vector,i))
		{
			PACKET_LINK *link_ptr;
			WORD youngest;
			
			link_ptr=link[i];
			youngest					=link_ptr->prev[0];
			link_ptr->prev[index]		=youngest;
			link_ptr->next[index]		=0;
			link_ptr->prev[0]			=index;
			link_ptr->next[youngest]	=index;
							
			if (match(link_ptr->cancel_sent,c))
 			{
 				/* The new packet matches a previously sent request that has
 				   been canceled, act if a cancel of class+1 (or +2) has been
 				   sent and force a new request.
 				   
 				   The highest class of a packet that has to be sent
 				   is Nclass-2, so the routine below is allright
 				*/
 				
 				if (is_odd(c) != vector_elem(arrowhead,i))
 					link_ptr->cancel_sent=c+2;
 				else
 					link_ptr->cancel_sent=c+1;
 				link_ptr->do_signal=True;
 				signal_channel(link_ptr);
			}
		}
	}
}

static VOID add_to_broadcast_toa(WORD index)
{
	/* used for broadcast routing to other procs, dest should be 'ToOthers',
	   class should be 0 */
	
	UWORD dest_vector;
	WORD i;
	VPACKET_BUFFER	*b;
	
	b		= buffer[index];

	/* see add_to_toa as well */
				
	dest_vector=brout[b->body.source];
	b->replicate=bitcnt(dest_vector)-1;			/* do not send to source proc */
	
	/* send over all connected real links, but not over virtual one */
		
	for(i=0; i<NLinks; i++)
	{	if (vector_elem(dest_vector,i))
		{
			PACKET_LINK *link_ptr;
			WORD youngest;
			
			link_ptr=link[i];
			youngest					=link_ptr->prev[0];
			link_ptr->prev[index]		=youngest;
			link_ptr->next[index]		=0;
			link_ptr->prev[0]			=index;
			link_ptr->next[youngest]	=index;
			
			/* class is 0 */
			
			if (match(link_ptr->cancel_sent,0))
 			{
 				/* see routine above */
 				
 				if (vector_elem(arrowhead,i))
 					link_ptr->cancel_sent=2;
 				else
 					link_ptr->cancel_sent=1;
 				link_ptr->do_signal=True;
 				signal_channel(link_ptr);
			}
		}
	}
}


static VOID remove_from_toa(WORD index)
{
	/* used for point-to-point communications */
	
	WORD dest;
	WORD i;
	UWORD dest_vector;
	
	dest=buffer[index]->header.destination;

	if (dest<0)
		dest_vector =urout[dest-UniqBase];
	else
		dest_vector	=rout[dest];
	
	for(i=0; i<VLinks; i++)
	{	if (vector_elem(dest_vector,i))
		{
			PACKET_LINK *link_ptr;
			WORD next,prev;

			link_ptr=link[i];
			next = link_ptr->next[index];
			prev = link_ptr->prev[index];
			link_ptr->prev[next]=prev;
			link_ptr->next[prev]=next;
		}
	}
}

static VOID remove_from_single_toa(WORD index, PACKET_LINK *link)
{
	/* used for broadcast routeing */
	
	WORD prev,next;

	next = link->next[index];
	prev = link->prev[index];
	link->prev[next]=prev;
	link->next[prev]=next;
}

static WORD find_packet(PACKET_LINK *link_ptr, WORD request)
{
	WORD try;
	WORD *next_list;

	next_list=link_ptr->next;
	
	try = next_list[0];
	while (try && (!match(request,buffer[try]->header.class)))
		try = next_list[try];
	return try;
}

static VOID release(WORD c) 	/* c is class to be released */
{
	WORD low_class;
	WORD i;

	/* adjust counts of class administration */
	
	low_class = c;
	while ((low_class) && (class[low_class]->count))
	{	class[low_class]->count--;
		low_class--;
	}
	
	/* adjust reservability of class administration */
	
	if (low_class)
	{
		/* free non-zero class buffer, the privilege to allocate this buffer
		   class has been passed earlier, as this routine cannot determine who
		   has allocated the buffer, see reserve_storage_unit as well */
		
		class[low_class]->reservable=True;
		if (low_class < lowest_class_free)
			lowest_class_free=low_class;
		else
			low_class=lowest_class_free;
	}
	else
	{
		/* free zero class buffer */
		
		if (reservable_zero < VLinks)
		{
			/* Pass on the privilege to allocate a zero class buffer. */
			 
			round_robin=round_robin->next_link;
			while (round_robin->access_zero)
				round_robin=round_robin->next_link;
			round_robin->access_zero=True;
		}
		reservable_zero++;
	}
	
	/* See some processes are waiting for such a buffer as this, and wake them
	   up if so. */
	
	for(i=0; i<VLinks; i++)
	{
		PACKET_LINK *link_ptr;

		link_ptr=link[i];
		if (is_odd(low_class) != vector_elem(arrowhead,i))
		{	if (link_ptr->cancel_received>(low_class))
			{	link_ptr->do_request=True;
				signal_channel(link_ptr);
			}
		} else
		{	if (link_ptr->cancel_received>(low_class+1))
			{	link_ptr->do_request=True;
				signal_channel(link_ptr);
			}
		}
	}
}

static VOID exchange(WORD low_class, WORD high_class)
{
	WORD c;
	
	/* exchange the low class buffer for a higher class buffer */
	
	c = high_class;
	
	while ((c>low_class) && (!class[c]->reservable))
	{	class[c]->count++;
		c--;
	}
	
	if (c > low_class)
	{	class[c]->reservable=False;
		release(low_class);
	}
}

static VOID set_next_fair(PACKET_LINK *link_ptr, WORD c)
{
	PACKET_LINK *next_ptr;
	
	/* This routine should only be used after a buffer of class c > 0 has been
	   allocated by this link. It passes on the privilege to allocate the buffer */
	
	/* find a link that needs this privilege */
	
	next_ptr=link_ptr->next_link;
	while(next_ptr != link_ptr)
	{
		if (c < next_ptr->lowest_priviliged)
		{
			/* this is a better privilege */

			if (next_ptr->cancel_received == NoCancelReceived)
			{
				/* this link has not (yet) received a cancel. This means it is
				   currently requesting a packet. give it the privilige, but make
				   sure it throws it away on reception of a cancel if it is too
				   high.
				*/
				
				class[c]->exclusive_access=next_ptr;
				class[next_ptr->lowest_priviliged]->exclusive_access=NULL; /* forget old privilige, this one's better */
				next_ptr->lowest_priviliged=c;
				return;
			}
			
			/* this link has received a cancel, so if the privilege is lower
			   than the cancel, the link gets the privilege. Otherwise it does
			   not need the privilige, and the next link may be tried... */
			
			if (is_odd(c) != vector_elem(arrowhead,next_ptr->link_number))
			{
				if (next_ptr->cancel_received>c)
				{
					class[c]->exclusive_access=next_ptr;
					class[next_ptr->lowest_priviliged]->exclusive_access=NULL; /* forget old privilige, this one's better */
					next_ptr->lowest_priviliged=c;
					return;
				}
							
			} else
			{	if (next_ptr->cancel_received>(c+1))
				{
					class[c]->exclusive_access=next_ptr;
					class[next_ptr->lowest_priviliged]->exclusive_access=NULL;	/* forget old privilige, this one's better */
					next_ptr->lowest_priviliged=c;
					return;
				}
			}
		}
		next_ptr=next_ptr->next_link;
	}
}

static WORD reserve_storage_unit(PACKET_LINK *link_ptr)
{
	WORD c;
	
	if (link_ptr->access_zero)
	{
		/* Reserve a zero class buffer and throw away any privilege as this
		   is the best buffer that can be allocated
		*/
		
		reservable_zero--;
		if (reservable_zero < VLinks)
			link_ptr->access_zero=False;

		/* Throw away privilege */
		
		class[link_ptr->lowest_priviliged]->exclusive_access=NULL;
		link_ptr->lowest_priviliged=net_info.Nclass;
		return 0;
	}


	/* A zero class buffer cannot be allocated, so try to allocate a non-zero
	   class buffer. First find real lowest class that is free */
	
	for(; lowest_class_free<net_info.Nclass; lowest_class_free++)
		if (class[lowest_class_free]->reservable)
			break;
	
	for(c=lowest_class_free; c<link_ptr->cancel_received; c++)
	{
		PACKET_CLASS *class_ptr;

		/* class should be reservable AND reservable by this particular link */
		
		class_ptr=class[c];
		if (class_ptr->reservable)
		{
			/* This buffer is reservable, see if privileges allow to reserve it */
			
			if (class_ptr->exclusive_access==NULL)
			{
				/* Anyone may allocate this buffer. Note that you could have a
				   privilege for another (lower) class buffer that has not been freed
				   yet. Keep this privilege. Otherwise fairness is not guaranteed.
				   Pass on the privilege for the current buffer.
				*/
				
				class_ptr->reservable=False;
				set_next_fair(link_ptr,c);
				return c;
			}
			
			if (class_ptr->exclusive_access==link_ptr)
			{
				/* you are using your privilege, throw it away and pass it on */
				
				class_ptr->exclusive_access=NULL;
				link_ptr->lowest_priviliged=net_info.Nclass;

				class_ptr->reservable=False;
				set_next_fair(link_ptr,c);
				return c;
			}
		}
	}
	
	/* No buffer available within non-canceled area. Keep any privilege
	   you have got and wait for suitable buffer to be freed.
	*/

	return -1;
}

static WORD setup_request(PACKET_LINK *link_ptr)
{
	/* this routine is used by the output process. If possible
	   requests are combined with RoutCancel/RoutSignal/PACKET messages */
	
	WORD c;
	
	link_ptr->do_request=False;
	c = reserve_storage_unit(link_ptr);	

	if (c < 0)
		return -1;
	
	if (is_odd(c) != vector_elem(arrowhead,link_ptr->link_number))
	{
		/* is_even(class) <=> arrowhead[link]
		   this means a request for this class may be sent */
		
		link_ptr->cancel_received=NoCancelReceived;
		link_ptr->output_request.class = c;
		return c;
	}
	
	if (c < net_info.Nclass-1)
	{
		WORD high_class;

		/* the class of the reserved storage unit isn't right, it should
		   be increased by 1 (if this is possible). */
		
		link_ptr->cancel_received=NoCancelReceived;
		high_class=c+1;
		link_ptr->output_request.class = high_class;

		if (class[high_class]->reservable)
		{	class[high_class]->reservable=False;
			if (c)
			{	class[c]->reservable=True;
				return high_class;
			}
			else
			{	if (reservable_zero < VLinks)
				{	round_robin=round_robin->next_link;
					while (round_robin->access_zero)
						round_robin=round_robin->next_link;
					round_robin->access_zero=True;
				}
				reservable_zero++;
				return high_class;
			}
		} else
		{	class[high_class]->count++;
			return high_class;
		}
	}

	/* class==Nclass-1 the class can't be increased */
	
	link_ptr->cancel_received=net_info.Nclass-1;		/* maximum class for this link is Nclass-2 */
	release(c);
	return -1;
}

static VOID emit_packet(PACKET_LINK *l)
{
	/* this routine outputs a packet on a link */
	
	WORD request;
	WORD i;
		
	request=l->do_packet;
	l->do_packet=NoRequest;
	
	if (i=find_packet(l,request))
	{
		VPACKET_BUFFER *b;
		WORD c, dest;

		b=buffer[i];
		dest= b->header.destination;

		if (dest >= 0 || dest < ToAll)
		{
			/* point to point communication */
			
			remove_from_toa(i);
	
			c=b->header.class;
			if (is_odd(c) != is_odd(request))
				b->header.class=c+1;
			
			if(l->do_request)
				b->header.request=setup_request(l);
			else
				b->header.request=-1;
			
#if SINGLE_WRITE
			chwrite(l->out.channel_p,&b->header,sizeof(PACKET_BUFFER_HEADER)+b->header.length);
#else
			chwrite(l->out.channel_p,&b->header,sizeof(PACKET_BUFFER_HEADER));
			if (b->header.length>0)
				chwrite(l->out.channel_p,&b->body,b->header.length);
#endif
			
			link[0]->next[i] = freehead;
			freehead = i;
	
			release(c);
			return;
			
		} else
		{
			/* broadcast, don't change buffer header, but use copy instead */
			
			PACKET_BUFFER_HEADER bhead;
			
			remove_from_single_toa(i,l);
			
			bhead.kind			=b->header.kind;
			bhead.length		=b->header.length;
			bhead.destination	=b->header.destination;
			
			c=b->header.class;
			if (is_odd(c) != is_odd(request))
				bhead.class=c+1;
			else
				bhead.class=c;
			
			if(l->do_request)
				bhead.request=setup_request(l);
			else
				bhead.request=-1;

			/* single write not possible here as header and body are
			   placed in different locations */
			
			chwrite(l->out.channel_p,&bhead,sizeof(PACKET_BUFFER_HEADER));
			if (bhead.length>0)
				chwrite(l->out.channel_p,&b->body,bhead.length);
			
			/* if last replica, release buffer, else keep it */
			
			if (--(b->replicate) == 0)
			{
				/* put buffer in freelist */
				
				link[0]->next[i] = freehead;
				freehead = i;
	
				/* release class c */
				
				release(c);
			}
			return;
		}
	}
	else
	{
		l->cancel_sent				=request;
		l->output_header.kind		=RoutCancel;
		l->output_header.class		=request;
		
		if (l->do_request)
			l->output_header.request=setup_request(l);
		else
			l->output_header.request=-1;
		
		chwrite(l->out.channel_p,&l->output_header,sizeof(PACKET_BUFFER_HEADER));
		return;
	}
}

/********************************************************************
 *																	*
 *		router processes											*
 *																	*
 ********************************************************************/

static VOID input_process(PACKET_LINK *l)
{
	WORD max_class;
	
	max_class=net_info.Nclass;
		
	while (router_running)
	{
		WORD message_kind;

		chread(l->in.channel_p,&l->input_header,sizeof(PACKET_BUFFER_HEADER));
		
		/* a request may be piggy backed on a RoutSignal/RoutCancel/PACKET
		   message */
		
		message_kind=l->input_header.kind;
		
		if (message_kind < 0)	/* protocol message */
		{
			if (message_kind==RoutRequest)
			{
				/* it could be you've already started sending a signal
				   This is no problem: the other side will ignore it. Just
				   search for a packet and send it. */
				
				l->do_packet=l->input_header.class;
				l->do_signal=False;
				l->cancel_sent=max_class;	/* prohibits signals to be sent */
				signal_channel(l);
				
				continue;
				
			} else if (message_kind==RoutCancel)
			{
				/* a cancel message should always be the response to a
				   request. */
				   
				l->cancel_received=l->input_header.class;
				
				/* throw away privilige if it is higher than the canceled
				   class. If it is lower, the process will use it */

				if (l->lowest_priviliged >= (l->cancel_received - 1))
				{
					/* Cancel received for class c. This means that class
					   c and class c-1 can't be used (c-1 will be exchanged for
					   c), so priviliges >= c-1 are useless */
					
					class[l->lowest_priviliged]->exclusive_access=NULL;
					l->lowest_priviliged=net_info.Nclass;
				}

				release(l->output_request.class);
				
				if (l->input_header.request>=0)
				{
					l->do_packet=l->input_header.request;
					l->do_signal=False;
					l->cancel_sent=max_class;	/* prohibits signals to be sent */
					signal_channel(l);
				}

				continue;
			} else if (message_kind==RoutSignal)
			{
				/* a signal indicates that the other side wants to send
				   something new of higher class than before. The class of
				   the signal indicates the maximum class that is needed
				   to receive the message. Note that you may have started to
				   send a request already. In this case the signal should
				   be ignored */
				
				if(l->cancel_received>=0) /* if you're not sending a request already */
				{
					l->do_request=True;
					l->cancel_received=l->input_header.class;
					signal_channel(l);
				}
				
				if (l->input_header.request>=0)
				{	l->do_packet=l->input_header.request;
					l->do_signal=False;
					l->cancel_sent=max_class;	/* prohibits signals to be sent */
					signal_channel(l);
				}

				continue;
			} else if (message_kind==RoutExit)
			{
				kill_router();
				continue;
			}
			continue;
		} else
		{
			VPACKET_BUFFER *b;
			WORD i;

			/* default a packet should be the answer to a request.
			   Receive it and send a new request. To speed up
			   execution first wake up process to send request */
			
			if (l->input_header.request>=0)
			{	l->do_packet=l->input_header.request;
				l->do_signal=False;
				l->cancel_sent=max_class;	/* prohibits signals to be sent */
			}

			l->do_request=True;
			l->cancel_received=max_class;	/* any request is right */
			signal_channel(l);
			
			exchange(l->output_request.class,l->input_header.class);
			i = freehead;
			freehead = link[0]->next[i];	
	
			b=buffer[i];
			b->header=l->input_header;
			if (l->input_header.length > 0)
				chread(l->in.channel_p,&b->body,l->input_header.length);

			add_to_toa(i);
			
			continue;
		}
	}
}

static VOID output_process(PACKET_LINK *link_ptr)
{
	PACKET_LINK *l;
	WORD c;
		
	l=link_ptr;
	
	while (router_running)
	{		
		if (l->do_packet>=0)
		{	emit_packet(l);
			continue;
		}
				
		if (l->do_signal)
		{	l->do_signal = False;
			l->output_header.kind = RoutSignal;
			l->output_header.class = l->cancel_sent;
			
			if (l->do_request)
				l->output_header.request=setup_request(l);
			else
				l->output_header.request=-1;
			
			chwrite(l->out.channel_p,&l->output_header,sizeof(PACKET_BUFFER_HEADER));
			continue;
		}
		
		if (l->do_request)
		{	c = setup_request(l);
	
			if (c >= 0)
			{	/* the request field isn't initialised for request messages */
	
				l->output_request.kind = RoutRequest;
				chwrite(l->out.channel_p,&l->output_request,sizeof(PACKET_BUFFER_HEADER));
			}
			continue;
		}

		/* nothing more to send, fall asleep */
		
		wait_channel(l);
	}
	node_info.neighbour[l->link_number]=LStopped;
}

/************************************************************************************
 *																					*
 *		special link processes. These should be used on the 'borders' of the		*
 *		network. This means that (a) packets of any class must be output within		*
 *		a finite amount of time and (b) only packets of class 0 are allowed to be	*
 *		read from a link.															*
 *																					*
 ************************************************************************************/

/* Normally the link is asleep (i.e. the output process has sent a cancel and
   no request has been received yet). In theory, there are 2 cases the output
   process has to be woken up: when a buffer is freed at the other side, or
   when a packet arrives. The first case is not considered: if a packet cannot
   be accepted the outputting routine itself should suspend. Thus, the only
   case the output process is woken up, is when a packet arrives. As the link
   is asleep the 'do_signal' boolean will be set and the link woken up. Normally
   a signal-request-packet-request-... is the result. In this case the signals
   and the requests are rubbish: just send packets as long as there are any.
   If no packets can be output anymore, just fall asleep again (i.e. perform
   a 'cancel')
   
   In addition to the normal 'output' sequence this routine should also wake up
   the corresponding input process if it should issue a request...
*/

static VOID border_output_process(PACKET_LINK *l,CHANNEL *i_chan,VOID (*out)(VPACKET_BUFFER *),BOOL (*test)(VOID))
{
	/* the 'test' parameter is not used for the output process */
		
	while (router_running)
	{
		while (l->do_signal)
		{
			WORD i;

			/* wake up input process if necessary (don't send request yourself) */
			
			if (l->do_request)
				signal_channel(i_chan);
		
			if (i=find_packet(l,0))	/* find any packet */
			{
				VPACKET_BUFFER *b;
				WORD c,dest;

				b=buffer[i];
				dest= b->header.destination;

				if (dest >= 0 || dest < ToAll)
				{
					/* point-to-point communications */
					
					c=b->header.class;
					remove_from_toa(i);
					
					(*out)(b);
				
					link[0]->next[i] = freehead;
					freehead = i;
		
					release(c);
					continue;
					
				} else
				{
					/* broadcast, see emit_packet as well */
					
					c=b->header.class;
					remove_from_single_toa(i,l);
					
					(*out)(b);
				
					if (--(b->replicate) == 0)
					{
						link[0]->next[i] = freehead;
						freehead = i;
		
						release(c);
					}
					continue;
				}
			} else
			{
				l->do_signal=False;
				l->cancel_sent=0;		/* no packet available, wake up if one arrives */
				break;
			}
		}
		
		/* wake up input process if necessary (don't send request yourself) */

		if (l->do_request)
			signal_channel(i_chan);
		
		wait_channel(l);
	}
}

/* Normally the input process is asleep (i.e. a request has been answered by a
   cancel). As only 0 class messages may be submitted from 'outside' to the
   network, this means that (a) no 0 class buffers were available and a higher
   class has been canceled or (b) the other side has no (0 class) packets
   available. When (a) is solved the input process is woken by an
   output process. When (b) is solved the input process is woken from 'outside'.
   Once woken the input proces tries to input as many (0 class) packets as possible.
   If it can't input another packet or when no 0 class buffers are available it
   falls asleep. Note the input process waits on a special channel (as no
   hardware channel is available).
*/

static VOID border_input_process(PACKET_LINK *link_ptr,CHANNEL *i_chan,VOID (*in)(VPACKET_BUFFER *),BOOL (*test)(VOID))
{
	PACKET_LINK *l;
	WORD try_insert_message;
	
	l=link_ptr;
	try_insert_message=0; /* you haven't tried to insert a message yet */
	
	while (router_running)
	{
		while (l->do_request)
		{

			/* this routine tries to insert a packet into the network. If the
			   number of free class 0 buffers is small this means
			   the network is quiet full. In this case the insertion of the
			   message is delayed to allow the network to reach a lower
			   load. This is done by releasing the buffer, thereby allowing
			   other links to reserve it. To ensure fairness this is done
			   only a number of times. As soon as a buffer of class 0 is
			   released you're woken up again. Note that there's always
			   some other buffer of class 0 in use if the number of free class 0
			   buffers is small */
	
			WORD c;
			WORD i;
			VPACKET_BUFFER *b;
	
			c=reserve_storage_unit(l);
			l->do_request=False;
	
			if (c<0)
				break;

			if (!(*test)())						/* there's nothing to input */
			{
				l->cancel_received=0;			/* cancel received of class 0 */
				release(c);
			    break;
			}

			if (c>0)
			{
				l->cancel_received=1;			/* only accept class 0 packets */
				release(c);
				break;
			}
	
			/* a buffer of class 0 is available */
	
			if (reservable_zero < VLinks && try_insert_message < VLinks-reservable_zero)
			{
				/* not many class 0 buffers available, so delay insertion
				   of message */
				
				try_insert_message++;
				
				/* act if temporary no cancel received, to avoid waking up this process
				   by the following release... */
				
				l->cancel_received=NoCancelReceived;
				release(c);
				
				/* now reset cancel_received: wake up if class 0 buffer is freed */
				
				l->cancel_received=1;
				break;
			}
	
			try_insert_message=0;
			i = freehead;
			freehead = link[0]->next[i];	
			
			b=buffer[i];
			(*in)(b);	
			b->header.class=0;

			/* announce the arrival of the packet, and take care of special
			   broadcast messages */
			
			if (b->header.destination==ToOthers)
				add_to_broadcast_toa(i);
			else
				add_to_toa(i);
			
			l->do_request=True;		/* and request the next packet */
			l->cancel_received=1;	/* request for class 0 is right */
			
			continue;
		}
		wait_channel(i_chan);
	}
}

/********************************************************************************
 *																				*
 *	initialisation																*
 *																				*
 ********************************************************************************/

static PACKET_LINK* allocate_link_data(WORD link_number)
{
	PACKET_LINK	*l;
	WORD		*next_list;
	WORD		*prev_list;
	
	if(!(l=push_mem(sizeof(PACKET_LINK))))
		return NULL;
	if(!(next_list=push_mem((net_info.Nbuffs+1)*sizeof(WORD))))
		return NULL;
	if(!(prev_list=push_mem((net_info.Nbuffs+1)*sizeof(WORD))))
		return NULL;
	
	chreset(&l->sync_channel);
	
	if (link_number < NLinks)
	{	l->out.channel_p			=link_to_chanout(link_number);
		l->in.channel_p				=link_to_chanin(link_number);
	} else
	{	chreset(&l->out.channel);
		chreset(&l->in.channel);
	}
	
	/* default setting prohibit use of links without processes */
	
	l->do_request				=False;					/* do not send request	*/
	l->do_signal				=False;					/* do not send signal	*/
	l->do_packet				=NoRequest;				/* do not output packet	*/
	l->cancel_sent				=NoCancelSent;			/* no cancel sent, prohibits signals */
	l->cancel_received			=NoCancelReceived;		/* no cancel received, prohibits requests */
	l->next						=next_list;
	l->prev						=prev_list;
	l->access_zero				=True;
	l->lowest_priviliged		=net_info.Nclass;		/* no priviliges for link */
	l->link_number				=link_number;
	
	return l;
}

static VPACKET_BUFFER **allocate_buffers(VOID)
{
	WORD i;
	VPACKET_BUFFER **buffers;
	
	if(!(buffers=push_mem((net_info.Nbuffs+1)*sizeof(VPACKET_BUFFER *))))
		return NULL;
	
	buffers[0]=NULL;					/* buffer 0 is unused */
	for (i=1; i<=net_info.Nbuffs; i++)
	{	if(!(buffers[i]=push_mem(vsizeof(VPACKET_BUFFER)+options.packet_size)))
			return NULL;
	}
	
	return buffers;
}

static PACKET_CLASS **allocate_classes(VOID)
{
	WORD i;
	PACKET_CLASS **classes;
	
	if(!(classes=push_mem((net_info.Nclass+1)*sizeof(PACKET_CLASS *))))
		return NULL;
	
	for (i=0; i<=net_info.Nclass; i++)
	{	if(!(classes[i]=push_mem(sizeof(PACKET_CLASS))))
			return NULL;
		
		classes[i]->reservable			=True;
		classes[i]->count				=0;
		classes[i]->exclusive_access	=NULL; /* any link may try to access buffer of this class */
	}
	
	return classes;
}

static BOOL initialise_routtable(VOID)
{
	WORD i,j;
	
	if(!(rout=push_mem((net_info.size+1)*sizeof(UWORD))))
		return False;
	if(!(urout=push_mem((net_info.size+1)*sizeof(UWORD))))
		return False;
	if(!(brout=push_mem((net_info.size+1)*sizeof(UWORD))))
		return False;
		
	for (i=0; i<net_info.size+1; i++)
	{
		rout	[node_info.rout_data[i].proc]=node_info.rout_data[i].outlinks;
		brout	[node_info.rout_data[i].proc]=node_info.rout_data[i].outblinks;
	}
	
	for (i=0; i<net_info.size+1; i++)
	{	urout[i]=0;
		for (j=0; j<VLinks; j++)
		{	if(vector_elem(rout[i],j))
			{	set_vector_elem(urout[i],j);
				break;
			}
		}
	}
	return True;
}

static WORD next_running_link(WORD i)
{
	WORD next;
	
	next=i+1;
	if (next > NLinks)
		next=0;
	
	while (next != i)
	{
		/* next <= NLinks */
		
		if (next == NLinks)
			return next;			/* virtual links always connected */
		
		/* next < NLinks */
		
		if (node_info.neighbour[next]>=0)
			return next;			/* neigbour is connected */
		
		/* neighbour is not connected, try next */
		
		if (++next > NLinks)
			next=0;
	}
	
	/* found yourself, error */
	
	return i;
}

BOOL initialise_router(VOID)
{
	WORD i;
	router_running=True;

	/* impose some acyclic graph on the network... */
	
	arrowhead = 0;	
	for (i=0;i<NLinks;i++)
	{
		if (node_info.neighbour[i] <=0 || node_info.id < node_info.neighbour[i])
			set_vector_elem(arrowhead,i);
	}
	set_vector_elem(arrowhead,NLinks); /* packets come in according to graph */

	/* initialise routing tables... */
	
	if(!initialise_routtable())
		return False;
	
	net_info.Nclass	=net_info.diameter+1;		/* 1 extra buffer needed: internal links increase diameter */
	net_info.Nbuffs	=options.n_buffers+net_info.Nclass;
	
	for (i=0; i<VLinks; i++)
	{	if(!(link[i]=allocate_link_data(i)))
			return False;
	}
	for (i=0; i<=NLinks; i++)
		link[i]->next_link=link[next_running_link(i)];
	
	if (!(buffer=allocate_buffers()) ||
		!(class=allocate_classes()))
		return False;
		
	/* empty outgoing lists */
	
	for(i=0;i<VLinks;i++)
	{	link[i]->next[0]	=0;
		link[i]->prev[0]	=0;
	}
	
	/* init free list */
	
	for(i=1;i<net_info.Nbuffs;i++)
		link[0]->next[i]=i+1;
	link[0]->next[net_info.Nbuffs]=0;
	freehead=1;
	
	lowest_class_free		=1;
	reservable_zero			=net_info.Nbuffs+1-net_info.Nclass;
	
	/* the internal link is always available */
	
	round_robin				=link[NLinks];
			
	return True;
}

/********************************************************************************
 *																				*
 *		router interface														*
 *																				*
 ********************************************************************************/

BOOL start_virtual_link(WORD linkno, VOID (*out) (VPACKET_BUFFER *), WORD out_stack_size, VOID (*in) (VPACKET_BUFFER *), WORD in_stack_size, BOOL (*test)(VOID))
{
	/* this routine starts up processes for the indicated virtual link
	   this is always a 'border' link, as virtual links provide access from
	   outside to the network. The virtual links should be started before
	   the router is started, but after its data structures have been initialised.
	   
		-	out: a packet (in buff_ptr) has arrived for the current transputer.
		This routine should read the buffer and take appropriate actions. The buffer
		will be reclaimed by the router immediately afterwards.
	
		-	in: the router will execute this routine when a packet is available
		(see routine above) and a buffer (buff_ptr) is free for this packet. The routine
		should fill this buffer with the packet.
	
		-	test: by means of this routine the router tests wheter a
		packet is available. If it will use in_packet to get the packet if a buffer
		is available.
	
	*/
	
	link[linkno]->cancel_sent=0;				/* wake up (signal) if any packet arrives */
	link[linkno]->cancel_received=1;			/* only request for class 0 may be sent */

	/* allocate a channel to be used for access to the router */

	chreset(&link[linkno]->in.channel);

	/* start processes on virtual link */

	if (!push_process(border_input_process,in_stack_size,4,link[linkno],&link[linkno]->in.channel,in,test))
		return False;
	if (!push_process(border_output_process,out_stack_size,4,link[linkno],&link[linkno]->in.channel,out,test))
		return False;
	
	procresch();
	
	return True;
}

BOOL start_router(VOID)
{
	WORD i;
	
	for(i=0; i<NLinks; i++)
	{	if (node_info.neighbour[i]>0)
		{
			/* there's a transputer at the other side (not the host) */
			
			link[i]->cancel_sent=0;					/* wake (signal) if any packet arrives */
			link[i]->cancel_received=0;				/* force requests to be sent (on signal) */
			
		}
	}
		
	for (i=0; i<NLinks; i++)
	{
		if (node_info.neighbour[i]>0)
		{
			/* neighbour is a 'normal' transputers */
			
			if (!push_process(output_process,KiloByte,1,link[i]))
				return False;
			
			if (!push_process(input_process,KiloByte,1,link[i]))
				return False;
		}
	}
	
	procresch();
	
	return True;
}

VOID wake_virtual_link(WORD linkno)
{
	PACKET_LINK *l;
	
	l=link[linkno];
	
	l->do_request=True;		/* internal link should send request for this packet */
	l->cancel_received=1;	/* and it should wait for buffers of class 0 */
	
	signal_channel(&l->in.channel);
}

/********************************************************************
 *																	*
 *		routines used to kill the router correctly (more or less)	*
 *																	*
 ********************************************************************/

static VOID kill_input_processes(VOID)
{
	WORD i;
	
	/* an input process is always waiting on a link or commmunicating
	   over it, so a resetch is enough to kill the process */
	
	for (i=0; i<NLinks; i++)
	{	if(node_info.neighbour[i]>0)
			chreset(link_to_chanin(i));
	}
}

static VOID kill_output_processes(VOID)
{
	WORD i;
	
	/* an output process is either communicating, or waiting for an
	   assignment. In the first case it will be killed by resetting
	   the channel. In the second case it kills itself after unblocking
	   it: it will find router_running to be False. This routine will
	   only kill output processes that won't be needed anymore to
	   kill the rest of the network (i.e. if no transputer has been
	   booted via this link).
	*/
	
	for (i=0;i<NLinks;i++)
	{	if(node_info.neighbour[i]>0)
		{	if (node_info.nbooted[i] == 0)
			{	while(node_info.neighbour[i]>0)
				{
					WSP *proc_desc;
					
					proc_desc=chreset(link_to_chanout(i));
					
					if (proc_desc!=NotProcessP)
						node_info.neighbour[i]=LStopped;
					else
					{	signal_channel(link[i]);
						procresch();
					}
				}
			}
		}
	}
}

static VOID kill_sons(VOID)
{
	WORD i;
	
	for(i=0;i<NLinks;i++)
	{	if(node_info.neighbour[i]>0)
		{	if(node_info.nbooted[i]>0)
			{
				while(node_info.neighbour[i]>=0)
				{	signal_channel(link[i]);
					procresch();
				}
				
				/* output proc ready: now send an exit message to son */
				
				link[i]->output_header.kind=RoutExit;
				sendlink(i,&link[i]->output_header,sizeof(PACKET_BUFFER_HEADER));
			}
		}
	}
	/* now all links are killed */
}

VOID kill_router(VOID)
{
	router_running=False;
	
	kill_input_processes();
	kill_output_processes();
	
	/* empty transputer queues, prio should be high */
	
	sthf(NotProcessP);	
	stlf(NotProcessP);

	kill_sons();
	
	if (node_info.id != RootId)
		procstop();
}
